package de.unisb.prog.typestate;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.Map.Entry;

import de.unisb.prog.analysis.AnalysisContext;
import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.Call;
import de.unisb.prog.jfirm.ir.MethodGraph;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.typestate.IViolationPoint.Severity;
import de.unisb.prog.typestate.violation.CallViolationInfo;
import de.unisb.prog.typestate.violation.IViolationInfo;
import de.unisb.prog.typestate.violation.MissingSpecInfo;

public class AnalysisResults implements IAnalysisResults, IViolationStore {
	private static final char			VIOSTORE_LOG_TAG	= 'X';
	private static final AnalysisLogger	log					= AnalysisLogger.getInstance();

	/**
	 * Private class for caching reported violations
	 */
	private class CallVioStub implements Comparable<CallVioStub> {

		ContextAndNode	allocSite;
		Set<IState>		oldState, newState;
		boolean			hasBad;
		IViolationInfo	callVioInfo;
		private String	className;
		private boolean	hasUnspecd;

		/**
		 * Stub for violations from CALLs.<br>
		 * <strong>Warning:</strong> Better do not use this in Hash-Collections because of the
		 * {@link #equals(Object)} implementation.
		 * 
		 * @param className
		 *            the class in which the affected call is located
		 * @param allocSite
		 *            the affected allocationSite
		 * @param oldState
		 *            state of instance before call
		 * @param newState
		 *            state of instance after call
		 * @param hasBad
		 *            if true, at least one of the transitions was an error transition
		 * @param hasUnspecd
		 *            if true, at least one of the transitions was not specified in the type's specification
		 * @param callVioInfo
		 *            {@link CallViolationInfo} giving pre/post states and potential corrections
		 */
		public CallVioStub(String className, ContextAndNode allocSite, boolean hasBad, boolean hasUnspecd,
				CallViolationInfo callVioInfo) {
			this.className = className;
			this.allocSite = allocSite;
			this.oldState = callVioInfo.getPreState();
			this.newState = callVioInfo.getPostState();
			this.hasBad = hasBad;
			this.hasUnspecd = hasUnspecd;
			this.callVioInfo = callVioInfo;
		}

		/**
		 * Only considers the instance.<br>
		 * For equality of <code>oldState</code> and <code>newState</code> use:
		 * <code>this.equals(other) && this.compareTo(other) == 1 && other.compareTo(this) == 1</code>
		 */
		public boolean equals(Object o) {
			if (o instanceof CallVioStub) {
				CallVioStub c = (CallVioStub) o;
				return this.allocSite == c.allocSite;
			} else
				return false;
		}

		/**
		 * <strong>Warning: Inconsistent with {@link #equals(Object)}!!</strong>
		 * 
		 * @return -1 if this stub contains LESS information than the argument (
		 *         <code>this.oldState &lt; o.oldState && this.newState &lt; o.newState && (!this.hasBad || o.hasBad) && (!o.hasUnspecd || this.hasUnspecd) </code>
		 *         )<br>
		 *         1 if this stub contains MORE information than the argument (
		 *         <code>this.oldState &gt; o.oldState && this.newState &gt; o.newState && (this.hasBad || !o.hasBad) && (o.hasUnspecd || !this.hasUnspecd)</code>
		 *         )<br>
		 *         0 if this and argument stub are not comparable (<code>this.instance != o.instance</code> or
		 *         none of the above predicates hold)
		 */
		public int compareTo(CallVioStub o) {
			if (allocSite.equals(o.allocSite)) {
				if (oldState.containsAll(o.oldState) && newState.containsAll(o.newState)
						&& (!o.hasBad || this.hasBad) && (!o.hasUnspecd || this.hasUnspecd)) {
					// greater or equal
					return 1;
				} else if (o.oldState.containsAll(oldState) && o.newState.containsAll(newState)
						&& (!this.hasBad || o.hasBad) && (o.hasUnspecd || !this.hasUnspecd)) {
					// less (or equal, but caught by previous case)
					return -1;
				}
			}
			// all other cases: instances are not equal or fell through both conditions above
			return 0;
		}

	}

	// private class MissingSpecVioStub {
	// ValueNode action;
	// ContextAndNode allocSite;
	//
	// public MissingSpecVioStub(ValueNode action, ContextAndNode allocSite) {
	// this.action = action;
	// this.allocSite = allocSite;
	// }
	//
	// public boolean equals(Object o) {
	// if (o instanceof MissingSpecVioStub) {
	// MissingSpecVioStub other = (MissingSpecVioStub) o;
	// return this.action == other.action && this.allocSite == other.allocSite;
	// }
	// return false;
	// }
	//
	// public int hashCode() {
	// return action.hashCode() + 37 * allocSite.hashCode();
	// }
	//
	// }

	private static final String							emptyString	= "";

	/** All violation points are stored here */
	private HashMap<String, List<IViolationPoint>>		violationPoints;
	private int											violationCount;

	/**
	 * Call violations of the current execution run. Do NOT replace list with Set because of
	 * {@link CallVioStub#equals(Object)}-implementation
	 */
	private HashMap<ContextAndNode, List<CallVioStub>>	callViolations;

	/**
	 * List of specifications which are required but missing, mapped from Type name to Nodes which placed
	 * requests.
	 */
	// private HashMap<String, Set<MissingSpecVioStub>> missingSpecs;
	private HashSet<String>								missingSpecs;

	/** The current class name and methodGraph signature */
	private String										className;
	private String										methodId;

	private MethodGraph									methodGraph;

	/** If true, missing specification will generate NOSPEC violations */
	private boolean										logNoSpec;
	private HashSet<ContextAndNode>						mutedInstances;

	public AnalysisResults() {
		violationPoints = new HashMap<String, List<IViolationPoint>>();
		callViolations = new HashMap<ContextAndNode, List<CallVioStub>>();
		// missingSpecs = new HashMap<String, Set<MissingSpecVioStub>>();
		missingSpecs = new HashSet<String>();
		className = methodId = emptyString;
		violationCount = 0;
		logNoSpec = false;
		mutedInstances = new HashSet<ContextAndNode>();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.unisb.prog.typestate.IAnalysisResults#getNumberOfViolations()
	 */
	public int getNumberOfViolations() {
		return violationCount;
	}

	public void setNoSpecLogging(boolean on) {
		assert !isInited() : "can only change logNoSpec flag when no run is currently open";
		if (on)
			logNoSpec = true;
		else
			logNoSpec = false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.unisb.prog.typestate.IAnalysisResults#getViolationList(java.lang.String)
	 */
	public List<IViolationPoint> getViolationList(String className) {
		// First check if the class' violations have already been finalized
		if (className.equals(this.className)) {
			throw new IllegalStateException(
					"ViolationStore: Current analysis results have not been finalized yet");
		}
		List<IViolationPoint> vps = violationPoints.get(className);
		if (vps == null)
			return null;
		Collections.sort(vps);
		return Collections.unmodifiableList(vps);
	}

	/**
	 * For importing violations from cache. Does not effect initialization state of the store.
	 * 
	 * @param className
	 * @param vioPoints
	 */
	public void addViolationPoints(String className, Collection<IViolationPoint> vioPoints) {
		List<IViolationPoint> vps = violationPoints.get(className);
		if (vps == null) {
			vps = new ArrayList<IViolationPoint>();
			violationPoints.put(className, vps);
		}
		vps.addAll(vioPoints);
	}

	/**
	 * Initializes a the violation store for a new analysis run by setting the current class name and
	 * methodGraph id. MUST be called prior to storing violations. A previous run must first be finalized
	 * before a new one can be initialized.
	 * 
	 * @param className
	 *            the name of class the current methodGraph belongs to
	 * @param methodId
	 *            the ID of the methodGraph in analysis, for example the methodGraph signature
	 * @return true if the initialization was successful (i.e. it was not initialized before), false if not.
	 */
	public boolean initRun(String className, String methodId, MethodGraph methodGraph) {
		if (isInited() || className == null || methodId == null || methodGraph == null
				|| emptyString.equals(className) || emptyString.equals(methodId)) {
			return false;
		}
		this.className = className;
		this.methodId = methodId;
		this.methodGraph = methodGraph;
		return true;
	}

	/**
	 * Returns the state of the violation store. It is initialized iff it has a className and methodId set.
	 * 
	 * @return whether is initialized for a run or not
	 */
	private boolean isInited() {
		// comparison with == is valid here, as finalizeRun() always sets fields to emptyString
		return !(emptyString == this.className || emptyString == this.methodId);
	}

	/**
	 * Adds all violations from the current run to the final class->violation mapping, and closes the current
	 * run. MUST be called before violations of the current class might be requested.
	 * 
	 * @return true if the current run was finalized, false if the violation store was not initialized for a
	 *         run.
	 */
	public boolean finalizeRun() {
		if (!isInited()) {
			return false;
		}

		// get the current list of violations for this class or create it, if not existent
		List<IViolationPoint> vios = violationPoints.get(className);
		if (vios == null) {
			vios = new ArrayList<IViolationPoint>();
			violationPoints.put(className, vios);
		}

		// add all current violations
		finalizeCallViolations();
		if (logNoSpec) {
			finalizeMissingSpecs();
		}

		// reset class and methodGraph references
		className = methodId = emptyString;

		// Unmute all items
		mutedInstances.clear();
		return true;

	}

	/* ------------------- */
	/* CALL */
	/* ------------------- */

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.unisb.prog.typestate.IViolationStore#addViolation(int, java.lang.String,
	 * de.unisb.prog.typestate.IViolationPoint.Severity)
	 */
	public void addCallViolation(ContextAndNode call, ContextAndNode allocSite, boolean hasBad,
			boolean hasUnspecd, IViolationInfo vioInfo) {
		assert isInited() : "AnalysisResults.addCallViolation: ViolationStore part not inited";
		assert call.getNode() instanceof Call : "not a call node";
		assert vioInfo instanceof CallViolationInfo : "ViolationStore.addCallViolation with !(vioinfo instanceof CallViolationInfo) --> null or wrong type";

		if (mutedInstances.contains(allocSite)) {
			if (log.debug())
				log.debug("ignoring call violation for instance " + allocSite
						+ " because it is muted. Vio would be: " + vioInfo);
			return;
		}
		CallViolationInfo callVioInfo = (CallViolationInfo) vioInfo;

		Set<IState> oldState = callVioInfo.getPreState(), newState = callVioInfo.getPostState();
		if (log.debug())
			log.debug("  VIOLATION for " + call + ": " + oldState + "-->" + newState, VIOSTORE_LOG_TAG);
		if (!isInited()) {
			throw new IllegalStateException("ViolationStore is not initialized");
		}
		List<CallVioStub> cvs = callViolations.get(call);
		if (cvs == null) {
			cvs = new ArrayList<CallVioStub>();
			callViolations.put(call, cvs);
		}

		// The name of the class WHERE the violation occurs can only be retrieved as the class of the current
		// call in the context
		String className;
		try {
			Call lastCall = call.getContext().getLast();
			// get classname from the last call on the stack
			className = lastCall.getTargetMethod().getOwner().toString();
		} catch (NoSuchElementException e) {
			// empty call stack, we're in the current base class
			className = this.className;
		}

		CallVioStub newCallVio = new CallVioStub(className, allocSite, hasBad, hasUnspecd, callVioInfo);

		for (Iterator<CallVioStub> it = cvs.iterator(); it.hasNext();) {
			CallVioStub oldStub = it.next();

			switch (oldStub.compareTo(newCallVio)) {
			case -1: // old < new
				// remove old one
				it.remove();

				// nothing more to do
				continue;
			case 1: // old > new
				// no new information in new violation, so cancel
				return;

			case 0:
			default:
				// not comparable, so go on
			}
		}

		// if this point is reached, the new information was not contained in any of the existing ones
		cvs.add(newCallVio);
	}

	/**
	 * Read stored stubs and convert them into {@link IViolationPoint}s which are then stored.
	 */
	private void finalizeCallViolations() {
		assert isInited() : "AnalysisResults.addCallViolation: ViolationStore part not inited";
		List<IViolationPoint> finalVios = violationPoints.get(className);

		for (Entry<ContextAndNode, List<CallVioStub>> entry : callViolations.entrySet()) {
			ContextAndNode callVersioned = entry.getKey();
			List<CallVioStub> vios = entry.getValue();

			for (CallVioStub cvs : vios) {
				violationCount++;
				//MUST/MAY violation
				Severity severity;
				//if only one item and this one is unspecd, generate unspecd violation
				if (cvs.hasUnspecd && !cvs.hasBad) {
					severity = Severity.UNSPEC;
				} else {
					//hasUnspecd and hasBad can not be 0 simultaneously, therefore
					//in this case hasBad > 0, which dominates hasUnspecd
					
					severity = cvs.newState.size() - (cvs.hasUnspecd?1:0) == 1 ? Severity.FAIL : Severity.WARN;
				}
				
				finalVios.add(new ViolationPoint(cvs.className, cvs.callVioInfo, callVersioned,
						severity, methodGraph));
			}

		}

		// clear current violation set
		callViolations.clear();
	}

	/* ------------------- */
	/* MISSING SPEC */
	/* ------------------- */

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.unisb.prog.typestate.IViolationStore#addMissingSpec(de.unisb.prog.jfirm.ir.ValueNode,
	 * de.unisb.prog.jfirm.ir.ValueNode)
	 */
	public void addMissingSpec(String typeName, ContextAndNode actionNode, ContextAndNode allocSite) {
		assert isInited() : "AnalysisResults.addCallViolation: ViolationStore part not inited";

		// Logging of no spec violations might be disabled
		if (!logNoSpec)
			return;

		if (typeName == null || actionNode == null || allocSite == null) {
			throw new IllegalArgumentException(
					String
							.format(
									"AnalysisResults.addMissingSpec(typeName==%s, actionNode==%s, instance==%s): no null allowed",
									typeName, actionNode, allocSite));
		}

		missingSpecs.add(typeName);
		// Set<MissingSpecVioStub> missVios = missingSpecs.get(typeName);
		// if (missVios == null) {
		// // first request for typeName
		// missVios = new HashSet<MissingSpecVioStub>();
		// missingSpecs.put(typeName, missVios);
		// }
		//
		// missVios.add(new MissingSpecVioStub(actionNode, instance));

	}

	/**
	 * Read stored stubs and convert them into {@link IViolationPoint}s which are then stored.
	 */
	public void finalizeMissingSpecs() {
		assert isInited() : "AnalysisResults.addCallViolation: ViolationStore part not inited";
		List<IViolationPoint> finalVios = violationPoints.get(className);

		// for (Entry<String, Set<MissingSpecVioStub>> entry : missingSpecs.entrySet()) {
		// for (MissingSpecVioStub stub : entry.getValue()) {
		// MissingSpecInfo missingInfo = new MissingSpecInfo(stub.instance, entry.getKey());
		// //Use the current context because spec violations are only counted per analysis-run
		// ViolationPoint vp = new ViolationPoint(className, missingInfo, stub.action, Severity.INFO,
		// methodGraph,
		// context);
		// finalVios.add(vp);
		// }
		// }
		for (String type : missingSpecs) {
			MissingSpecInfo missingInfo = new MissingSpecInfo(null, type);

			// FIXME: Best approximation of the first line in the methodGraph...
			ContextAndNode topNode = new ContextAndNode(new AnalysisContext(), (ValueNode) methodGraph
					.getParameters()[0].usersIterator().next());

			// Use the current context because spec violations are only counted per analysis-run
			ViolationPoint vp = new ViolationPoint(className, missingInfo, topNode, Severity.NOSPEC,
					methodGraph);
			finalVios.add(vp);
			violationCount++;
		}

		missingSpecs.clear();
	}

	/* ------------------- */
	/* NULL PTR DEREF */
	/* ------------------- */

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.unisb.prog.typestate.IViolationStore#addNullPtrDerefViolation(de.unisb.prog.jfirm.ir.ValueNode,
	 * de.unisb.prog.jfirm.ir.ValueNode, boolean)
	 */
	public void addNullPtrDerefViolation(ContextAndNode actionNode, ContextAndNode allocSite,
			boolean mustViolation) {
		assert isInited() : "AnalysisResults.addCallViolation: ViolationStore part not inited";
		log.debug("AnalysisResults.addNullPtrDerefViolation(..): unimplemented!!");
		log.info("Found null pointer dereferencing violation but handling it is unimplemented (Call: "
				+ actionNode + ", on node:" + allocSite + ", must-violation: " + mustViolation + ")");

	}

	public void disableViolationsForInstance(ContextAndNode instance) {
		if (instance == null) {
			throw new IllegalArgumentException(
					"AnalysisResults.disableViolationsForInstance: Illegal instance: null");
		}
		mutedInstances.add(instance);
	}

	public void enableViolationsForInstance(ContextAndNode instance) {
		if (instance == null) {
			throw new IllegalArgumentException(
					"AnalysisResults.enableViolationsForInstance: Illegal instance: null");
		}
		mutedInstances.remove(instance);
	}

	public boolean isDisabled(ContextAndNode actualNode) {
		return mutedInstances.contains(actualNode);
	}

}
