package de.unisb.prog.typestate;

import java.net.URLClassLoader;
import java.util.Arrays;

import org.apache.log4j.Logger;

import de.unisb.prog.jfirm.ir.MethodGraph;
import de.unisb.prog.typestate.ir.FirmWalker;

/**
 * The analysis can be run by using an instance of this class.
 * 
 * @author mikerofone
 * 
 */
public class AnalysisRunner implements IAnalysis {

	/**
	 * Runs the analysis on the given {@link IInputDataSet} and returns the {@link IAnalysisResults}
	 * afterwards.
	 * 
	 * @param inputData
	 *            The input data set.
	 * @return The analysis results
	 * @throws InterruptedException 
	 */
	public final IAnalysisResults run(IInputDataSet inputData, Logger logger) {

		// **
		// init logger
		// **
		AnalysisLogger.setLogger(logger);
		logger = AnalysisLogger.getInstance().log;
		logger.info("Analysis started.");
		if (AnalysisLogger.getInstance().info()) {
			URLClassLoader classLoader = (URLClassLoader) ClassLoader.getSystemClassLoader();
			logger.info("Classpath used: " + Arrays.toString(classLoader.getURLs()));
		}

		// **
		// init result list
		// **
		AnalysisResults results = new AnalysisResults();
		// configure violation store
		results.setNoSpecLogging(inputData.getConfig().logUnknownSpecification());

		// Create a specification repository
		SpecificationRepository specRep = new SpecificationRepository(inputData);

		// **
		// init cache provider
		// **
		// Initialize the cache logic with the given cache provider
		ICacheLogic cache = getCacheLogic(inputData.getCacheProvider());

		// The jFirm construction unit
		FirmConnector firmConn = FirmConnector.getInstance();
		firmConn.init(inputData);

		// The FirmWalker, which will process each jFirm Graph
		FirmWalker fw = new FirmWalker(specRep, results, inputData.getConfig());

		// the class whose methods are currently processed
		String currentClassName = "";

		// **
		// for each method in inputset (inputData returns methods sorted by class!)
		// **
		for (IMethodIdentifier methodIdent : inputData.getAllMethods()) {

			logger.info("Current method: " + methodIdent.getMethodId());

			// **
			// if in new class
			// **
			if (!currentClassName.equals(methodIdent.getClassName())) {

				logger.trace(" Reading class " + methodIdent.getClassName());

				currentClassName = methodIdent.getClassName();

				// **
				// init cache logic for this class
				// **
				cache.useClass(inputData.getBytecodeStream(currentClassName));

			}
			
			// **
			// if cacheprovider has cached version
			// **
			if (cache.hasCachedEntry(methodIdent)) {

				logger.trace(" Retrieving cached result from cache");

				// **
				// add cached results to result list
				// **
				results.addViolationPoints(currentClassName, cache.getCachedResults(methodIdent));
			}

			// **
			// if not
			// **
			else {

				logger.trace(" Cache miss, running analysis for " + methodIdent.getMethodName());

				// **
				// build firm graph for method
				// **

				MethodGraph methodGraph = firmConn
						.getFirmGraph(currentClassName, methodIdent.getMethodName(), methodIdent.getSignature());
				if (methodGraph == null) {
					// MethodGraph not found
					logger.error("MethodGraph '" + methodIdent.getMethodId() + "' could not be built, skipping");
					continue;
				}

				// **
				// start analysis for method (the actual work)
				// **

				// init IViolationStore part of AnalysisResults
				results.initRun(currentClassName, methodIdent.getMethodId(), methodGraph);

				// logger.setLevel(Level.INFO);

				// Run the analysis: Reset walker and then walk over jFirm graph
				fw.reset();
				try {
					fw.walk(methodGraph);
				} catch (Exception e) {
					logger.error("Could not analyse method " + methodGraph.getName() + ": " +e);
				}

				// logger.setLevel(Level.ALL);

				// finalize current run in IViolationStore part of AnalysisResults
				results.finalizeRun();

				// **
				// add analysis results to cache provider
				// **
				cache.cacheResults(methodIdent, results.getViolationList(currentClassName));
			}
			logger.trace("\n\nMethod Analysis ends\n\n\n\n\n");
		}

		// **
		// return result set
		// **
		logger.info(String.format("Analysis ends."));
		return results;
	}

	/**
	 * Returns the {@link ICacheLogic} which shall be used for the analysis
	 * 
	 * @param cacheProvider
	 *            The cache provider which enables the logic to write out and read objects.
	 * @return The initialized cache logic, or, if <code>cacheProvider</code> is <code>null</code>, a
	 *         {@link DummyCacheLogic} which just disables caching completely
	 */
	private ICacheLogic getCacheLogic(ICacheProvider cacheProvider) {

		// test if cache provider actually specified
		if (cacheProvider == null) {
			// none given, so use default dummy implementation
			return new DummyCacheLogic();
		}

		// TODO actually use some real caching logic
		return new DummyCacheLogic(cacheProvider);
	}
}
