package de.unisb.prog.typestate;

import java.util.HashMap;

import de.unisb.prog.typestate.IAnalysisConfiguration.TopStateMode;
import de.unisb.prog.typestate.IAnalysisConfiguration.UnknownTransitionMode;
import edu.uci.ics.jung.graph.DirectedGraph;

/**
 * This class parses JUNG {@link DirectedGraph}s and converts them to the internal, more time/space efficient
 * finite state automata implementation.
 * 
 * @author mikerofone
 * 
 */
public abstract class FSAParser {

	public static Specification parseGraph(ITypeStateAutomaton tsa, UnknownTransitionMode transMode, TopStateMode topMode) {

		if (tsa == null) {
			throw new IllegalArgumentException("FSAParser.parse(null)");
		}

		DirectedGraph<String, ITransition> graph = tsa.getAutomaton();
		HashMap<String, IState> states = new HashMap<String, IState>(graph.getVertexCount());
		
		// The specification object which is to be returned
		Specification spec = new Specification();
		
		// Build and mark the error State
		State errorState = new State(tsa.getErrorState(),spec);
		errorState.setBad(true);
		states.put(tsa.getErrorState(), errorState);

		// Build the state for signaling unspecified transitions
		UnspecifiedState unspecState = new UnspecifiedState(spec);
		
		// Create States
		for (String vert : graph.getVertices()) {
			State state = new State(vert, spec);

			//Error state is already in map
			if (!errorState.equals(state)) {
				states.put(vert, state);
			}
			
			// Set the default transition according to the selected mode
			switch (transMode) {
			case ERROR_TRANSITION:
				// transition to error
				state.setDefaultTransition(errorState);
				break;
			case SAME_STATE:
				// stay in same state
				state.setDefaultTransition(state);
				break;
			case REPORT_UNSPEC:
				// transition to signaling state
				state.setDefaultTransition(unspecState);
				break;
			default:
				throw new IllegalArgumentException("FSAParser.parseGraph: Unknown Mode specified: " + transMode);
			}
		}

		// Link transitions
		for (ITransition trans : graph.getEdges()) {

			// The source and destination state are the same for all outgoing transitions
			State sourceState = (State) states.get(trans.getSourceName());
			State destState = (State) states.get(trans.getDestinationName());

			// Check if this is a regular transition or a default transition specification (overriding the
			// default transition set according to mode)
			if (trans.getMarking().isEmpty()) {
				// default transition
				sourceState.setDefaultTransition(destState);
			} else {
				// regular transition

				// For each method firing this transition:
				// Link source to destination state and mark transition with method name + signature
				// Put destination state into transitionTargets-mapping
				for (IMethodIdentifier meth : trans.getMarking()) {

					// Linking source to dest
					String transName = meth.getMethodName() + meth.getSignature();
					sourceState.extend(transName, destState);
				}
			}
		}

		// Fill data into the Specification
		spec.init(tsa.getName(), states, tsa.getInitState(), tsa.getErrorState(), topMode);

		// Return the specification
		return spec;
	}
}
