package de.unisb.prog.typestate;

/**
 * Classes implementing this interface combine several settings and values which tweak the analysis.
 * 
 * @author mikerofone
 * 
 */
public interface IAnalysisConfiguration {

	/**
	 * This enum determines how the TOP state is handled in case a transition must be performed.
	 * 
	 * @author mikerofone
	 * 
	 */
	public enum TopStateMode {

		/**
		 * If transition is performed, every possible state performs the given transition. Using this mode is
		 * required for a sound analysis but will likely produce massively many false negatives.
		 */
		ASSUME_EVERY_STATE,

		/**
		 * If transition is performed, every possible state performs the given transition, but error states
		 * will be ignored. This makes the analysis unsound but also dramatically reduce the amount of false
		 * negatives.
		 */
		FIRST_ALWAYS_OK
	}

	/**
	 * Controls the way how an element found in TOP state will be handled during transitions. See
	 * {@link TopStateMode} for details.
	 * 
	 * @return the corresponding {@link TopStateMode} value
	 */
	public TopStateMode getTopStateMode();

	/**
	 * This enum determines how a request for a transition unknown to the current state in the specification
	 * will be handled.
	 * 
	 * @author mikerofone
	 * 
	 */
	public enum UnknownTransitionMode {

		/**
		 * An unknown transition does not change the current state (equivalent to ignoring it). This option
		 * will make the analysis unsound as an unknown transition might also perform an error transition.
		 */
		SAME_STATE,

		/**
		 * An unknown transition takes the instance to the ERROR state, regardless what state it currently is
		 * in. This is required to have a sound analysis, but results and rate of false detections strongly
		 * depend on the specification quality and completeness.
		 */
		ERROR_TRANSITION,

		/**
		 * An unknown transition generates a distinct violation explicitly reporting that an unspecified
		 * transition was fired.
		 */
		REPORT_UNSPEC
	}

	/**
	 * Controls the behaviour of the analysis when an instance is asked to perform a transition which is not
	 * known in the current state.
	 * 
	 * @return the corresponding {@link UnknownTransitionMode} value
	 */
	public UnknownTransitionMode getUnknownTransitionMode();

	/**
	 * This value determines the maximum depth of the interprocedural analysis for method invocations.
	 * 
	 * @return the maximum depth of nested methods
	 */
	public int getMaxDepth();

	/**
	 * Returns whether there should be violations generated if there is no specification for a type or not.
	 * 
	 * @return true, if logging of missing specification should be enabled
	 */
	public boolean logUnknownSpecification();

	public boolean walkerUseOnlyOneWorklist();
}
