package de.unisb.prog.typestate;

import java.io.InputStream;
import java.io.Serializable;
import java.util.List;

import org.apache.log4j.Logger;

/**
 * This interface provides the necessary mechanisms for caching analysis results. A cache provider must
 * support the following operations:<br>
 * <ul>
 * <li>Parse the bytecode of all methods of a class to create some fingerprint, hashcode etc. which can be<br>
 * <ul>
 * <li>either compared to previously analyzed versions of a method for finding out whether there were changes
 * or not,</li>
 * <li>or, if there was a change, stored as a reference against which later versions of the method are
 * compared.</li>
 * </ul>
 * </li>
 * <li>Restore the previously stored results for a version of a method, or report that there are no cached
 * results.</li>
 * <li>Store a set of analysis results for the current version of the method.</li>
 * </ul>
 * For permanently storing data, the cache logic can use an {@link ICacheProvider} which can write and read
 * {@link Serializable} objects to some permanent memory.<br>
 * In addition, the cache provider can output messages to a <code>log4j</code> logger given by a call to
 * {@link setLogger}.
 * 
 * @author mikerofone
 * 
 */
public interface ICacheLogic {

	/**
	 * Sets a <code>log4j</code> logger where messages related to caching can be output.
	 * 
	 * @param logger
	 *            The logger to use
	 */
	public void setLogger(Logger logger);

	/**
	 * Tells the cache provider that from now on methods from the specified class will be processed. Upcoming
	 * cache queries will relate to this class and its methods. This method must be called prior to any other
	 * cache-related method!
	 * 
	 * @param classCode
	 *            a stream providing the bytecode of the desired class
	 */
	public void useClass(InputStream classCode);

	/**
	 * Determines if there is a cached version of the analysis for the given method is available or not.
	 * <p>
	 * <strong>Note:</strong> Prior to calling this method, the CacheProvider must have been initialized for
	 * the current class by {@link useClass}. Otherwise, the output and behavior of the method is undefined.
	 * 
	 * @param method
	 *            The method for which the presence of a cached entry is queried.
	 * @return True if there is a cached version available, false if not or if not inited.
	 */
	public boolean hasCachedEntry(IMethodIdentifier method);

	/**
	 * Returns the cached set of {@link IAnalysisResults} for this version of the given method.
	 * <p>
	 * <strong>Note:</strong> Prior to calling this method, the CacheProvider must have been initialized for
	 * the current class by {@link useClass}. Otherwise, the output and behavior of the method is undefined.
	 * 
	 * @param method
	 *            The method for which the cached analysis results are queried.
	 * @return A set of cached analysis results, or <code>null</code> if no cached version is available
	 */
	public List<IViolationPoint> getCachedResults(IMethodIdentifier method);

	/**
	 * Stores a set of analysis results for the current version of the given method.
	 * <p>
	 * <strong>Note:</strong> Prior to calling this method, the CacheProvider must have been initialized for
	 * the current class by {@link useClass}. Otherwise, the output and behavior of the method is undefined.
	 * 
	 * @param method
	 *            The method for which analysis results should be stored.
	 * @param methodAnalysisResults
	 *            The set of {@link IViolationPoints} to put into the cache.
	 */
	public void cacheResults(IMethodIdentifier method, List<IViolationPoint> foundViolations);
}
