package de.unisb.prog.typestate;

import java.io.InputStream;
import java.util.List;

/**
 * Classes implementing this interface can be used as input sets for the analysis.<br>
 * <br>
 * The analysis needs the following information to run:<br>
 * <ul>
 * <li>A list of {@link IMethodIdentifier}s determining all methods which are to be analyzed.</li>
 * <li>A mapping from class names to InputStreams providing the bytecode of the classes. Such a stream must be
 * available for every class of which a member method is to be analyzed.</li>
 * <li>A set containing all type names for which specification automata exist.</li>
 * <li>A mapping from type names to {@link ITypeStateAutomaton}s which are the specification FSA for the
 * corresponding type.
 * </ul>
 * Additionally, an {@link ICacheProvider} can be specified which is used to cache analysis results for later
 * analysis runs, if the underlying methods have not changed. Caching is disabled completely if no
 * {@link ICacheProvider} is specified.
 * <p>
 * <strong>Note:</strong> The analysis can be sped up if the list of class files only contains classes, whose
 * methods have been changed. Either this should be considered if this information is available when starting
 * the analysis (e.g. derived from the command line of the compiler in scenarios like IDE integration), or one
 * should specify a cache provider which can decide this at runtime.
 * 
 * @author mikerofone
 * 
 */
public interface IInputDataSet {

	/**
	 * Gets an InputStream returning the bytecode of the specified class.
	 * 
	 * @param method
	 *            The fully qualified Java name of the desired class.
	 * @return An InputString returning the class' bytecode.
	 */
	public InputStream getBytecodeStream(String className);

	/**
	 * Gets a List of {@link IMethodIdentifier}s for all methods which are to be analyzed.
	 * 
	 * @return A List containing the method's {@link IMethodIdentifier}, while methods of the same class are
	 *         grouped together.
	 */
	public List<IMethodIdentifier> getAllMethods();

	/**
	 * Gets the {@link ICacheProvider} specified for this analysis run, enabling caching of analysis results
	 * for future runs.<br>
	 * If no cache provider is specified, caching will be completely disabled.
	 * 
	 * @return The {@link ICacheProvider} which shall be used for caching, or <code>null</code> if caching is
	 *         disabled.
	 */
	public ICacheProvider getCacheProvider();

	/**
	 * Gets a specification automaton for the given type.
	 * 
	 * @param type
	 *            The fully qualified Java name of the requested type (for example "java/util/HashMap$Entry").
	 * @return The specification {@link ITypeStateAutomaton}, or <code>null</code> if there is no spec for
	 *         this type.
	 */
	public ITypeStateAutomaton getSpecification(String type);

	/**
	 * Gets the names of all types for which specification automata exist.
	 * 
	 * @return A list of fully qualified Java typenames.
	 */
	public List<String> getAllSpecNames();
	
	/**
	 * Gets the configuration for this analysis run, which controls some parameters of the analysis' behaviour.
	 * @return the {@link IAnalysisConfiguration} object
	 */
	public IAnalysisConfiguration getConfig();
}
