package de.unisb.prog.typestate;

/**
 * Describes a called method which is about to be analyzed. The description contains the following
 * information:<br>
 * <ul>
 * <li>The name of the MethodGraph</li>
 * <li>The name of the class it's a member of</li>
 * <li>The signature of the method (number of arguments and their types)</li>
 * <li>The return type of the method</li>
 * </ul>
 * <code>equals()</code> and <code>hashCode()</code> must be overwritten to allow comparing methods by their
 * origin and signature (meaning containing class, method name, number and types of parameters)
 * 
 * @author mikerofone
 * 
 */
public interface IMethodIdentifier {

	/**
	 * Gets the method's name.
	 * 
	 * @return The method name
	 */
	public String getMethodName();

	/**
	 * Gets the fully qualified class name which this method is a member of.
	 * 
	 * @return The name formatted as
	 *         <code>&lt;package&gt;/.../&lt;package&gt;/&lt;class&gt;$&lt;subclass&gt;</code>
	 */
	public String getClassName();

	/**
	 * Gets the method signature as specified by the java virtual machine specification.
	 * 
	 * @return The encoded signature
	 * @see http://java.sun.com/docs/books/jvms/second_edition/html/Concepts.doc.html#34442
	 */
	public String getSignature();

	/**
	 * Returns the hash code based on the name of the containing class and the method's signature (name,
	 * #parameters and their types)
	 * 
	 * @return the hashcode
	 */
	@Override
	public int hashCode();

	/**
	 * Compare to another object
	 * 
	 * @param o
	 *            the other object
	 * @return true, if <code>o</code> is an {@link IMethodIdentifier} and has the same class, method name and
	 *         signature.
	 */
	@Override
	public boolean equals(Object o);

	/**
	 * Return an id for this method which identifies it <strong>uniquely</strong> in this class, i.e. a
	 * combination of method name and signature.
	 * 
	 * @return a unique id for this method in this class
	 */
	public String getMethodId();

}
