package de.unisb.prog.typestate;

import java.util.Map;
import java.util.Set;

/**
 * Every instance of this class represents a one concrete state of one concrete type. Every <code>State</code>
 * has a name which also serves as the <code>State</code>'s id. Other than being unique, the names do not have
 * any further meaning or requirements. The instance also contains the named transitions to the
 * <code>State</code>s which are reachable from this <code>State</code>.<br>
 * A <code>State</code> instance is immutable, so it can be used in an unlimited number of contexts without
 * any side effects.<br>
 * <code>State</code>s must implement hashCode() and equals() such that they can be compared by the type and
 * their name.
 * 
 * @author mikerofone
 * 
 */
public interface IState {

	/**
	 * Indicates whether the state is bad or not
	 * 
	 * @return true if it's bad.
	 */
	public boolean isBad();

	/**
	 * Get the name of the state (for debugging purposes)
	 * 
	 * @return The name of the state
	 */
	public String getName();

	/**
	 * @return the {@link Specification} to which this State belongs or <code>null</code>, if this is an
	 *         abstract state not belonging to any concrete specification
	 */
	public Specification getTypeSpecification();

	/**
	 * Return the resulting <code>IState</code>(s) from performing the transition labeled 'action'.
	 * 
	 * @param action
	 *            The name of the transition to perform. The transition name "" (empty string) is reserved for
	 *            the default transition which is fired if the specified one is unknown.
	 * @return A Set of states reached by following the specified transition, if it's known to the state. If
	 *         it's not, then either the default transition with name "" is taken (if specified) or the
	 *         current state is returned.
	 */
	public Set<IState> performTransition(String action);

	/**
	 * Returns all transitions which can be taken from this <code>IState</code>.
	 * 
	 * @return a mapping from transition to sets of <code>IState</code>
	 */
	public Map<String, Set<IState>> getTransitions();

	/**
	 * Override hashCode such that it's based on the type- and the <code>IState</code>'s name.
	 */
	@Override
	public int hashCode();

	/**
	 * Equals method which only depends on the type- and the <code>IState</code>'s name.
	 */
	@Override
	public boolean equals(Object o);

}
