package de.unisb.prog.typestate;

import edu.uci.ics.jung.graph.DirectedGraph;

/**
 * Every instance of this class represents a typestate specification for one type. A typestate specification
 * is a finite state machine (FSA) with unnamed states (i.e. the names have no semantics but may be used for
 * messages) and named transitions representing function calls. The FSA's name is that of the type it stands
 * for and it has a designated initial state and one error state.
 * <p>
 * Every state may define a default transition marked with an empty set. The default transition is fired if
 * the desired one is not specified for the current state and can be used to implement white- or blacklists
 * for actions, e.g. "all transitions except close() lead to the error state".<br>
 * If no default action is specified in the automaton, the outcome of a transition request is defined by
 * {@link IAnalysisConfiguration#getUnknownTransitionMode()} called on the config object in the given
 * {@link IInputDataSet}.
 * <p>
 * The FSA are <strong>non-multi</strong> {@link DirectedGraph}s (JUNG, {@link http://jung.sourceforge.net/}),
 * whose nodes are marked with <code>String</code> and the edges with
 * {@link ITransition}, allowing multiple actions taking the same transition.
 * 
 * @author mikerofone
 * 
 */
public interface ITypeStateAutomaton {

	/**
	 * Gets the name of the FSA determining the type for which it is the specification.
	 * 
	 * @return The name of the automaton
	 */
	public String getName();

	/**
	 * Gets the name of the initial state of the FSA.
	 * 
	 * @return The name of the initial state
	 */
	public String getInitState();

	/**
	 * Gets the name of the bad/error state of the FSA.
	 * 
	 * @return The name of the bad/error state
	 */
	public String getErrorState();

	/**
	 * Returns the whole automaton as a {@link DirectedGraph}.
	 * 
	 * @return The complete automaton
	 */
	public DirectedGraph<String, ITransition> getAutomaton();

}
