package de.unisb.prog.typestate;

import java.util.List;

import de.unisb.prog.typestate.violation.IViolationInfo;

/**
 * Represents one point in the analyzed program where a potential typestate violation was found. It provides
 * the following details about the violation:
 * <ul>
 * <li>The name of the class.</li>
 * <li>The line number in the source code, where the violation was detected.</li>
 * <li>A textual description of the violation.</li>
 * <li>The severity of the violation.</li>
 * </ul>
 * <code>IViolationPoint</code>s must implement {@link Comparable} lexicoprahically, the first criterion being
 * the class name and the second being the line number. Violation points of the same class are thus put
 * together (the order of the classes may be arbitrary) and those in the same class are ordered by their
 * appearance in the code.
 * 
 * @author mikerofone
 * 
 */
public interface IViolationPoint extends Comparable<IViolationPoint> {

	/**
	 * Determines the type of the violation:
	 * <p>
	 * <code>INFO</code> - only informative (remarks etc, no violation)<br>
	 * <code>WARN</code> - warning: for example, if the analysis is too weak to decide whether there might be
	 * a circumstance where this is in fact a violation<br>
	 * <code>FAIL</code> - real violation<br>
	 * <code>UNSPEC</code> - the transition fired was unspecified in the type's specifiation<br>
	 * <code>NOSPEC</code> - for some type there is no specification given
	 * 
	 * @author mikerofone
	 * 
	 */
	public enum Severity {
		INFO, WARN, FAIL, UNSPEC, NOSPEC;
	}

	/**
	 * @return The fully qualified Java class name where the violation was found.
	 */
	public String getClassName();

	/**
	 * Gets the code line number of the violation, if known.
	 * 
	 * @return The line number where the violation was found in the class' source code, or <code>-1</code> if
	 *         it is unknown.
	 * 
	 */
	public int getLineNumber();
	
	/**
	 * Gives information about which method was being analyzed when the violation occurred.
	 * @return the state of the method call stack (the call hierarchy) when the violation was detected.
	 */
	public List<String> getLocation();

	/**
	 * @return Information about the violation as a list of {@link IViolationInfo}
	 */
	public List<IViolationInfo> getInfo();

	/**
	 * Returns the severity of the violation.
	 * 
	 * @return The {@link Severity} value
	 */
	public Severity getSeverity();

	/**
	 * Compares violation points 1. by class name and 2. by line number in the source code.
	 * 
	 * @param other
	 *            the other <code>IViolationPoint</code>
	 * @return The result of lexicographic comparison of class name and line number according to the
	 *         guidelines in {@link Comparable}.
	 */
	public int compareTo(IViolationPoint other);
}
