package de.unisb.prog.typestate;

import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.Call;
import de.unisb.prog.jfirm.ir.MethodGraph;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.typestate.violation.IViolationInfo;

/**
 * Used to store Violation points during the analysis phase walking over the graph.
 * 
 * @author mikerofone
 */
public interface IViolationStore {

	/**
	 * Initializes a the violation store for a new analysis run by setting the current class name and method
	 * id. MUST be called prior to storing violations. A previous run must first be finalized before a new one
	 * can be initialized.
	 * 
	 * @param className
	 *            the name of class the current method belongs to
	 * @param methodId
	 *            the ID of the method in analysis, for example the method signature
	 * @param methodGraph
	 *            the jFirm method object belonging to the above id
	 * @return true if the initialization was successful (i.e. it was not initialized before), false if not.
	 */
	public boolean initRun(String className, String methodId, MethodGraph methodGraph);

	/**
	 * Finalizes the current run. MUST be called before violations of the current class might be requested.
	 * 
	 * @return true if the current run was finalized, false if the violation store was not initialized for a
	 *         run.
	 */
	public boolean finalizeRun();

	/**
	 * Reports a violation during execution of a {@link Call}.
	 * 
	 * @param callNode
	 *            the {@link Call} Node which caused the violation
	 * @param instance
	 *            the instance {@link ValueNode} which performed a transition to a bad state
	 * @param hasBad
	 *            if true, at least one of the states in oldState perform a transition to the error state
	 * @param hasUnspecd
	 * 			  if true, the performed transition was not specified for at least one of the states in oldState 
	 * @param suggestInfo
	 *            {@link IViolationInfo} containing fix callVioInfo
	 */
	public void addCallViolation(ContextAndNode callNode, ContextAndNode instance, boolean hasBad,
			boolean hasUnspecd, IViolationInfo suggestInfo);

	/**
	 * Reports a null pointer dereferencing violation.
	 * 
	 * @param actionNode
	 *            the action which dereferenced the null pointer
	 * @param instance
	 *            the instance which is null at the moment
	 * @param mustViolation
	 *            if true, <code>instance</code> can ONLY be null at this moment. if false, it MIGHT be null
	 *            or instantiated
	 */
	public void addNullPtrDerefViolation(ContextAndNode actionNode, ContextAndNode instance,
			boolean mustViolation);

	/**
	 * Reports that specification for a type could not be found.
	 * 
	 * @param typeName
	 *            the fully qualified type name for which no spec is found
	 * @param actionNode
	 *            the action which requested the unknown specification
	 * @param instance
	 *            the instance whose type is unknown
	 */
	public void addMissingSpec(String typeName, ContextAndNode actionNode, ContextAndNode instance);

	/**
	 * Enables or disables logging if an encountered type has no specification. It defaults to off. This can
	 * only be changed before {@link #initRun(String, String, MethodGraph)} or after {@link #finalizeRun()},
	 * otherwise an exception will be thrown.
	 * 
	 * @param on if <code>true</code>, enables logging of no specs.
	 */
	public void setNoSpecLogging(boolean on);

	public void disableViolationsForInstance(ContextAndNode instance);
	public void enableViolationsForInstance(ContextAndNode instance);
	public boolean isDisabled(ContextAndNode actualNode);
}
