package de.unisb.prog.typestate;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Map.Entry;

import org.softevo.util.source.ClassFileSource;
import org.softevo.util.source.ClassPathClassFileSource;

import de.unisb.prog.typestate.io.TypeStateReader;
import edu.uci.ics.jung.graph.DirectedGraph;

public class InputDataSet implements IInputDataSet {

	private static final String						START_STATE	= "start";

	private static final String						ERR_STATE	= "ex";

	private List<IMethodIdentifier>					methods;

	private ICacheProvider							cache;

	private List<ClassPathClassFileSource>					classes;

	private IAnalysisConfiguration					config;

	private HashMap<String, ITypeStateAutomaton>	specs;

	private AnalysisLogger							log;

	public InputDataSet() {
		log = AnalysisLogger.getInstance();
		methods = new ArrayList<IMethodIdentifier>();
		specs = new HashMap<String, ITypeStateAutomaton>();
		cache = null;
		config = new AnalysisConfiguration();
		classes = new ArrayList<ClassPathClassFileSource>();
	}

	public List<IMethodIdentifier> getAllMethods() {
		return Collections.unmodifiableList(methods);
	}

	public List<String> getAllSpecNames() {
		return new ArrayList<String>(specs.keySet());
	}

	public InputStream getBytecodeStream(String className) {
		for (ClassFileSource src : classes) {
			if (src.containsClass(className)) {
				try {
					return new ByteArrayInputStream(src.getClassForName(className).getClassBytes());
				} catch (NoSuchElementException e) {
					// Do nothing, search in next
					continue;
				} catch (IOException e) {
					log.warn("I/O error while reading bytecode for class " + className + ": "
							+ e.getMessage());
				}
			}

		}

		log.warn("Could not retrieve bytecode for class " + className + " because it was not found.");
		// Some error
		return null;

	}

	public ICacheProvider getCacheProvider() {
		return cache;
	}

	public ITypeStateAutomaton getSpecification(String type) {
		return specs.get(type);
	}

	public IAnalysisConfiguration getConfig() {
		return this.config;
	}

	/**
	 * Adds a method to the methods list. Before a method can be added, the class containing the method must
	 * have been added by {@link addClass}.
	 * 
	 * @param method
	 *            The method to add
	 * @return true, if adding was sucessful, false if not (class not already added or method already in the
	 *         set)
	 */
	public boolean addMethod(IMethodIdentifier method) {
		if (method == null || methods.contains(method)) {
			// illegal args or method already known
			return false;
		}
		return methods.add(method);
	}

	/**
	 * Sets the cache provider.
	 * 
	 * @param cache
	 *            the cache provider to set
	 */
	public void setCache(ICacheProvider cache) {
		this.cache = cache;
	}

	public void setConfig(IAnalysisConfiguration config) {
		this.config = config;
	}

	public void readTypeStateAutomata(String specificationFolderName) throws FileNotFoundException {
		File specificationFolder = new File(specificationFolderName);
		if (!specificationFolder.isDirectory()) {
			throw new FileNotFoundException("Error reading specification: File " + specificationFolder
					+ " does not exist or is not a directory.");
		}

		if (specs == null) {
			specs = new HashMap<String, ITypeStateAutomaton>();
		}
		File[] specificationFiles = specificationFolder.listFiles();
		for (File file : specificationFiles) {
			if (file.getName().endsWith(".xml")) {
				try {
					Map<String, DirectedGraph<String, ITransition>> graphs = TypeStateReader
							.readTypeStates(file);
					for (Entry<String, DirectedGraph<String, ITransition>> entry : graphs.entrySet()) {
						String className = entry.getKey();
						DirectedGraph<String, ITransition> graph = entry.getValue();
						ITypeStateAutomaton automaton = new TypeStateAutomaton(className, graph, START_STATE,
								ERR_STATE);
						specs.put(className, automaton);
					}
				} catch (Exception readException) {
					AnalysisLogger.getInstance().warn(
							"An exception occured reading type states: " + readException.getClass().getName()
									+ ": " + readException.getMessage());
				}
			}
		}
	}

	public void addClassFileDirectory(ClassPathClassFileSource classFolder) {
		classes.add(classFolder);
	}

}
