package de.unisb.prog.typestate;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;

import org.apache.log4j.Appender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.softevo.util.asm.ASMUtilities;
import org.softevo.util.asm.MethodIdentifier;
import org.softevo.util.source.ClassPathClassFileSource;

import de.unisb.prog.typestate.IAnalysisConfiguration.TopStateMode;
import de.unisb.prog.typestate.IAnalysisConfiguration.UnknownTransitionMode;
import de.unisb.prog.typestate.util.ClassPathModificator;
import de.unisb.prog.typestate.util.Types;
import de.unisb.prog.typestate.violation.IViolationInfo;

public class Main {

	private static HashSet<String>	relevantClasses	= new HashSet<String>();

	/**
	 * @param args
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	public static void main(String[] args) {
		long initTime, runTime;

		initTime = System.nanoTime();
		InputDataSet inputDataSet = null;

		// Default logger
		AnalysisLogger.setLogger(null);
		Logger log = AnalysisLogger.getInstance().getLogger();
		for (Enumeration<Appender> apps = log.getAllAppenders(); apps.hasMoreElements();) {
			apps.nextElement().setLayout(new PatternLayout("%-5p %d{DATE}: %m%n"));
		}
		log.setLevel(Level.INFO);

		try {
			inputDataSet = init(args);
		} catch (IllegalArgumentException e) {
			log.error("Wrong arguments specified: " + e.getMessage());
			printUsage();
			return;
		} catch (ArrayIndexOutOfBoundsException e) {
			log.error("Incorrect number of arguments specified.");
			printUsage();
			return;
		}

		if (inputDataSet == null) {
			log.error("Input data set could not be built. Program terminating.");
			return;
		}

		AnalysisRunner runner = new AnalysisRunner();
		initTime = System.nanoTime() - initTime;
		runTime = System.nanoTime();
		IAnalysisResults results = runner.run(inputDataSet, log);
		runTime = System.nanoTime() - runTime;
		if (results.getNumberOfViolations() > 0) {
			System.out.println(" --- Found violations ---\n");
			int vioNr = 0;
			for (String cls : relevantClasses) {
				List<IViolationPoint> vios = results.getViolationList(cls);
				if (vios != null) {
					for (IViolationPoint vp : vios) {
						System.out.println(String.format("%3d: [%s]%s@%d, call stack %s:", vioNr++, vp
								.getSeverity(), vp.getClassName(), vp.getLineNumber(), vp.getLocation()
								.toString()));
						for (IViolationInfo vi : vp.getInfo()) {
							System.out.println("     " + vi.getType() + ":"
									+ vi.text().replace("\n", "\n     ") + "\n");
						}
					}
				}
			}
		} else {
			log.info(" -- No violations found --\n");
		}

		System.out.println("Finished. Timing statistics in secs:\nINIT: "
				+ String.format("%.3f", initTime / 1000000000.) + "\nANALYSIS: "
				+ String.format("%.3f", runTime / 1000000000.));
	}

	private static InputDataSet init(String[] args) {
		InputDataSet ids = new InputDataSet();
		LinkedList<String> specFolders = new LinkedList<String>();
		LinkedList<String> classFolders = new LinkedList<String>();
		LinkedList<String> classesToAnalyze = new LinkedList<String>();
		LinkedList<String> methodsToAnalyze = new LinkedList<String>();
		// Default, not-so conservative config
		AnalysisConfiguration cfg = new AnalysisConfiguration(TopStateMode.FIRST_ALWAYS_OK,
				UnknownTransitionMode.ERROR_TRANSITION, 1, false, false);
		ids.setConfig(cfg);
		Logger log = AnalysisLogger.getInstance().getLogger();

		/* CMDLINE PARSING */
		for (int arg = 0, n = args.length; arg < n; arg++) {

			if ("--specfiles".equals(args[arg])) {
				specFolders.add(args[++arg]);
			} else if ("--classfiles".equals(args[arg])) {
				classFolders.add(args[++arg]);
			} else if ("--help".equals(args[arg])) {
				printUsage();
				System.exit(0);
			} else if ("--testclass".equals(args[arg])) {
				classesToAnalyze.add(args[++arg]);
			} else if ("--testmethod".equals(args[arg])) {
				methodsToAnalyze.add(args[++arg]);
			} else if ("--showNoSpecWarnings".equals(args[arg])) {
				cfg.setLogUnknownSpec(true);
				log.info("Enabling output of missing specification warnings");
			} else if ("--topPolicy".equals(args[arg])) {
				String policy = args[++arg];
				if ("everyState".equals(policy)) {
					cfg.setTopStateMode(TopStateMode.ASSUME_EVERY_STATE);
				} else if ("firstOk".equals(policy)) {
					cfg.setTopStateMode(TopStateMode.FIRST_ALWAYS_OK);
				} else {
					log.warn("unrecognized TOP policy: " + policy);
					continue;
				}
				log.info("Set --topPolicy to " + policy);
			} else if ("--maxDepth".equals(args[arg])) {
				int depth = 1;
				try {
					depth = Integer.valueOf(args[++arg]);
				} catch (NumberFormatException e) {
					log.warn("illegal max depth specified: " + args[arg]);
					continue;
				}
				if (depth < 0) {
					log.warn("illegal max depth specified: " + args[arg]);
					continue;
				}
				cfg.setMaxDepth(depth);
				log.info("Set --maxDepth to " + depth);
			} else if ("--unknownTransition".equals(args[arg])) {
				String policy = args[++arg];
				if ("ignore".equals(policy)) {
					cfg.setUnknownTransitionMode(UnknownTransitionMode.SAME_STATE);
				} else if ("error".equals(policy)) {
					cfg.setUnknownTransitionMode(UnknownTransitionMode.ERROR_TRANSITION);
				}  else if ("report".equals(policy)) {
					cfg.setUnknownTransitionMode(UnknownTransitionMode.REPORT_UNSPEC);
				} else {
					log.warn("unrecognized transition policy: " + policy);
					continue;
				}
				log.info("Set --unknownTransition to " + policy);
			} else if ("--oneWorklist".equals(args[arg])) {
				cfg.setWalkerUseOnlyOneWorklist(true);
				log.info("Will only use one worklist in analysis algorithm");
			} else if ("--loglevel".equals(args[arg])) {
				String levelStr = args[++arg];
				Level level = null;
				if ("ALL".equals(levelStr)) {
					level = Level.ALL;
				} else if ("TRACE".equals(levelStr)) {
					level = Level.TRACE;
				} else if ("DEBUG".equals(levelStr)) {
					level = Level.DEBUG;
				} else if ("INFO".equals(levelStr)) {
					level = Level.INFO;
				} else if ("WARN".equals(levelStr)) {
					level = Level.WARN;
				} else if ("ERROR".equals(levelStr)) {
					level = Level.ERROR;
				} else if ("OFF".equals(levelStr)) {
					level = Level.OFF;
				} else {
					log.warn("unrecognized logging level: " + levelStr);
					continue;
				}
				log.setLevel(level);
				log.info("Set --loglevel to " + level);
			} else {
				log.warn("unrecognized parameter: " + args[arg]);
				continue;
			}
		}

		if (specFolders.isEmpty()) {
			throw new IllegalArgumentException("parameter --specfiles missing");
		}
		if (classFolders.isEmpty()) {
			throw new IllegalArgumentException("parameter --classfiles missing");
		}

		/* INITIALIZATION */
		log.info("Initializing input data structures");

		// Specs
		try {
			for (String specFolder : specFolders) {
				log.info("Reading typestate specification from " + specFolder);
				ids.readTypeStateAutomata(specFolder);
			}
		} catch (FileNotFoundException e) {
			log.error("Error while reading specification:\n\t" + e.getMessage());
			return null;
		}

		// Class folders
		try {
			for (String folderName : classFolders) {
				File resource = new File(folderName);
				if (!resource.exists()) {
					log.warn("Specified classpath container does not exist: " + folderName);
					continue;
				}
				addResourceToClasspath(ids, log, folderName);

				if (!isJarFile(folderName)) {
					String[] jarFilesInFolder = resource.list(new FilenameFilter() {

						public boolean accept(File dir, String name) {
							return isJarFile(name);
						}
					});

					for (String jarFileName : jarFilesInFolder) {
						addResourceToClasspath(ids, log, folderName + "/" + jarFileName);
					}
				}
			}
		} catch (IOException e) {
			log.error("Could not open class files directory:\n\t" + e.getMessage());
			return null;
		}

		// Classes to analyze
		classesLoop: for (String className : classesToAnalyze) {

			for (String folderName : classFolders) {
				ClassPathClassFileSource classFolder;
				try {
					classFolder = new ClassPathClassFileSource(folderName);
					if (classFolder.containsClass(className)) {
						Set<MethodIdentifier> idents = ASMUtilities.getAllMethods(classFolder
								.getClassForName(className).getClassBytes());
						for (MethodIdentifier ident : idents) {
							ids.addMethod(new de.unisb.prog.typestate.MethodIdentifier(ident.getMethodName(),
									ident.getSignature(), ident.getClassName()));
						}
						relevantClasses.add(className);
						continue classesLoop;

					}
				} catch (NoSuchElementException e) {
				} catch (IOException e) {
					log.warn("Error while searching for class " + className + " in " + folderName
							+ ", skipping because: " + e);
				}
			}
			// Class not found
			log.warn("Requested class " + className + " not found in the given class folders. Skipping");
		}

		// Methods to analyze
		for (String methodFullName : methodsToAnalyze) {
			ids.addMethod(new de.unisb.prog.typestate.MethodIdentifier(methodFullName));
			relevantClasses.add(Types.getClassName(methodFullName));
		}
		return ids;
	}

	private static boolean isJarFile(String filename) {
		return filename.length() >= 4 && filename.substring(filename.length() - 4).equalsIgnoreCase(".jar");
	}

	/**
	 * @param ids
	 * @param log
	 * @param resourceName
	 * @throws IOException
	 */
	private static void addResourceToClasspath(InputDataSet ids, Logger log, String resourceName)
			throws IOException {
		ClassPathClassFileSource classFolder = new ClassPathClassFileSource(resourceName);
		log.info("Adding " + resourceName + " to the class path");
		ids.addClassFileDirectory(classFolder);
		// String currentCP = System.getProperty("java.class.path");
		// if (!currentCP.contains(folderName) ){
		// System.setProperty("java.class.path", currentCP + ":" + folderName);
		// }

		ClassPathModificator.appendPath(resourceName);
	}

	public static void printUsage() {
		System.out.println("\nTypestate verifier program usage:");
		System.out.println();
		System.out.println(String.format("\t%-40s%s", "--help", "prints this usage information and quits"));
		System.out.println();
		System.out.println(" Parameters which MUST be specified at least ONCE:");
		System.out.println(String.format("\t%-40s%s", "--classfiles <file/folder>",
				"resource (folder/archive) containing bytecode of all classes"));
		System.out.println(String.format("\t%-40s%s", "--specfiles <folder>",
				"folder containing specification XML files"));
		System.out.println();
		System.out.println(" Parameters which might be specified multiple times:");
		System.out.println(String.format("\t%-40s%s", "--testclass <class name>",
				"qualified class name (package/package/Class) of ONE class of which all"));
		System.out.println(String.format("\t%-40s%s", " ", "methods are to be analyzed"));
		System.out.println(String.format("\t%-40s%s", "--testmethod <method name>",
				"qualified method name (package/package/Class.method(Ljava/lang/Object;)V)"));
		System.out.println(String.format("\t%-40s%s", " ", "giving a method which is to be analyzed"));
		System.out.println();
		System.out.println(" Option switches:");
		System.out.println(String.format("\t%-40s%s", "--showNoSpecWarnings",
				"enables printing warnings about types without specification"));
		System.out.println();
		System.out.println(String.format("\t%-40s%s", "--oneWorklist", "uses only one worklist in the analysis algorithm"));
		System.out.println();
		System.out.println(String.format("\t%-40s%s", "--maxDepth <int> (default 1)",
				"maximum level to which the call analysis descends"));
		System.out.println();
		System.out.println(String.format("\t%-40s%s", "--topPolicy <policy>",
				"sets how calls on TOP state types are handled."));
		System.out.println(String.format("\t    %-36s%s", "<policy> == everyState",
				"assume object being in any possible state (except 'start' and 'ex')"));
		System.out.println(String.format("\t    %-36s%s", "<policy> == firstOk (default)",
				"assume that the first call on the object never produces an error"));
		System.out.println();
		System.out.println(String.format("\t%-40s%s", "--unknownTransition <policy>",
				"sets how a transition, which is unknown to the object's"));
		System.out.println(String.format("\t%-40s%s", " ",
				"specification, is handled (only if there is a spec. for this type)"));
		System.out.println(String.format("\t    %-36s%s", "<policy> == ignore",
				"the object does not change its state"));
		System.out.println(String.format("\t    %-36s%s", "<policy> == error (default)",
		"the object performs a transition to the error state"));
		System.out.println(String.format("\t    %-36s%s", "<policy> == report",
		"generate a distinct violation for unspecified transitions"));
		System.out.println();
		System.out.println(String.format("\t%-40s%s", "--loglevel <level>", "sets the logging level"));
		System.out.println(String.format("\t    %-36s%s", "<level> == ALL, TRACE, DEBUG,",
				"possible levels sorted by decreasing verbosity"));
		System.out.println(String.format("\t    %-36s", "           INFO (default),"));
		System.out.println(String.format("\t    %-36s", "           WARN, ERROR, OFF"));
		
	}

}
