package de.unisb.prog.typestate;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.unisb.prog.typestate.IAnalysisConfiguration.TopStateMode;

public class Specification {

	private String						typeName;
	private Map<String, IState>			states;
	private Map<String, Set<IState>>	transitionTargets;
	private IState						startState, badState;
	private TopStateMode				topMode;

	public Specification() {

	}

	public void init(String typeName, Map<String, IState> states, String startState, String badState,
			TopStateMode topMode) {
		this.typeName = typeName;
		this.states = Collections.unmodifiableMap(states);
		this.transitionTargets = new HashMap<String, Set<IState>>();
		this.startState = states.get(startState);
		this.badState = states.get(badState);
		this.topMode = topMode;
	}

	/**
	 * @return the name of the type for this specification
	 */
	public String getTypeName() {
		return typeName;
	}

	/**
	 * Gets the start state of the specification automaton.
	 * 
	 * @return The starting state of the specification in internal format {@link IState}
	 */
	public IState getStartingState() {
		return startState;
	}

	/**
	 * Gets the bad state of the specification automaton.
	 * 
	 * @return The bad {@link IState}
	 */
	public IState getBadState() {
		return badState;
	}

	/**
	 * @return a mapping from State name to IState
	 */
	public Map<String, IState> getStates() {
		return states;
	}

	/**
	 * Gets all states which are a possible outcome of performing <code>transition</code> on an instance of
	 * this type, taking the {@link TopStateMode} into account.
	 * 
	 * @param transition
	 *            the transition whose resulting states should be retrieved
	 * @return a Set containing all possible result states
	 */
	public Set<IState> getAllStatesAfter(String transition) {
		Set<IState> targets = transitionTargets.get(transition);
		if (targets == null) {
			// first occurrence
			targets = new HashSet<IState>();
			// enumerate all states
			for (IState state : states.values()) {
				// Starting state can be excluded because the constructor is already called
				if (state.equals(startState))
					continue;

				/*  ************* IF UNDEFINED TRANSITIONS SHOULD ALWAYS BE IGNORED, USE THIS ********* */
				// Set<IState> stateTargets = state.getTransitions().get(transition);
				// //if null, the transition is not defined from this spec. instead of taking the default
				// transition, just ignore it.
				// if (stateTargets!=null)
				// targets.addAll(stateTargets);
				/* *********************************************************************************** */

				/*  ********* IF UNDEFINED TRANSITIONS SHOULD ALWAYS BE CONSIDERED, USE THIS ********** */
				targets.addAll(state.performTransition(transition));
				/* *********************************************************************************** */

			}

			// if FIRST_ALWAYS_OK, remove bad state if any (and if not a must violation)
			switch (topMode) {
			case FIRST_ALWAYS_OK:
				// only remove bad state, if it is not a must violation
				if (!(targets.size() == 1 && targets.contains(badState))) {
					targets.remove(badState);
				}
				break;
			case ASSUME_EVERY_STATE:
				// do nothing
				break;
			}

			// Cache them in the spec
			transitionTargets.put(transition, targets);
		}
		return Collections.unmodifiableSet(targets);
	}

	public String toString() {
		return "Spec (" + getTypeName() + ")";
	}

	public void addStatesAfterTransition(Set<IState> retState) {
		// TODO Auto-generated method stub

	}
}
