package de.unisb.prog.typestate;

import java.util.HashMap;

import de.unisb.prog.jfirm.ir.Type;
import de.unisb.prog.typestate.IAnalysisConfiguration.TopStateMode;
import de.unisb.prog.typestate.IAnalysisConfiguration.UnknownTransitionMode;

/**
 * Takes an {@link IInputDataSet} and provides the specification as {@link IState}s. Specs are converted from
 * JUNG to {@link IState} on demand, and the results are cached once a specification has been converted once.
 * 
 * @author mikerofone
 * 
 */
public class SpecificationRepository implements ISpecificationRepository {

	private IInputDataSet					inputData;
	private HashMap<String, Specification>	specCache;
	private AnalysisLogger					log;
	private UnknownTransitionMode			defaultMode;
	private TopStateMode					topMode;

	public SpecificationRepository(IInputDataSet inputData) {
		this.inputData = inputData;
		this.defaultMode = inputData.getConfig().getUnknownTransitionMode();
		this.topMode = inputData.getConfig().getTopStateMode();
		this.specCache = new HashMap<String, Specification>();
		log = AnalysisLogger.getInstance();
	}

	/**
	 * Looks up the specification in the classCache, then in the {@link IInputDataSet} if not found. If not
	 * cached, the spec is converted and put into the classCache for further accesses.
	 * 
	 * @param typeName
	 *            the jFirm object for the desired type
	 * @return the corresponding {@link Specification}
	 */
	public Specification getSpecification(Type type) {

		// "clean" the typename if it has bytecode format
		String typeName = trim(type.toString());
		if (log.trace())
			log.trace("Specification request for type " + typeName);

		Specification spec = null;
		if (!specCache.containsKey(typeName)) {
			// must convert and store first
			ITypeStateAutomaton jungSpec = inputData.getSpecification(typeName);
			if (jungSpec == null) {
				if (log.debug())
					log.debug(" SpecificationRepository: " + typeName + " is unknown!");
			} else {
				// Convert specification
				log.trace(" Converting spec");
				spec = FSAParser.parseGraph(jungSpec, defaultMode, topMode);
			}
			// put into specCache, if null or not -> if it's not known now, it will never be
			specCache.put(typeName, spec);
		} else {
			spec = specCache.get(typeName);
		}

		// return the specification
		return spec;
	}

	/**
	 * Removes leading/trailing extra characters.
	 * 
	 * @param typeString
	 * @return
	 */
	private String trim(String typeString) {
		// "clean" the typename if it has format "L<type>;"
		if (typeString.charAt(0) == 'L' && typeString.charAt(typeString.length() - 1) == ';') {
			// cut off first and last char
			typeString = typeString.substring(1, typeString.length() - 1);
		}
		return typeString;
	}
}
