package de.unisb.prog.typestate;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * A class representing a state in a finite automaton. It is connected to other states via named transitions.<br>
 * It supports one "catchall" transition which is the unnamed transition "". This will be taken if no existing
 * transition matches the required one (allows for positive specification like "except for transitions A B and
 * C, every transition leads to an error state.")<br>
 * If no default transition has been added, the State won't change if the requested transition is unknown.
 * 
 * @author mikerofone
 * 
 */
public class State implements Cloneable, IState {
	private boolean							bad_;
	private String							name_;			// must be unique, serves as id
	private Specification					spec_;

	/**
	 * Is null for bad states as they, once reached, cannot be left
	 */
	private HashMap<String, Set<IState>>	links_;

	/**
	 * A set only containing the default transition. Initially, the default transition is a loop returning
	 * this state again.
	 */
	private Set<IState>						defaultTrans;

	/**
	 * Construct a standard named non-bad State.
	 * 
	 * @param name
	 *            the name of the state
	 * @param spec
	 *            the specification for the State's type
	 */
	public State(String name, Specification spec) {
		bad_ = false;
		name_ = name;
		links_ = new HashMap<String, Set<IState>>();
		defaultTrans = new HashSet<IState>();
		defaultTrans.add(this);
		spec_ = spec;
	}

	/**
	 * Construct a named state which is initially bad or not.
	 * 
	 * @param name
	 *            the name of the state
	 * @param bad
	 *            determines if the state is bad or not
	 * @param spec
	 *            the specification for the State's type
	 */
	public State(String name, boolean bad, Specification spec) {
		this(name, spec);
		bad_ = bad;
	}

	/**
	 * Indicates whether the state is bad or not
	 * 
	 * @return true if it's bad.
	 */
	final public boolean isBad() {
		return bad_;
	}

	/**
	 * Get the name of the state (for debugging purposes)
	 * 
	 * @return The name of the state
	 */
	final public String getName() {
		return name_;
	}

	/**
	 * @return the {@link Specification} to which this State belongs.
	 */
	final public Specification getTypeSpecification() {
		return spec_;
	}
	
	/**
	 * Perform a transition labeled 'action'.<br>
	 * 
	 * @param action
	 *            The name of the transition to perform
	 * @return A set containing the states reached by following the specified transition, if it's known to the
	 *         State. If it's not, then the default transition is taken (which stays in the same state unless
	 *         specified differently)
	 */
	final public Set<IState> performTransition(String action) {
		// links_ is always != null
		// Get linked states
		Set<IState> states = links_.get(action);
		// Transition unknown, use default transition
		if (states == null)
			states = defaultTrans;
		return states;
	}

	/**
	 * Specifies a default transition taken whenever the automaton should fire a transition which it does not
	 * know. The default transition can only define <strong>one</strong> target state.
	 * 
	 * @param toState
	 *            The resulting state from taking the default transition
	 */
	public void setDefaultTransition(IState toState) {
		// Default transition can only contain one target
		defaultTrans.clear();
		defaultTrans.add(toState);
	}

	/**
	 * Add a new transition.
	 * 
	 * @param onAction
	 *            The name of the action firing the transition.
	 * @param toState
	 *            The state which is reached by firing the transition.
	 */
	public void extend(String onAction, IState toState) {
		// Fetch already existing transitions, if any
		Set<IState> trans = links_.get(onAction);
		if (trans != null) {
			// Add to the existing transitions
			trans.add(toState);
		} else {
			// New transition
			trans = new HashSet<IState>();
			trans.add(toState);
			links_.put(onAction, trans);
		}
	}

	/**
	 * Just override hashCode such that collections are happy.
	 */
	@Override
	final public int hashCode() {
		return name_.hashCode();
	}

	/**
	 * Equals method which only depends on the state's name (debugging)
	 */
	@Override
	final public boolean equals(Object o) {
		return (o instanceof IState && name_.equals(((IState) o).getName()));
	}

	/**
	 * Creates a clone of the object, copying the links_ HashMap but NOT performing a deep copy.
	 */
	@Override
	protected Object clone() {
		State clone = new State(name_, bad_, spec_);
		clone.links_ = links_ != null ? new HashMap<String, Set<IState>>(links_) : null;
		return clone;
	}

	/**
	 * Just return the type and name of the state
	 */
	public String toString() {
		return name_;
	}

	/**
	 * Changes the "bad" marking
	 * 
	 * @param bad
	 *            if true, the state is marked as bad; if false, the mark is removed.
	 */
	public void setBad(boolean bad) {
		this.bad_ = bad;

	}

	public Map<String, Set<IState>> getTransitions() {
		Map<String, Set<IState>> trans = new HashMap<String, Set<IState>>(links_);
		return trans;
	}

}