package de.unisb.prog.typestate;

import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.unisb.prog.typestate.IAnalysisConfiguration.TopStateMode;
import de.unisb.prog.typestate.IAnalysisConfiguration.UnknownTransitionMode;
import edu.uci.ics.jung.graph.DirectedGraph;
import edu.uci.ics.jung.graph.DirectedSparseGraph;

public class TestInputDataSet implements IInputDataSet {

	private List<IMethodIdentifier> methods;

	private HashMap<String, ITypeStateAutomaton> specs;
	
	private ICacheProvider cache;
	
	private HashMap<String, InputStream> classes;

	private IAnalysisConfiguration	config;
	
	public TestInputDataSet() {
		methods = new ArrayList<IMethodIdentifier>();
		specs = new HashMap<String, ITypeStateAutomaton>();
		cache = null;
		classes = new HashMap<String, InputStream>();
		config = new AnalysisConfiguration();
	}
	
	public List<IMethodIdentifier> getAllMethods() {
		return Collections.unmodifiableList(methods);
	}

	public List<String> getAllSpecNames() {
		return new ArrayList<String>(specs.keySet());
	}

	public InputStream getBytecodeStream(String className) {
		InputStream stream = classes.get(className);
		
		if (stream == null) {
			//not found
			return null;
		}
		
		//Create a copy of the stream
		try {
			int numbytes = stream.available();
			byte[] content = new byte[numbytes];
			
			stream.read(content);
			numbytes = stream.available();
			
			//The stream might have more available
			int index = 0;
			while (numbytes != 0) {
				//reinit content
				index = content.length;
				content = copyOf(content, content.length + numbytes);
				stream.read(content, index, numbytes);
				numbytes = stream.available();
			}
			
			//Put new Stream into the map
			classes.put(className, new ByteArrayInputStream(content));
			
			//Return another copy to the caller
			return new ByteArrayInputStream(content);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		//Some error
		return null;
		
	}

  public static byte[] copyOf(byte[] original, int newLength) {
    byte[] copy = new byte[newLength];
    System.arraycopy(original, 0, copy, 0,
                     Math.min(original.length, newLength));
    return copy;
  }

	public ICacheProvider getCacheProvider() {
		return cache;
	}

	public ITypeStateAutomaton getSpecification(String type) {
		return specs.get(type);
	}
	
	/**
	 * Adds a class to the classes list
	 * @param className The class' Java bytecode name
	 * @param byteCode An InputStream providing the class' bytecode
	 * @return true if class was added sucessfully, false if not (was already in, stream was null) 
	 */
	public boolean addClass(String className, InputStream byteCode) {
		if (className == null || byteCode == null || classes.containsKey(className)) {
			//class already in
			return false;
		}
		classes.put(className,byteCode);
		return true;
	}
	
	/**
	 * Adds a method to the methods list. Before a method can be added, the class containing the method must have been added by {@link addClass}.
	 * @param method The method to add
	 * @return true, if adding was sucessful, false if not (class not already added or method already in the set)
	 */
	public boolean addMethod(IMethodIdentifier method) {
		if (method == null || classes.get(method.getClassName()) == null || methods.contains(method)) {
			//no class bytecode found or already added
			return false;
		}
		return methods.add(method);
	}
	
	
	/**
	 * Adds a specification {@link ITypeStateAutomaton} to this data set.
	 * @param className The name of the class for which this automaton is a specification
	 * @param fsa The finite state automaton
	 * @return true, if adding was successful, false else (argument was null, specification was already added)
	 */
	public boolean addSpec(ITypeStateAutomaton fsa) {
		if ( fsa == null || specs.containsKey(fsa.getName())) {
			//class already in
			return false;
		}
		specs.put(fsa.getName(),fsa);
		return true;
	}
	
	/**
	 * Sets the cache provider.
	 * @param cache the cache provider to set
	 */
	public void setCache(ICacheProvider cache) {
		this.cache = cache;
	}

	public IAnalysisConfiguration getConfig() {
		return this.config;
	}
	

	public void setConfig(IAnalysisConfiguration config) {
		this.config = config;
	}
	
	/**
	 * Generates some example TestInputDataSet
	 * 
	 * @param inputData
	 *            the example TestInputDataSet
	 */
	@SuppressWarnings("unused")
	public static TestInputDataSet generateExampleInput() {
		int i = 0;
		TestInputDataSet inputData = new TestInputDataSet();
		try {
			inputData.addClass("de/unisb/prog/typestate/TestClass", new FileInputStream(
					"target/classes/de/unisb/prog/typestate/TestClass.class"));
			// inputData.addClass("de/unisb/prog/typestate/ir/oldstuff/TalkClass", new FileInputStream(
			// "target/classes/de/unisb/prog/typestate/ir/oldstuff/TalkClass.class"));

			// **
			// METHODS
			// **
			IMethodIdentifier failingFileTestMethod = new MethodIdentifier("failingFileTest",
					"(Lde/unisb/prog/typestate/dummyImpl/File;Lde/unisb/prog/typestate/dummyImpl/File;)V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier validFileTestMethod = new MethodIdentifier("validFileTest",
					"(Lde/unisb/prog/typestate/dummyImpl/File;Lde/unisb/prog/typestate/dummyImpl/File;)V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier bigTestMethod = new MethodIdentifier("bigTest", "()V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier hugeTestMethod = new MethodIdentifier("hugeTest",
					"(Lde/unisb/prog/typestate/dummyImpl/File;Lde/unisb/prog/typestate/dummyImpl/File;)V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier stackTestMethod = new MethodIdentifier("stackTest", "()V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier mergeTestMethod = new MethodIdentifier("mergeTest", "()V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier phiLoopTestMethod = new MethodIdentifier("phiLoopTest", "()V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier fuzzyTestMethod = new MethodIdentifier("fuzzyTest", "()V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier fieldTestMethod = new MethodIdentifier("fieldTest",
			// "(Lde/unisb/prog/typestate/dummyImpl/File;Lde/unisb/prog/typestate/dummyImpl/File;)Lde/unisb/prog/typestate/dummyImpl/File;",
					"()Lde/unisb/prog/typestate/dummyImpl/File;", "de/unisb/prog/typestate/TestClass");
			MethodIdentifier defaultConstructor = new MethodIdentifier("<init>", "()V",
					"de/unisb/prog/typestate/TestClass");
			IMethodIdentifier recursionTestMethod = new MethodIdentifier("t_recursionTest",
					"(Lde/unisb/prog/typestate/dummyImpl/File;)Lde/unisb/prog/typestate/dummyImpl/File;",
					"de/unisb/prog/typestate/TestClass");
//			inputData.addMethod(failingFileTestMethod);
//			inputData.addMethod(stackTestMethod);
//			inputData.addMethod(validFileTestMethod);
//			inputData.addMethod(mergeTestMethod);
			inputData.addMethod(bigTestMethod);
//			inputData.addMethod(hugeTestMethod);
//			inputData.addMethod(phiLoopTestMethod);
//			inputData.addMethod(fieldTestMethod);
//			inputData.addMethod(defaultConstructor);
//			inputData.addMethod(recursionTestMethod);
//			inputData.addMethod(fuzzyTestMethod);

			// // TALK CLASS
			// IMethodIdentifier talkSimpleReadMethod = new MethodIdentifier("readSimple",
			// "(Ljava/io/File;)B","de/unisb/prog/typestate/ir/oldstuff/TalkClass");
			// IMethodIdentifier talkUglyReadMethod = new MethodIdentifier("readUgly",
			// "(Ljava/io/File;)B","de/unisb/prog/typestate/ir/oldstuff/TalkClass");
			//			
			// System.out.println(inputData.addMethod(talkSimpleReadMethod));
			// System.out.println(inputData.addMethod(talkUglyReadMethod));

			// **
			// DUMMY IMPLEMENTATIONS - "THE FIRM WORLD"
			// **
			inputData.addClass("de/unisb/prog/typestate/dummyImpl/File", new FileInputStream(
					"target/classes/de/unisb/prog/typestate/dummyImpl/File.class"));

			inputData.addClass("de/unisb/prog/typestate/dummyImpl/Stack", new FileInputStream(
					"target/classes/de/unisb/prog/typestate/dummyImpl/Stack.class"));

		} catch (FileNotFoundException e) {
			AnalysisLogger.getInstance().debug(
					"generateSampleInput(): Could not open class files, only composing sample specs");
			// Only genereate sample specs
		}

		// **
		// SPEC FSA
		// **

		// ** FOR FILE **
		DirectedGraph<String, ITransition> fileGraph = new DirectedSparseGraph<String, ITransition>();
		// VERTICES
		fileGraph.addVertex("start");
		fileGraph.addVertex("!isOpen");
		fileGraph.addVertex("isOpen");
		fileGraph.addVertex("ex");

		// METHOD DESCRIPTORS
		IMethodIdentifier constructor = new MethodIdentifier("<init>", "()V",
				"de/unisb/prog/typestate/dummyImpl/File");
		IMethodIdentifier openMethod = new MethodIdentifier("open", "()V",
				"de/unisb/prog/typestate/dummyImpl/File");
		IMethodIdentifier closeMethod = new MethodIdentifier("close", "()V",
				"de/unisb/prog/typestate/dummyImpl/File");
		IMethodIdentifier readMethod = new MethodIdentifier("read",
				"()Lde/unisb/prog/typestate/dummyImpl/Obj;", "de/unisb/prog/typestate/dummyImpl/File");
		IMethodIdentifier writeMethod = new MethodIdentifier("write",
				"(Lde/unisb/prog/typestate/dummyImpl/Stack;)V", "de/unisb/prog/typestate/dummyImpl/File");
		IMethodIdentifier isOpenMethod = new MethodIdentifier("isOpen", "()Z",
				"de/unisb/prog/typestate/dummyImpl/File");
		// TRANSITIONS
		Set<IMethodIdentifier> set = new HashSet<IMethodIdentifier>();
		set.add(constructor);
		fileGraph.addEdge(new Transition("start", "!isOpen", set), "start", "!isOpen");

		set = new HashSet<IMethodIdentifier>();
		set.add(openMethod);
		fileGraph.addEdge(new Transition("!isOpen", "isOpen", set), "!isOpen", "isOpen");

		set = new HashSet<IMethodIdentifier>();
		set.add(readMethod);
		set.add(writeMethod);
		set.add(isOpenMethod);
		fileGraph.addEdge(new Transition("isOpen", "isOpen", set), "isOpen", "isOpen");

		set = new HashSet<IMethodIdentifier>();
		set.add(closeMethod);
		fileGraph.addEdge(new Transition("isOpen", "!isOpen", set), "isOpen", "!isOpen");

		set = new HashSet<IMethodIdentifier>();
		set.add(isOpenMethod);
		fileGraph.addEdge(new Transition("!isOpen", "!isOpen", set), "!isOpen", "!isOpen");

		// ex transitions
		set = new HashSet<IMethodIdentifier>();
		set.add(writeMethod);
		set.add(readMethod);
		set.add(closeMethod);
		fileGraph.addEdge(new Transition("!isOpen", "ex", set), "!isOpen", "ex");

		set = new HashSet<IMethodIdentifier>();
		set.add(writeMethod); // file full for example
		set.add(openMethod);
		fileGraph.addEdge(new Transition("isOpen", "ex", set), "isOpen", "ex");

		set = new HashSet<IMethodIdentifier>();
		set.add(writeMethod);
		set.add(readMethod);
		set.add(closeMethod);
		set.add(isOpenMethod);
		fileGraph.addEdge(new Transition("start", "ex", set), "start", "ex");
		
		// DEFAULT transitions
		// Marked with an empty set. As it will stay empty, the same set can be used multiple times.
		// set = new HashSet<IMethodIdentifier>();
		// fileGraph.addEdge(new Transition("isOpen", "ex", set), "isOpen", "ex");
		// fileGraph.addEdge(new Transition("!isOpen", "ex", set), "!isOpen", "ex");

		ITypeStateAutomaton fileSpec = new TypeStateAutomaton("de/unisb/prog/typestate/dummyImpl/File",
				fileGraph, "start", "ex");
		inputData.addSpec(fileSpec);

		// ** FOR STACK **
		DirectedGraph<String, ITransition> stackGraph = new DirectedSparseGraph<String, ITransition>();
		// VERTICES
		stackGraph.addVertex("start");
		stackGraph.addVertex("isEmpty");
		stackGraph.addVertex("!isEmpty");
		stackGraph.addVertex("ex");
		// METHOD DESCRIPTORS
		constructor = new MethodIdentifier("<init>", "()V", "de/unisb/prog/typestate/dummyImpl/Stack");
		IMethodIdentifier pushMethod = new MethodIdentifier("push",
				"(Lde/unisb/prog/typestate/dummyImpl/File;)V", "de/unisb/prog/typestate/dummyImpl/Stack");
		IMethodIdentifier popMethod = new MethodIdentifier("pop",
				"()Lde/unisb/prog/typestate/dummyImpl/File;", "de/unisb/prog/typestate/dummyImpl/Stack");
		IMethodIdentifier clearMethod = new MethodIdentifier("clear", "()V",
		"de/unisb/prog/typestate/dummyImpl/Stack");
		IMethodIdentifier isEmptyMethod = new MethodIdentifier("isEmpty", "()Z",
				"de/unisb/prog/typestate/dummyImpl/Stack");
		// TRANSITIONS
		set = new HashSet<IMethodIdentifier>();
		set.add(constructor);
		stackGraph.addEdge(new Transition("start", "isEmpty", set), "start", "isEmpty");
		
		set = new HashSet<IMethodIdentifier>();
		set.add(pushMethod);
		set.add(clearMethod);
		stackGraph.addEdge(new Transition("isEmpty", "!isEmpty", set), "isEmpty", "!isEmpty");
		
		set = new HashSet<IMethodIdentifier>();
		set.add(popMethod);
		stackGraph.addEdge(new Transition("!isEmpty", "isEmpty", set), "!isEmpty", "isEmpty");
		
		set = new HashSet<IMethodIdentifier>();
		set.add(isEmptyMethod);
		set.add(clearMethod);
		stackGraph.addEdge(new Transition("isEmpty", "isEmpty", set), "isEmpty", "isEmpty");
		
		set = new HashSet<IMethodIdentifier>();
		set.add(pushMethod);
		set.add(popMethod);
		set.add(isEmptyMethod);
		stackGraph.addEdge(new Transition("!isEmpty", "!isEmpty", set), "!isEmpty", "!isEmpty");
		
		// ex TRANSITIONS
		set = new HashSet<IMethodIdentifier>();
		set.add(pushMethod);
		set.add(popMethod);
		set.add(isEmptyMethod);
		set.add(clearMethod);
		stackGraph.addEdge(new Transition("start", "ex", set), "start", "ex");
		
		set = new HashSet<IMethodIdentifier>();
		set.add(popMethod);
		stackGraph.addEdge(new Transition("isEmpty", "ex", set), "isEmpty", "ex");
		
		// //DEFAULT transitions
		// set = new HashSet<IMethodIdentifier>();
		// stackGraph.addEdge(new Transition("isEmpty", "ex", set), "isEmpty", "ex");
		// stackGraph.addEdge(new Transition("!isEmpty", "ex", set), "!isEmpty", "ex");
		ITypeStateAutomaton stackSpec = new TypeStateAutomaton("de/unisb/prog/typestate/dummyImpl/Stack",
				stackGraph, "start", "ex");
		inputData.addSpec(stackSpec);

		// Configure parameters of analysis.
		inputData.setConfig(new AnalysisConfiguration(TopStateMode.FIRST_ALWAYS_OK,
				UnknownTransitionMode.ERROR_TRANSITION, 4,true,false));

		return inputData;
	}


	
	
}
