/**
 * 
 */
package de.unisb.prog.typestate;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.unisb.prog.typestate.IAnalysisConfiguration.TopStateMode;
import de.unisb.prog.typestate.util.Types;

/**
 * An abstract implementation of the TOP value for State information. It means that the instance is in
 * "unknown" state might be in every possible state which the specification for this object defines, excluding
 * BAD states.<br>
 * The set containing all states is not enumerated and stored, but it is represented by an abstract value
 * which will be interpreted accordingly if a transition is to be performed. Interpretation modes can be
 * switched by using the corresponding {@link TopState.Mode} parameter.
 * 
 * @author mikerofone
 * 
 */
public class TopState implements IState {

	private Specification	spec;

	private TopState(Specification spec) {
		this.spec = spec;
	}

	public static IState getTop(Specification spec, TopStateMode mode) {
		if (spec != null) {
			// specification is known, return appropriate TOP
			return new TopState(spec);
		} else {
			// specification unknown, so the state is unknown as well
			return UnknownState.getInstance();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.unisb.prog.typestate.IState#getName()
	 */
	public String getName() {
		return toString();
	}

	/**
	 * The unknown state does not know its possible transitions.
	 * 
	 * @return an empty mapping
	 * 
	 * @see de.unisb.prog.typestate.IState#getTransitions()
	 */
	public Map<String, Set<IState>> getTransitions() {
		return Collections.emptyMap();
	}

	/**
	 * Unknown state is not considered a bad state.
	 * 
	 * @return <code>false</code>
	 * 
	 * @see de.unisb.prog.typestate.IState#isBad()
	 */
	public boolean isBad() {
		return false;
	}

	/**
	 * The unknown state assumes that the very first transition that is fired is valid and results in all
	 * possible outcome-states of the transition.
	 * 
	 * @param action
	 *            the action to perform
	 * @return a Set containing all {@link IState}s reachable by firing <code>action</code> on this type,
	 *         possibly including the error state.
	 * 
	 * @see de.unisb.prog.typestate.IState#performTransition(java.lang.String)
	 */
	public Set<IState> performTransition(String action) {
		Set<IState> retState = new HashSet<IState>();
		if (spec != null) {
			// the type of the TOP state is known

			// if action is a constructor call, the affected instance must have been uninitialized
			// --> only consider starting state
			if (action.startsWith("<init>")) {
				return spec.getStartingState().performTransition(action);
			} else {
				retState = spec.getAllStatesAfter(action);
			}
		} else {
			// just stay in TOP state
			retState.add(UnknownState.getInstance());
		}
		return retState;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.unisb.prog.typestate.IState#getTypeSpecification()
	 */
	public Specification getTypeSpecification() {
		return spec;
	}

	public String toString() {
		return "<TOP (" + (spec != null ? Types.shortName(spec.getTypeName()) : "?") + ")>";
	}

	@Override
	public int hashCode() {
		return (spec != null ? spec.hashCode() : 0);
	}

	@Override
	public boolean equals(Object o) {
		if (o instanceof TopState) {
			TopState other = (TopState) o;
			return this.spec != null ? this.spec.equals(other.spec) : other.spec == null;
		}
		return false;
	}

}
