package de.unisb.prog.typestate;

import java.util.ArrayList;
import java.util.List;

import de.unisb.prog.analysis.Context;
import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.Call;
import de.unisb.prog.jfirm.ir.DebugInfo;
import de.unisb.prog.jfirm.ir.MethodGraph;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.typestate.util.Types;
import de.unisb.prog.typestate.violation.IViolationInfo;

public class ViolationPoint implements IViolationPoint {

	/** The name of the class containing this violation point. */
	private String					className;

	/** Additional info about the violation */
	private List<IViolationInfo>	info;

	/** The line number in the source code where this violation was found. */
	private int						lineNumber;

	/** The severity of the violation */
	private Severity				severity;

	/** The valid context when the violation was found */

	private MethodGraph				methodGraph;

	private Context					context;

	public ViolationPoint(String className, IViolationInfo singleInfo, ContextAndNode node,
			Severity severity, MethodGraph methodGraph) {
		if (className == null || singleInfo == null || node == null || severity == null
				|| methodGraph == null) {
			throw new IllegalArgumentException(String.format("ViolationPoint(%s,%s,%s,%s,%s,%s)", className,
					singleInfo, node, severity, methodGraph));
		}
		ArrayList<IViolationInfo> infoList = new ArrayList<IViolationInfo>(1);
		infoList.add(singleInfo);
		init(className, infoList, node, severity, methodGraph);
	}

	public ViolationPoint(String className, List<IViolationInfo> infoList, ContextAndNode node,
			Severity severity, MethodGraph methodGraph) {
		if (className == null || infoList == null || node == null || severity == null || methodGraph == null) {
			throw new IllegalArgumentException(String.format("ViolationPoint(%s,%s,%s,%s,%s,%s)", className,
					info, node, severity, methodGraph));
		}
		init(className, infoList, node, severity, methodGraph);
	}

	private void init(String className, List<IViolationInfo> infoList, ContextAndNode node,
			Severity severity, MethodGraph methodGraph) {
		this.className = className;
		this.info = infoList;
		ValueNode vNode = node.getNode();
		this.lineNumber = vNode == null ? -1 : (vNode.getDebugInfo() == null ? -vNode.getId() : vNode
				.getDebugInfo().getLine());
		this.severity = severity;
		this.methodGraph = methodGraph;
		this.context = node.getContext();
	}

	public String getClassName() {
		return className;
	}

	public List<IViolationInfo> getInfo() {
		return info;
	}

	public int getLineNumber() {
		return lineNumber;
	}

	public Severity getSeverity() {
		return severity;
	}

	/**
	 * Two {@link IViolationPoint}s are the same if class name, info, node and severity are the same.
	 */
	@Override
	public boolean equals(Object o) {
		if (o instanceof IViolationPoint) {
			IViolationPoint other = (IViolationPoint) o;
			return className.equals(other.getClassName()) && info.equals(other.getInfo())
					&& lineNumber == other.getLineNumber() && severity == other.getSeverity();
		} else
			return false;
	}

	/**
	 * Compare violation points 1. by class Name and 2. by line number in the source code.
	 * 
	 * @param other
	 *            the other <code>ViolationPoint</code>
	 * @return The result of this.className.compareTo(other.className) if both are not the same, and
	 *         this.lineNumber.compareTo(other.lineNumber) if they are.
	 */
	public int compareTo(IViolationPoint other) {
		int sameclass = this.className.compareTo(other.getClassName());
		return sameclass == 0 ? new Integer(this.lineNumber).compareTo(other.getLineNumber()) : sameclass;
	}

	public List<String> getLocation() {
		List<String> location = new ArrayList<String>();
		String str = methodGraph.getName();
		for (Call call : context.getCallStack()) {
			DebugInfo dbgi = call.getDebugInfo();
			if (dbgi != null) {
				str += "@" + dbgi.getLine();
			}
			location.add(str);

			str = Types.getFullShortName(call.getTargetMethod());
		}
		location.add(str);

		return location;
	}
}
