package de.unisb.prog.typestate.io;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.apache.commons.collections15.Transformer;
import org.softevo.util.asm.MethodIdentifier;

import de.unisb.prog.typestate.IMethodIdentifier;
import de.unisb.prog.typestate.ITransition;
import edu.uci.ics.jung.graph.DirectedGraph;
import edu.uci.ics.jung.graph.DirectedSparseGraph;
import edu.uci.ics.jung.io.GraphIOException;
import edu.uci.ics.jung.io.graphml.EdgeMetadata;
import edu.uci.ics.jung.io.graphml.GraphMLReader2;
import edu.uci.ics.jung.io.graphml.GraphMetadata;
import edu.uci.ics.jung.io.graphml.HyperEdgeMetadata;
import edu.uci.ics.jung.io.graphml.NodeMetadata;

public class TypeStateReader {

	public static Set<IMethodIdentifier> parseMethodIdentifiers(String input) {
		Set<IMethodIdentifier> result = new HashSet<IMethodIdentifier>();
		String[] methodDescriptors = input.split("\\|");
		for (String methodDescriptor : methodDescriptors) {
			MethodIdentifier identifier = MethodIdentifier.parseFromIdentifier(methodDescriptor);
			result.add(new MethodIdentifierWrapper(identifier));
		}
		return result;
	}

	private static class Transition implements ITransition {

		private Set<IMethodIdentifier>	identifiers;

		private String					target;

		private String					source;

		public Transition(String source, String target, Set<IMethodIdentifier> identifiers) {
			this.source = source;
			this.target = target;
			this.identifiers = identifiers;
		}

		public String getDestinationName() {
			return target;
		}

		public Set<IMethodIdentifier> getMarking() {
			return identifiers;
		}

		public String getSourceName() {
			return source;
		}

	}

	public static ITransition getTransition(String methodIdentifiers, String source, String target) {
		Set<IMethodIdentifier> identifiers = parseMethodIdentifiers(methodIdentifiers);
		return new Transition(source, target, identifiers);
	}

	public static Map<String, DirectedGraph<String, ITransition>> readTypeStates(File typeStateFile)
			throws IOException, GraphIOException {
		final Vector<String> classNameWrapper = new Vector<String>(1);
		classNameWrapper.add("null");
		Transformer<GraphMetadata, DirectedGraph<String, ITransition>> graphTransformer = new Transformer<GraphMetadata, DirectedGraph<String, ITransition>>() {
			public DirectedGraph<String, ITransition> transform(GraphMetadata metaData) {
				String className = metaData.getId();
				// This id might be space-separated and contain multiple classes. This will be handled later
				// when specs are stored inside the spec-map.
				classNameWrapper.set(0, className);
				return new DirectedSparseGraph<String, ITransition>();
			}
		};
		Transformer<NodeMetadata, String> vertexTransformer = new Transformer<NodeMetadata, String>() {
			public String transform(NodeMetadata metaData) {
				return metaData.getId();
			}
		};
		Transformer<EdgeMetadata, ITransition> edgeTransformer = new Transformer<EdgeMetadata, ITransition>() {
			public ITransition transform(EdgeMetadata metaData) {
				return getTransition(metaData.getProperty("key1").replace("[init]", "<init>"), metaData
						.getSource(), metaData.getTarget());
			}
		};
		Transformer<HyperEdgeMetadata, ITransition> hyperEdgeTransformer = new Transformer<HyperEdgeMetadata, ITransition>() {
			public ITransition transform(HyperEdgeMetadata metaData) {
				return null;
			}
		};
		GraphMLReader2<DirectedGraph<String, ITransition>, String, ITransition> reader = new GraphMLReader2<DirectedGraph<String, ITransition>, String, ITransition>(
				new FileReader(typeStateFile), graphTransformer, vertexTransformer, edgeTransformer,
				hyperEdgeTransformer);
		reader.init();
		Map<String, DirectedGraph<String, ITransition>> result = new HashMap<String, DirectedGraph<String, ITransition>>();
		DirectedGraph<String, ITransition> graph;
		do {
			try {
				graph = reader.readGraph();
			} catch (GraphIOException exception) {
				if (exception.getMessage().equals(
						"Unable to read Graph from document - the document could be empty")) {
					graph = null;
				} else {
					throw exception;
				}
			}
			if (graph != null) {
				String className = classNameWrapper.get(0);
				// Handle multiple classes
				// Graph id can contain multiple class names, separated by whitespace
				if (className != null) {
					className = className.trim();
					if (!"".equals(className)) {
						// className is a valid string
						String classes[] = className.split(" ");
						for (int i = 0, n = classes.length; i < n; i++)
							if (!"".equals(classes[i]))
								result.put(classes[i], graph);
					}
				}

			}
		} while (graph != null);
		reader.close();
		return result;
	}

}
