package de.unisb.prog.typestate.ir;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.Type;
import de.unisb.prog.typestate.AnalysisLogger;

public class AllocSite extends MutableStateTuple {
	private static final char			ALLOCSITE_LOG_TAG	= 'A';
	private static final AnalysisLogger	log					= AnalysisLogger.getInstance();

	private static int					counter				= 1;
	private int							id;

	private Set<Reference>				users;

	private ContextAndNode				source;

	/** New inited AllocSite from ContextAndNode constructor */
	public AllocSite(ContextAndNode source, Type type) {
		super(type);
		this.source = source;
		users = new HashSet<Reference>();
		this.id = counter++;
	}

	/** Cloning constructor */
	private AllocSite(AllocSite srcAlloc, boolean copyUsers) {
		super(srcAlloc);
		if (copyUsers) {

			this.users = new HashSet<Reference>(srcAlloc.users);
			for (Reference user : users) {
				user.registered(this);
			}
		} else
			this.users = new HashSet<Reference>();
		this.id = counter++;
	}

	/**
	 * Creates a copy of this allocation site, but the user references are translated to other references
	 * according to the given mapping. If no mapping is given (<code>null</code>), the returned allocation
	 * site has no users
	 */
	public AllocSite copy(HashMap<Reference, Reference> referenceMapping) {
		AllocSite retAlloc = new AllocSite(this, false);

		// map the old users to the new ones (or discard them, if no such mapping provided)
		if (referenceMapping != null) {
			for (Reference oldUser : users) {
				Reference newUser = referenceMapping.get(oldUser);

				// only add the user if it is contained in the map. if it's missing, just drop it
				if (newUser != null) {
					retAlloc.users.add(newUser);
					newUser.registered(retAlloc);
				} else {
					if (log.debug())
						log.debug("AllocSite.copy on " + this + " with mapping " + referenceMapping
								+ ": Dropped user " + oldUser);
				}
			}
		}

		return retAlloc;
	}

	/** Returns the {@link ContextAndNode} this AllocSite stands for */
	public ContextAndNode getSource() {
		return source;
	}

	/** Register new using reference */
	public boolean addUser(Reference usingRef) {
		boolean changed = users.add(usingRef);
		if (changed) {
			// notify user
			usingRef.registered(this);
		} else {
			if (log.debug())
				log.debug("Tried to register ref " + usingRef + " to alloc " + this
					+ " but was already registered");
		}

		return changed;
	}

	/** Remove using reference */
	public boolean removeUser(Reference usingRef) {
		boolean changed = users.remove(usingRef);
		if (changed) {
			// notify user
			usingRef.unregistered(this);
		}
		assert changed : "Tried to unregister ref " + usingRef + " from alloc " + this
				+ " but was not registered";
		return changed;
	}

	public boolean disconnect() {
		int oldSize = users.size();
		for (Iterator<Reference> it = users.iterator(); it.hasNext();) {
			it.next().unregistered(this);
			it.remove();
		}
		return oldSize != 0;
	}

	/** Merge the user lists as well as updating the states */
	public boolean update(AllocSite fromTuple, boolean strongUpdate,
			HashMap<Reference, Reference> referenceMapping) {
		boolean changed = false;

		// -- UPDATING OF USER INFORMATION --
		if (strongUpdate) {
			// Strong update does overwrite users as well, so disconnect before merging other data
			changed = disconnect();
		} else {
			// no strong update, so check if users need to be replaced by copies
			if (referenceMapping != null) {
				List<Reference> newUsers = new LinkedList<Reference>();
				// some of the current reference links might need to be mapped to new ones
				for (Iterator<Reference> it = users.iterator(); it.hasNext();) {
					Reference user = it.next();
					Reference newUser = referenceMapping.get(user);
					// only replace if necessary
					if (newUser != null) {
						// replace old with new
						user.unregistered(this);
						newUser.registered(this);

						it.remove();
						newUsers.add(newUser);
					}
				}
				users.addAll(newUsers);
			}
		}

		// migrate the (remapped) users from the source AllocSite
		if (referenceMapping != null) {
			for (Reference srcUser : fromTuple.users) {
				Reference newUser = referenceMapping.get(srcUser);
				assert newUser != null : ("mapping for new references does not contain reference " + srcUser
						+ " in source alloc " + fromTuple);
				changed = addUser(newUser) || changed;
			}
		} else {
			// not remapped, directly link
			for (Reference srcUser : fromTuple.users) {
				changed = addUser(srcUser) || changed;
			}
		}

		// -- UPDATING TYPESTATE INFORMATION --
		return super.update(fromTuple, strongUpdate) || changed;

	}

	/**
	 * Propagate a change of state to all of this AllocSite's users.
	 * 
	 * @param newState
	 * @param strongUpdateOnAlloc
	 *            whether it was a strong update on the <strong>AllocSite</strong>
	 * @param seen
	 *            set of {@link ContextAndNode}s which have already been updated (and thus can be skipped).
	 *            can be <code>null</code>, then nothing will be skipped
	 * @return whether this changed the state of at least one of the users
	 */
	public boolean propagate(StateTuple newState, boolean strongUpdateOnAlloc, HashSet<Reference> seen) {

		boolean changed = false;
		boolean useSeen = seen != null;
		if (log.trace())
			log.trace("allocsite " + this + " received update", ALLOCSITE_LOG_TAG);
		// propagate to users
		for (Reference user : users) {

			if (useSeen) {
				if (seen.contains(user)) {
					if (log.trace())
						log.trace("  skipping user " + user, ALLOCSITE_LOG_TAG);
					continue;
				}
				seen.add(user);
			}
			if (log.trace())
				log.trace("  propagating to user " + user, ALLOCSITE_LOG_TAG);

			changed = user.receiveUpdate(newState, strongUpdateOnAlloc) || changed;
		}

		return changed;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder(id + "A" + super.toString());
		for (Reference ref : users) {
			sb.append(" " + ref.getId());
		}
		return sb.toString();
	}

	/** Removes all users not existent in the given snapshot */
	public void removeUsersNotIn(MemorySnapshot currentMemSnapshot) {
		HashSet<Reference> usersCopy = new HashSet<Reference>(users);
		for (Reference user : usersCopy) {
			if (!currentMemSnapshot.contains(user)) {
				users.remove(user);
			}

		}

	}

}
