package de.unisb.prog.typestate.ir;

import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.ArrayLoad;
import de.unisb.prog.jfirm.ir.ArrayStore;
import de.unisb.prog.jfirm.ir.DataMemExec;
import de.unisb.prog.jfirm.ir.GetField;
import de.unisb.prog.jfirm.ir.GetStatic;
import de.unisb.prog.jfirm.ir.MemExec;
import de.unisb.prog.jfirm.ir.Mode;
import de.unisb.prog.jfirm.ir.NewArray;
import de.unisb.prog.jfirm.ir.PutField;
import de.unisb.prog.jfirm.ir.PutStatic;
import de.unisb.prog.jfirm.ir.Type;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.jfirm.ir.Type.ClassType.Field;
import de.unisb.prog.pointsto.PointsToSet;
import de.unisb.prog.typestate.IState;
import de.unisb.prog.typestate.Specification;
import de.unisb.prog.typestate.TopState;
import de.unisb.prog.typestate.UnknownState;
import de.unisb.prog.typestate.ir.FirmWalker.Globals;
import de.unisb.prog.typestate.util.NodeHelper;

public class ArrayFieldHandler {

	private Globals	g;

	public void init(Globals g) {
		this.g = g;
	}

	/* COMMON STUFF */
	/*
	 * Inputs to a putField: 0: Mem 1: Object containing affected field 2: Object to store in field
	 * 
	 * Inputs to a getField: 0: Mem 1: Object containing affected field
	 * 
	 * put/getStatic: ref to Object containing affected field missing
	 */

	/* "GET" NODES */
	public ValueNode handleGetStatic(GetStatic getStatic) {

		Field field = getStatic.getTargetField();

		// Only care about reference type fields
		if (field.getType().getMode() == Mode.A) {

			// static field reference, so no handling of objects necessary
			FieldID staticField = new FieldID(field, null);
			Reference staticFieldRef = retrieveFieldRef(staticField);
			FieldReference getStaticRef = new FieldReference(staticFieldRef.getType());
			getStaticRef.update(staticFieldRef, true);
			getStaticRef.addField(staticField);

			// store new reference for the getField PROJ DATA node
			ContextAndNode getStaticNode = new ContextAndNode(g.context, getStatic);
			g.methodMem.putAsReference(getStaticNode, getStaticRef);
			g.methodMem.printDumpCurrMem();
		}
		return getStatic.getValue(DataMemExec.MEM);
	}

	public ValueNode handleGetField(GetField getField) {

		Field field = getField.getTargetField();

		// Only care about reference type fields
		if (field.getType().getMode() == Mode.A) {

			// 1.for each referenced field:
			// 1a. retrieve fieldReference
			// 1b. merge field's status into getFieldRef
			// 2. link getFieldRef to all alloc sites referenced by the getfield PROJ DATA
			FieldReference getFieldRef = new FieldReference(NodeHelper.getType(getField));
			ContextAndNode getFieldObjectNode = new ContextAndNode(g.context, NodeHelper
					.getInput(getField, 1));
			g.log.debug("PRE get ptsto for different field hosting objects");
			PointsToSet pts = g.methodMem.getAllocSites(getFieldObjectNode);
			g.log.debug("POST get ptsto for different field hosting objects " + pts);

			// potentially more than one field affected by the getfield, so read state from all of them
			// and combine then in one new reference for the getField
			g.log.debug("PRE connect fieldRef to fields");
			for (ContextAndNode allocSite : pts) {
				FieldID fieldID = new FieldID(field, allocSite);
				// get reference of the field
				Reference fieldRef = retrieveFieldRef(fieldID);

				// retrieved field might never be null, as no information is equivalent to TOP
				getFieldRef.update(fieldRef, false);
				getFieldRef.addField(fieldID);
			}
			g.log.debug("POST connect fieldRef to fields, state " + getFieldRef.getStateSet());
			
			ContextAndNode getFieldNode = new ContextAndNode(g.context, getField);
			g.log.debug("PRE put getFieldRef as reference");
			g.methodMem.putAsReference(getFieldNode, getFieldRef);
			g.log.debug("POST put getFieldRef as reference");
			g.methodMem.printDumpCurrMem();
		}

		return getField.getValue(DataMemExec.MEM);
	}

	public ValueNode handleLoad(ArrayLoad loadNode) {

		// Only care about reference type fields
		ValueNode value = NodeHelper.getDataNode(loadNode);
		if (value.getMode() == Mode.A) {
			ContextAndNode aLoadNode = new ContextAndNode(g.context, loadNode);

			// Create a (new) Reference Node for this getField which stores the current state of the instance
			// and which is linked to all allocation sites which this getField can reach
			g.methodMem.makeReference(aLoadNode, null);
//			g.methodMem.putAsReference(aLoadNode, new Reference(NodeHelper.getType(value)));
			
		}
		// Those are EnumTupleNode<DataMemExec>
		return loadNode.getValue(DataMemExec.MEM);
	}

	public ValueNode handlePutStatic(PutStatic putStaticNode) {

		Field field = putStaticNode.getTargetField();

		// Only care about reference type fields
		if (field.getType().getMode() == Mode.A) {

			ContextAndNode nodeToPut = new ContextAndNode(g.context, NodeHelper.getInput(putStaticNode, 1));
			StateTuple state = g.methodMem.readState(nodeToPut);
			assert state != null : "The object " + nodeToPut + " is unknown while storing in field " + field;
			// g.methodMem.setGlobal(nodeToPut);

//			//create reference
//			Reference fieldRef = g.methodMem.getLinkedReference(nodeToPut);
//			fieldRef.update(state, true);

			// Static field, so only exactly one field to update
			FieldID fieldID = new FieldID(field, null);

			// store reference in memory
			g.methodMem.setFieldRef(fieldID, state, nodeToPut, true);
			g.methodMem.printDumpCurrMem();
		}
		return putStaticNode.getValue(MemExec.MEM);

	}

	public ValueNode handlePutField(PutField putFieldNode) {

		Field field = putFieldNode.getTargetField();

		// Only care about reference type fields
		if (field.getType().getMode() == Mode.A) {

			ContextAndNode nodeToPut = new ContextAndNode(g.context, NodeHelper.getInput(putFieldNode, 2));
			StateTuple state = g.methodMem.readState(nodeToPut);
			assert state != null : "The object " + nodeToPut + " is unknown while storing in field " + field;


			// For every possibly referenced object, which contains the affected field, store the new
			// reference
			ContextAndNode nodeContainingField = new ContextAndNode(g.context, NodeHelper.getInput(
					putFieldNode, 1));
			PointsToSet pts = g.methodMem.getAllocSites(nodeContainingField);

			// potentially more than one field affected by the getfield, so read state from all of them
			// and combine then in one new reference for the getField
			for (ContextAndNode allocSite : pts) {
				FieldID fieldID = new FieldID(field, allocSite);

				// store reference in memory
				g.methodMem.setFieldRef(fieldID, state, nodeToPut, pts.isSingleConcrete());
			}

			g.methodMem.printDumpCurrMem();
		}
		return putFieldNode.getValue(MemExec.MEM);
	}

	public ValueNode handleStore(ArrayStore storeNode) {
		// Don't do anything - an ArrayLoad will just return the types of all affected allocation sites.
		// As every access to an array can only create weak updates - because the really affected object
		// can not be determined, it is valid to not store any references at all.
		
		// Return memory (those are EnumTupleNode<DataMemExec>)
		return storeNode.getValue(MemExec.MEM);
	}

	/* NEW ARRAY */
	public ValueNode handleNew(NewArray newNode) {
		// Don't do anything - an ArrayLoad will just return the types of all affected allocation sites.
		// As every access to an array can only create weak updates - because the really affected object
		// can not be determined, it is valid to not store any references at all.
		Reference arrayRef = new Reference(NodeHelper.getType(newNode));
		arrayRef.addState(UnknownState.getInstance());
		ContextAndNode newArrayCNode = new ContextAndNode(g.context, newNode);
		g.methodMem.putAsReference(newArrayCNode, arrayRef);
		
		// Return memory (those are EnumTupleNode<DataMemExec>)
		return newNode.getValue(DataMemExec.MEM);
	}

	private Reference retrieveFieldRef(FieldID fieldID) {
		Reference fieldRef = g.methodMem.getFieldRef(fieldID);
		// Check if field already defined
		if (fieldRef == null) {
			// Not defined yet, so use before first explicit definition -> TOP state
			g.log.debug("Field " + fieldID + " has not been discovered yet, assigning TOP");

			// Determine type and build TOP state
			Type fieldType = fieldID.getField().getType();
			Specification spec = g.specRep.getSpecification(fieldType);
			IState state;
			if (spec == null) {
				//specification unknown, so state is as well
				state = UnknownState.getInstance();
			} else {
				state = TopState.getTop(spec, g.analysisConfig.getTopStateMode());
			}

			// Create the new field Reference
//			fieldRef = new Reference(fieldType);
//			fieldRef.addState(state);
//			fieldRef.setContainsTop(true);
			MutableStateTuple stateTuple = new MutableStateTuple(fieldType);
			stateTuple.addState(state);
			fieldRef =g.methodMem.setFieldRef(fieldID, stateTuple, null, true);
		}
		return fieldRef;
	}

}
