package de.unisb.prog.typestate.ir;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import de.unisb.prog.analysis.Context;
import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.Call;
import de.unisb.prog.jfirm.ir.Constant;
import de.unisb.prog.jfirm.ir.DataMemExec;
import de.unisb.prog.jfirm.ir.MethodGraph;
import de.unisb.prog.jfirm.ir.Mode;
import de.unisb.prog.jfirm.ir.Type;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.typestate.FirmConnector;
import de.unisb.prog.typestate.IState;
import de.unisb.prog.typestate.NullState;
import de.unisb.prog.typestate.Specification;
import de.unisb.prog.typestate.TopState;
import de.unisb.prog.typestate.TypestateHelper;
import de.unisb.prog.typestate.UnspecifiedState;
import de.unisb.prog.typestate.ir.FirmWalker.Globals;
import de.unisb.prog.typestate.util.NodeHelper;
import de.unisb.prog.typestate.util.Types;
import de.unisb.prog.typestate.violation.AlternateMethodsSuggestion;
import de.unisb.prog.typestate.violation.CallViolationInfo;
import de.unisb.prog.typestate.violation.ISuggestion;
import de.unisb.prog.typestate.violation.IViolationInfo;
import de.unisb.prog.typestate.violation.StateExclusionSuggestion;

public class CallHandler {

	private static final char	CALLHANDLER_LOG_TAG	= 'C';

	private Globals				g;

	/** If > 0, no violations are reported. int instead of boolean to support nested dis/enable requests */
	private int					ignoreErrors;

	public void init(Globals g) {
		this.g = g;
		this.ignoreErrors = 0;
	}

	public ValueNode evaluateCall(Call call) {

		ValueNode nextMem = call.getValue(DataMemExec.MEM);
		assert nextMem != null : "Malformed jFirm graph: Call node does not have a result-memory predecessor (getValue(MEM)==null)";
		// Find out whether this is an analysis instruction or a regular call
		if ("de/unisb/prog/typestate/TypestateHelper".equals(call.getTargetMethod().getOwner().toString())) {

			// TYPESTATE ANALYSIS INSTRUCTION
			if (g.log.debug())
				g.log.debug("Encountered analysis instruction: " + call, CALLHANDLER_LOG_TAG);
			try {
				handleInstruction(call);
			} catch (Exception e) {
				// TODO Generate appropriate violation
				e.printStackTrace();
			}

		} else {
			// REGULAR METHOD INVOCATION
			handleInvocation(call);
		}
		g.methodMem.printDumpCurrMem();
		return nextMem;
	}

	/**
	 * Evaluates a CALL node of methods on {@link TypestateHelper}, which are interpreted as annotations.
	 * 
	 * @param call
	 */
	private void handleInstruction(Call call) {
		de.unisb.prog.jfirm.ir.Type.ClassType.Method instrType = call.getTargetMethod();
		String instrName = instrType.getName();

		// determine which instruction TYPE this is
		if (instrName.startsWith("state")) {

			// The first parameter always denotes the affected instance
			ContextAndNode affectedNode = new ContextAndNode(g.context, NodeHelper.getInput(call, 1));

			// skip "state"
			instrName = instrName.substring(6);

			// second input is the name of the state
			ValueNode stringParam = call.getIn(2);
			if (!(stringParam instanceof Constant && stringParam.getMode() == Mode.A && ((Constant) stringParam)
					.getValue() instanceof String)) {
				throw new IllegalArgumentException("The second parameter of " + call
						+ " must be a non-null String constant (was " + stringParam + ")");
			}

			// This must be a string
			String stateName = (String) ((Constant) stringParam).getValue();

			// Retrieve the current states of the affected input target
			// Is it a reference or an allocation site?
			StateTuple targetState = g.methodMem.readState(affectedNode);
			if (targetState == null) {
				g.log.error("Could not find state information for " + affectedNode
						+ " in analysis instruction " + call + ". Instruction is ignored",
						CALLHANDLER_LOG_TAG);
				return;
			}

			if ("print".equals(instrName)) {
				// Just output the current state together with the message
				if (g.log.info())
					g.log.info("State info for " + affectedNode + ": Typestate " + targetState
							+ ", pointsto " + g.methodMem.getAllocSites(affectedNode) + " -- " + stateName);
				return;
			}

			// Find spec
			Type affectedNodeType = NodeHelper.getType(affectedNode.getNode());
			Specification spec = g.specRep.getSpecification(affectedNodeType);
			if (spec == null) {
				if (g.log.warn())
					g.log.warn("The specification for the type " + affectedNodeType + " of node "
							+ affectedNode + " in state " + targetState
							+ " could not be determined for evaluation of an analysis instruction",
							CALLHANDLER_LOG_TAG);
				return;
			}
			// END find spec

			// find the given state
			IState state = spec.getStates().get(stateName);
			if (state == null) {
				if (g.log.warn())
					g.log
							.warn("The type of alloc. site " + targetState + " for reference node "
									+ affectedNode + " does not know a state named " + stateName,
									CALLHANDLER_LOG_TAG);
				return;
			}

			MutableStateTuple newState = targetState.mutableClone();

			if ("ignore".equals(instrName)) {
				// Ignore the state in parameter 2
				newState.removeState(state);
			} else if ("assume".equals(instrName)) {
				// Add the state in parameter 2
				newState.addState(state);
			} else if ("set".equals(instrName)) {
				newState.clearTypestate();
				newState.addState(state);
			}

			// update state
			try {
				// perform strong update because the given changes must DEFINITELY be performed
				g.methodMem.updateState(affectedNode, newState, true);
			} catch (IllegalArgumentException e) {
				g.log.error("Could not find state information for " + affectedNode
						+ " in analysis instruction " + call + ". Instruction is ignored",
						CALLHANDLER_LOG_TAG);
				return;
			}
		} else if (instrName.startsWith("ignoreErrors")) {
			// Turn error reporting on or off
			if (instrName.endsWith("begin")) {
				// Turn off error reporting
				if (this.ignoreErrors < Integer.MAX_VALUE) {
					this.ignoreErrors++;
				}
			} else if (instrName.endsWith("end")) {
				// Turn on error reporting
				if (this.ignoreErrors > 0) {
					this.ignoreErrors--;
				}
			}
		}
	}

	/**
	 * Evaluates a CALL node for an ordinary method invocation.
	 * 
	 * @param call
	 * @throws InterruptedException
	 */
	private void handleInvocation(Call call) {

		// OVERVIEW:
		// 1. Gather states of all allocation sites pointed at by "this" when the method IS CALLED
		// 2. Perform transitions on all parameters and the return value (if any), either by descending
		// into
		// the method call or setting to some meta state
		// 3. Perform transition given by the specification of the "this"' type on all alloc sites from 1.

		boolean isStatic = call.getTargetMethod().is_static();

		ContextAndNode callNode = new ContextAndNode(g.context, call);

		/* ---------------------------------- */
		/* PROCESSING STATE OF THIS REFERENCE */
		/* ---------------------------------- */

		// Remember the bad state (if reached) separately for giving error messages. It will not be kept in
		// the target state.
		IState badState = null;

		// Same for the unspecified state
		IState unspecState = null;

		// The signature of the Call-MethodGraph
		String methodSignature = Types.getSignature(call.getTargetMethod());

		// The "this" reference (if method is not static)
		ContextAndNode thisRef = null;

		// The new state for the this reference (if method is not static);
		MutableStateTuple newState = null;

		// The state of the "this" reference (if method is not static)
		StateTuple oldThisState = null;

		// For debug output: Carries the message about the transition (must be built prior to recursive
		// evaluation because otherwise the affected StateTuples might have changed
		String transString = "";

		boolean analyzeTransition;
		if (!isStatic) {
			// The "this" reference
			ValueNode firstParam = NodeHelper.getInput(call, 1);
			thisRef = new ContextAndNode(g.context, firstParam);

			analyzeTransition = NodeHelper.isDataNode(firstParam)
					&& !g.vioStore.isDisabled(g.methodMem.getActualNode(thisRef));
		} else {
			analyzeTransition = false;
		}

		// Perform transition on the this reference if this method is not static. The rest is unknown for
		// now
		if (analyzeTransition) {

			oldThisState = g.methodMem.readState(thisRef);

			assert oldThisState != null : "State of input " + thisRef + " to call " + call + " is unknown";

			// HashMap<ContextAndNode, StateTuple> oldStates = new HashMap<ContextAndNode, StateTuple>();
			//			
			// PointsToSet allocSites = g.methodMem.getAllocSites(thisRef);
			//
			// strongUpd = allocSites.size() == 1;// allocSites.isSingleConcrete();
			// if (allocSites.isTop()) {
			// // FIXME: ignored for now, but must be considered when pointsToTargets information becomes
			// better
			// g.log.debug("CallHandler.evaluateCall(call==" + call + "): pointsToTargets information for "
			// + thisRef + " returned TOP " + allocSites, CALLHANDLER_LOG_TAG);
			// }
			//
			// // Iterate over all affected instances
			// for (ContextAndNode allocSite : allocSites) {
			//
			// // Get the old state of the current this reference
			// StateTuple oldState = g.methodMem.readAllocSite(allocSite);
			//
			// // if no states are stored, then nothing may ever change -> just skip this instance
			// if (oldState.isBottom()) {
			// g.log.trace("No states for " + allocSite + ", skipping", CALLHANDLER_LOG_TAG);
			// continue;
			// }
			//
			// oldStates.put(allocSite, oldState);
			// }

			// First try transition on the reference state.
			// If no error:
			// --Update reference, done.
			// If error:
			// --Cheap: Generate violation and suggestions depending on reference state.
			// --Better: Get allocation sites, perform transition for each, and generate
			// ----violations/suggestions for each allocSite.

			// If transition to bad state, gather potential alternative methods here
			Set<String> potentialAlternativeMethods = new HashSet<String>();

			// If set, it might be an option to propose ignoring any bad transitions in this state because
			// of
			// some transition nondeterministically performing good and bad transitions
			boolean mightIgnoreBadDestState = false;

			// If all trans from those states are bad we must ignore them to resolve conflict
			Set<IState> mustIgnoreSourceStates = new HashSet<IState>();

			// for (Entry<ContextAndNode, StateTuple> entry : oldStates.entrySet()) {
			int numBad = 0, numUnspec = 0, numGood = 0;

			// ContextAndNode thisRef = entry.getKey();
			// StateTuple oldThisState = entry.getValue();

			newState = new MutableStateTuple(oldThisState.getType());
			for (IState state : oldThisState) {
				// Ignore states which are already bad
				if (state.isBad()) {
					g.log.warn(this + " called on " + thisRef + " in bad state - should not happen",
							CALLHANDLER_LOG_TAG);

					// more than one bad state not allowed in specification
					assert badState == null || badState == state : "two different bad states for " + thisRef
							+ ": " + badState + " and " + state;
					badState = state;

					continue;
				}

				Specification spec = state.getTypeSpecification();

				// Even if an allocSite weakly contains more than one state, a call to <init> does !only!
				// affect the starting state of the object (as it is directly called after object
				// creation)
				if (methodSignature.contains("<init>") && oldThisState.stateCount() > 1) {
					if (spec.getStartingState() != state) {
						// this is an uninited state and should only be considered for constructor
						// calls OR when this is the only state
						if (g.log.debug())
							g.log.debug("skipping source state " + state
									+ " because it's not start state of " + spec, CALLHANDLER_LOG_TAG);

						// keep unmodified state
						newState.addState(state);
						continue;
					}
				}

				// check if the "this" reference might be null
				if (NullState.isNull(state)) {

					// add a new violation. It is a must violation, if null is the only source state
					g.vioStore.addNullPtrDerefViolation(callNode, thisRef, oldThisState.stateCount() == 1);
					continue;
				}

				// Save current values to determine if there are some good/bad transitions for this
				// state
				int numGoodOld = numGood, numBadOld = numBad;

				Set<IState> newSubState = state.performTransition(methodSignature);

				for (IState nState : newSubState) {
					// Check if a transition from good to bad happened
					if (nState.isBad()) {
						if (nState instanceof UnspecifiedState) {
							// The requested transition was not recognised in the specification and
							// unspecified transitions are to be reported
							numUnspec++;

							unspecState = nState;
						} else {
							// count it as bad state
							numBad++;

							// more than one bad state not allowed in specification
							assert badState == null || badState == nState : "two different bad states for "
									+ thisRef + ": " + badState + " and " + nState;

							badState = nState;
						}
						// For computation of fix-callVioInfo, collect all methods which could be
						// alternatively called from this source state
						potentialAlternativeMethods.addAll(state.getTransitions().keySet());
					} else {

						// count it as good state
						numGood++;
					}
				}
				if (numBad - numBadOld != 0) {
					// new bad states

					if (numGood - numGoodOld != 0) {
						// bad and good states reachable
						mightIgnoreBadDestState = true;
					} else {
						// only bad ones reachable
						mustIgnoreSourceStates.add(state);
					}
				}
				// Add the new States from the transition to the new state set
				newState.update(newSubState, false);

			}

			if (g.log.debug()) {
				transString = String.format("%-15s on %-15s: %s --%s--> %s", call, thisRef, oldThisState,
						methodSignature, newState);
			}

			// If there were errors/unspec'd transitions (and error reporting is turned on)...
			if (numBad + numUnspec != 0 && this.ignoreErrors == 0) {
				/*                               */
				/* GATHER CORRECTION SUGGESTIONS */
				/*                               */

				/* ... collect alternative methods which might be called ... */

				// do not propose the current method as an alternative - for obvious reasons ;)
				potentialAlternativeMethods.remove(Types.getSignature(call.getTargetMethod()));
				ISuggestion suggAlternate = testTransitions(potentialAlternativeMethods, oldThisState
						.getStateSet(), badState, methodSignature);

				/* ... as well as states which might be ignored to resolve the problem. */
				ISuggestion suggIgnore = new StateExclusionSuggestion(mustIgnoreSourceStates,
						mightIgnoreBadDestState);
				/* END CORRECTION COMPUTATION */

				// Check if this is surely a violation or just a may-violation
				if (numGood == 0) {
					// All reached states are bad, so this is a must-violation
					// --> The only solution to this violation is choosing another method, ignoring
					// some state is useless
					IViolationInfo callVio = new CallViolationInfo(thisRef, new StateTuple(oldThisState),
							new StateTuple(newState), Types.getSignature(call.getTargetMethod()),
							suggAlternate);
					g.vioStore.addCallViolation(callNode, thisRef, numBad > 0, numUnspec > 0, callVio);

				} else {
					// Good and bad states -> may violation
					// So propose alternate methods as well as ignoring some states
					List<ISuggestion> suggs = new ArrayList<ISuggestion>(2);
					suggs.add(suggAlternate);
					suggs.add(suggIgnore);

					IViolationInfo callVio = new CallViolationInfo(thisRef, new StateTuple(oldThisState),
							new StateTuple(newState), Types.getSignature(call.getTargetMethod()), suggs);
					g.vioStore.addCallViolation(callNode, thisRef, numBad > 0, numUnspec > 0, callVio);

				}
			}
			// Don't store error state in new state
			if (badState != null) {
				newState.removeState(badState);
			}
			// Same for unspec state
			if (unspecState != null) {
				newState.removeState(unspecState);
			}

			// Do not update state of the reference unless the method call has been evaluated
			// --> the reference performs the actual transition AFTER execution of the method body

		}

		/* ----------------------------------------------- */
		/* PROCESSING STATE OF PARAMETERS AND RETURN VALUE */
		/* ----------------------------------------------- */

		boolean interprocEvalSuccessful;

		// If this method call is non static, we need to mute the violation repository for the "this"
		// reference
		if (!isStatic) {
			// No violations for "this" reference
			g.vioStore.disableViolationsForInstance(thisRef);

			// Perform interprocedural call analysis
			interprocEvalSuccessful = evalParamsRecursive(call, isStatic);

			g.vioStore.enableViolationsForInstance(thisRef);
			// Enable violations again
		} else {
			// No "this" to mute, just perform interprocedural call analysis
			interprocEvalSuccessful = evalParamsRecursive(call, isStatic);
		}

		if (!interprocEvalSuccessful) {
			// maximum level of recursion reached or no source for called method --> "guess" states

			// The index of the first real (not counting "this") parameter
			// set the state of all parameters to TOP
			for (int i = isStatic ? 1 : 2, n = call.getN(); i < n; i++) {
				ContextAndNode paramNode = new ContextAndNode(g.context, NodeHelper.getInput(call, 1));

				// Ignore non-reference parameters
				if (paramNode.getNode().getMode() != de.unisb.prog.jfirm.ir.Mode.A)
					continue;

				Type paramType = NodeHelper.getType(paramNode.getNode());
				Specification spec = g.specRep.getSpecification(paramType);
				// // FIXME replace following with a "getType" method, will often not find the correct
				// // supertype on phis
				// StateTuple curParamState = g.methodMem.readState(paramNode);
				// if (curParamState != null) {
				// for (IState state : curParamState) {
				// spec = state.getTypeSpecification();
				// if (spec != null)
				// break;
				// }
				// } else {
				// g.log.error("Could not determine state of " + paramNode
				// + " while deriving type information for TOP state creation", CALLHANDLER_LOG_TAG);
				// }

				if (spec == null) {
					g.log.debug("Could not get specification for type " + paramType + " of " + paramNode
							+ " while deriving type information for TOP state creation", CALLHANDLER_LOG_TAG);
				}
				// // Mark all used allocation sites as possibly escaping the method
				// for (ContextAndNode allocSiteNode : g.methodMem.getAllocSites(paramNode)) {
				// g.methodMem.setGlobal(allocSiteNode);
				// }

				// Create TOP state
				MutableStateTuple paramTopState = new MutableStateTuple(paramType);
				paramTopState.addState(TopState.getTop(spec, g.analysisConfig.getTopStateMode()));
				paramTopState.setContainsTop(true);

				// Update the reference to TOP state
				g.methodMem.makeReference(paramNode, paramTopState);

			}

			// set the state of the return value to top if it is a reference type
			ContextAndNode retNode = new ContextAndNode(g.context, call.getValue(DataMemExec.DATA));
			if (retNode.getNode().getMode() == Mode.A) {
				Set<IState> retState = new HashSet<IState>();
				Type returnType = NodeHelper.getType(call);
				Specification spec = g.specRep.getSpecification(returnType);
				retState.add(TopState.getTop(spec, g.analysisConfig.getTopStateMode()));

				// connect the returned value to the call node (as this is the only visible allocation
				// site if the analysis does not descend into the call)
				//
				// // The call itself will be the allocation site
				// g.methodMem.addPointsToTarget(retNode, callNode);

				// create allocation site in call
				MutableStateTuple retAllocSite = new MutableStateTuple(returnType);
				retAllocSite.overwrite(retState);
				retAllocSite.setContainsTop(true);
				g.methodMem.createAllocSite(callNode, retAllocSite);

				// //update state
				// PointsToSet retTargetSites = g.methodMem.getAllocSites(retNode);
				// for (ContextAndNode retTargetSite : retTargetSites) {
				// // The allocation site of the return value might have escaped the method (which was
				// // skipped)
				// g.methodMem.setGlobal(retTargetSite);
				// // always strong update to top? or better some top-state which knows the previous ones
				// AllocSite retAllocSite = new AllocSite(retTargetSite);
				// retAllocSite.overwrite(retState);
				// retAllocSite.setTop(true);
				// g.methodMem.updateAllocSite(retTargetSite, retAllocSite, true);
				// }
			}
		}

		/* ------------------------------------ */
		/* UPDATE STATE OF THE "THIS" REFERENCE */
		/* ------------------------------------ */
		if (analyzeTransition) {

			g.log.debug(transString, CALLHANDLER_LOG_TAG);

			// Finally write the new state of the reference (always perform strong update on the target)
			g.methodMem.updateState(thisRef, newState, true);
		}
	}

	/**
	 * @param call
	 * @param isStatic
	 * @throws InterruptedException
	 */
	private boolean evalParamsRecursive(Call call, boolean isStatic) {
		if (g.interprocDepth < g.analysisConfig.getMaxDepth()) {
			// analyze the target method to find the resulting state of the parameters

			de.unisb.prog.jfirm.ir.Type.ClassType.Method methodType = call.getTargetMethod();
			MethodGraph method = FirmConnector.getInstance().getFirmGraph(methodType);
			if (method != null) {
				// MethodGraph is available

				ContextAndNode[] paramNodes = new ContextAndNode[call.getN()];

				// process parameters (including "this" parameter, will be overwritten when the context is
				// left)
				for (int i = 1, n = call.getN(); i < n; i++) {
					// uncomment this if "this" parameter should not be included
					// if (i == 1 && !isStatic) {
					// continue;
					// }
					paramNodes[i] = new ContextAndNode(g.context, NodeHelper.getInput(call, i));
				}

				g.log.debug(">> Context switching to " + methodType.getName() + " (cur. recursion depth "
						+ g.interprocDepth + ") >>", CALLHANDLER_LOG_TAG);
				g.log.setIndentation(++g.interprocDepth);
				try {

					// store the current memory and context to be able to revert if an error occurs
					MethodMemory oldMem = g.methodMem;
					Context oldContext = g.context;
					Context newContext = g.context.addCall(call);

					// the destination node for the return value
					// ...but only consider it if it is a reference type
					ContextAndNode returnValNode = null;
					if (call.getValue(DataMemExec.DATA).getMode() == de.unisb.prog.jfirm.ir.Mode.A) {
						returnValNode = new ContextAndNode(g.context, call);// .getValue(DataMemExec.DATA));
					}
					oldMem.printDumpCurrMem();
					// perform context switch to another memory (which starts with the state of the current
					// mem)
					// and map the method's parameters to the "real" ValueNodes
					MethodMemory newMem = oldMem.contextSwitchOut(method, newContext, returnValNode,
							paramNodes); // !isStatic,
					// // put new memory in invocation history
					// invocations.put(invoc, newMem);

					// set new memory as current
					g.methodMem = newMem;

					// analyze the called method with the new memory
					g.context = newContext;
					try {
						g.walker.walk(method);
					} catch (Exception e) {
						g.log.debug("!! Interproc eval of " + call + " aborted due to " + e + " !!",
								CALLHANDLER_LOG_TAG);
						// reload old memory and context
						g.methodMem = oldMem;
						g.context = oldContext;
						// trigger intraprocedural evaluation
						return false;
					}
					g.context = g.context.returnFromCall();

					// from the context switched memory, retrieve states of all globally visible instances
					// which
					// are not destroyed when the called method returns
					oldMem.contextSwitchBack(g.methodMem, newContext);
					g.methodMem = oldMem;

					g.log.debug("<< Returning from context switch to " + methodType.getName()
							+ " (cur. recursion depth " + g.interprocDepth + ") <<", CALLHANDLER_LOG_TAG);
					// recursive parameter evaluation successful
					return true;
				} finally {
					g.log.setIndentation(--g.interprocDepth);
				}

			}
			// method not available
			return false;
		} else {
			g.log.debug("<> No context switch because recursion depth == max recursion depth ("
					+ g.analysisConfig.getMaxDepth() + ") <>", CALLHANDLER_LOG_TAG);
			return false;
		}
	}

	private enum TransRating {
		UNCHECKED, SAMESTATE, SAFECHANGE, SAFEENABLEDESIREDTRANS, ENABLEDESIREDTRANS, CHANGESTATE, BAD;
	}

	private AlternateMethodsSuggestion testTransitions(Set<String> transitions, Set<IState> sourceState,
			IState badState, String desiredTransition) {

		AlternateMethodsSuggestion ret = new AlternateMethodsSuggestion();

		// Store if a transition is
		// - not yet checked,
		// - staying in the same state,
		// - safely changing state,
		// - enabling the initally called transition
		// - changing state while reaching also bad states,
		// - disabled because at least one state deterministically would go to error.
		HashMap<String, TransRating> transitionRatings = new HashMap<String, TransRating>(transitions.size());

		// init transitionTypes
		for (String trans : transitions) {
			transitionRatings.put(trans, TransRating.UNCHECKED);
		}

		for (IState state : sourceState) {

			for (Iterator<String> it = transitions.iterator(); it.hasNext();) {
				String trans = it.next();

				// get current rating for this transition
				TransRating currentTransitionRating = transitionRatings.get(trans);

				// from this, compute the new rating
				TransRating newTransitionRating = rateStates(state, state.performTransition(trans), badState,
						desiredTransition, currentTransitionRating);

				// store the new rating
				transitionRatings.put(trans, newTransitionRating);

				if (newTransitionRating == TransRating.BAD) {
					// BAD means that the transition causes one state to deterministically go to error, so
					// remove from the toDo-List
					it.remove();
				}
			}
		}

		// Now fill the transitions into the result sets, directly skipping impossible transitions
		for (String trans : transitions) {
			TransRating rating = transitionRatings.get(trans);
			switch (rating) {
			case SAFEENABLEDESIREDTRANS:
				ret.safeEnablingTrans.add(trans);
				break;
			case ENABLEDESIREDTRANS:
				ret.unsafeEnablingTrans.add(trans);
				break;
			case SAMESTATE:
				ret.sameState.add(trans);
				break;
			case SAFECHANGE:
				ret.safeChanging.add(trans);
				break;
			case CHANGESTATE:
				ret.changing.add(trans);
				break;
			default:
				// May never happen
				g.log.warn("Call.testTransitions(): had transition '" + trans + "' marked " + rating
						+ " when composing callVioInfo for transitions " + transitions + " from "
						+ sourceState, CALLHANDLER_LOG_TAG);
			}
		}

		return ret;
	}

	private TransRating rateStates(IState sourceState, Set<IState> targetStates, IState badState,
			String desiredTransition, TransRating currentTransitionRating) {

		// a bad state always stays bad
		if (currentTransitionRating == TransRating.BAD)
			return TransRating.BAD;

		if (!targetStates.contains(badState)) {
			// this method is safe

			// check if method actually can change state of object
			if (targetStates.size() > 1 || !targetStates.contains(sourceState)) {
				// reaches different but no unsafe state

				if (desiredTransition != null) {
					// check if method enables the desired transition
					TransRating desiredTransRating = TransRating.UNCHECKED;

					// for each target state perform desired transition and check if there is some potential
					// error
					for (IState tState : targetStates) {
						desiredTransRating = rateStates(tState, tState.performTransition(desiredTransition),
								badState, null, desiredTransRating);
					}

					switch (desiredTransRating) {
					case SAMESTATE:
					case SAFECHANGE:
						// enables desired trans -> rate highest
						return TransRating.SAFEENABLEDESIREDTRANS;
					case CHANGESTATE:
						// enables desired trans but can also introduce non-det bad transitions
						return TransRating.ENABLEDESIREDTRANS;
					default:
						// no enabling of current transition, just go on with normal rating
					}
				}

				if (currentTransitionRating.ordinal() < TransRating.SAFECHANGE.ordinal())
					return TransRating.SAFECHANGE;
			} else {
				// no change in state
				if (currentTransitionRating.ordinal() < TransRating.SAMESTATE.ordinal())
					return TransRating.SAMESTATE;
			}
		} else {
			// // This source state may lead to a bad state, so tag as "has bad transition"

			if (targetStates.size() > 1) {
				// not only reaches error state but others as well
				if (currentTransitionRating.ordinal() < TransRating.CHANGESTATE.ordinal())
					return TransRating.CHANGESTATE;

			} else {
				// ONLY error state reached -> the current method can not be proposed.
				if (currentTransitionRating.ordinal() < TransRating.BAD.ordinal())
					return TransRating.BAD;
			}
		}
		// if this point is reached, nothing has changed, so just return old rating
		return currentTransitionRating;
	}
}
