package de.unisb.prog.typestate.ir;

import java.util.Iterator;
import java.util.Stack;

import de.unisb.prog.analysis.AnalysisContext;
import de.unisb.prog.analysis.Context;
import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.ArrayLoad;
import de.unisb.prog.jfirm.ir.ArrayStore;
import de.unisb.prog.jfirm.ir.Block;
import de.unisb.prog.jfirm.ir.Call;
import de.unisb.prog.jfirm.ir.Cast;
import de.unisb.prog.jfirm.ir.DataMemExec;
import de.unisb.prog.jfirm.ir.EnumTupleNode;
import de.unisb.prog.jfirm.ir.GetField;
import de.unisb.prog.jfirm.ir.GetStatic;
import de.unisb.prog.jfirm.ir.MemExec;
import de.unisb.prog.jfirm.ir.MethodGraph;
import de.unisb.prog.jfirm.ir.Mode;
import de.unisb.prog.jfirm.ir.NewArray;
import de.unisb.prog.jfirm.ir.NewObject;
import de.unisb.prog.jfirm.ir.Node;
import de.unisb.prog.jfirm.ir.Param;
import de.unisb.prog.jfirm.ir.Phi;
import de.unisb.prog.jfirm.ir.PutField;
import de.unisb.prog.jfirm.ir.PutStatic;
import de.unisb.prog.jfirm.ir.Throw;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.typestate.AnalysisLogger;
import de.unisb.prog.typestate.IAnalysisConfiguration;
import de.unisb.prog.typestate.ISpecificationRepository;
import de.unisb.prog.typestate.IViolationStore;

/**
 * 
 * @author mikerofone
 * 
 */
public class FirmWalker {

	private class WorklistEntry {
		Block		prevBlock;
		ValueNode	firstBlockVNode;

		public WorklistEntry(Block prevBlock, ValueNode firstBlockMem) {
			// this.prevBlock = prevBlock;
			this.firstBlockVNode = firstBlockMem;
		}

		@Override
		public String toString() {
			return /* prevBlock + ":" + */firstBlockVNode.toString();
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#hashCode()
		 */
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((this.firstBlockVNode == null) ? 0 : this.firstBlockVNode.hashCode());
			result = prime * result + ((this.prevBlock == null) ? 0 : this.prevBlock.hashCode());
			return result;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj instanceof WorklistEntry) {
				WorklistEntry other = (WorklistEntry) obj;
				return (this.firstBlockVNode == null ? other.firstBlockVNode == null : this.firstBlockVNode
						.equals(other.firstBlockVNode))
						&& (this.prevBlock == null ? other.prevBlock == null : this.prevBlock
								.equals(other.prevBlock));
			} else {
				return false;
			}
		}

	}

	private static final char	WALKER_LOG_TAG	= 'W';

	/* Elements for global state */
	protected class Globals {
		public FirmWalker				walker;
		public MethodMemory				methodMem;
		public ISpecificationRepository	specRep;
		public IViolationStore			vioStore;
		public IAnalysisConfiguration	analysisConfig;
		public AnalysisLogger			log;
		public Context					context;
		public int						interprocDepth;
	}

	private AnalysisLogger		log;
	private Globals				globals;

	/* Handlers for the different Types of Nodes */
	// TODO: Create Handler class with map from Opcode to HandlerClass -> Handler<OPCODE> interface providing
	// handle/reset methods
	private CallHandler			callHandler			= new CallHandler();
	private NewHandler			newHandler			= new NewHandler();
	private ParamHandler		paramHandler		= new ParamHandler();
	private ReturnHandler		returnHandler		= new ReturnHandler();
	private ArrayFieldHandler	arrayFieldHandler	= new ArrayFieldHandler();
	private BlockHandler		blockHandler		= new BlockHandler();
	private CastHandler			castHandler			= new CastHandler();

	public FirmWalker(ISpecificationRepository specs, IViolationStore vioStore,
			IAnalysisConfiguration anaConfig) {

		// Init global structures
		globals = new Globals();
		globals.specRep = specs;
		globals.vioStore = vioStore;

		globals.walker = this;
		globals.analysisConfig = anaConfig;
		this.log = globals.log = AnalysisLogger.getInstance();

		// Init per-method-analysis-structures
		reset();
	}

	/**
	 * Initializes the FirmWalker before starting a new analysis run for a method.
	 */
	public void reset() {
		globals.interprocDepth = 0;
		globals.methodMem = null;
		globals.context = null;

		// Init handlers
		this.callHandler.init(globals);
		this.newHandler.init(globals);
		this.paramHandler.init(globals);
		this.returnHandler.init(globals);
		this.arrayFieldHandler.init(globals);
		this.blockHandler.init(globals);
		this.castHandler.init(globals);

	}

	/**
	 * Inits the method-local structures and starts walking over the jFirm graph.
	 * 
	 * @param methodGraph
	 *            The jFirm {@link MethodGraph} which is to be analyzed
	 * @throws InterruptedException
	 */
	public void walk(MethodGraph methodGraph) {
		log.trace("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!", WALKER_LOG_TAG);
		log.debug("-- WALKER: BEGINNING WITH METHOD " + methodGraph.getName() + " --", WALKER_LOG_TAG);
		log.trace("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!", WALKER_LOG_TAG);
		log.trace("Walking down graph of " + methodGraph.getName(), WALKER_LOG_TAG);

		// Set context if not already done so
		if (globals.context == null) {
			// this is the start of a new analysis run
			globals.context = new AnalysisContext();
		}

		// Initialize PARAMS and get first MethodMemory in Block 0
		Stack<WorklistEntry> workstack = initBlockZero(methodGraph);

		// Start processing the workstack
		process(workstack, globals);

	}

	/**
	 * Initializes the global structures for Block 0:<br>
	 * <ul>
	 * <li>Finds and initializes the PARAMs</li>
	 * <li>Creates a new workstack and fills it with the users of the first memory Node in Block 0</li>
	 * <li>Returns the pre-filled workstack</li>
	 * </ul>
	 * 
	 * @param methodGraph
	 *            the jFirm method
	 * @return the workstack which must be processed
	 */
	private Stack<WorklistEntry> initBlockZero(MethodGraph methodGraph) {
		Block startBlock = methodGraph.getStartBlock();

		Stack<WorklistEntry> workstack = new Stack<WorklistEntry>();

		// init first memory for Block 0

		if (globals.methodMem == null) {
			// Beginning of method analysis, so create empty new memory
			globals.methodMem = new MethodMemory(methodGraph);
			globals.methodMem.createInBlock(startBlock);
			MethodMemory.initMemory(methodGraph, globals.analysisConfig.getMaxDepth());
		} else {
			// some memory already loaded (context switch), do nothing
		}
		globals.methodMem.loadBlockMem(startBlock);

		// find first memory node
		ValueNode[] blockZeroNodes = methodGraph.getParameters();
		assert blockZeroNodes != null && blockZeroNodes.length > 0 : "method.getParameters() returned null or zero-length array";
		ValueNode firstMemory = blockZeroNodes[0];

		// iterate over parameters and initialize them if necessary
		ContextAndNode[] parameterTargets = globals.methodMem.getParameterMapping();
		for (int i = 1, n = blockZeroNodes.length; i < n; i++) {

			if (i >= parameterTargets.length || parameterTargets[i] == null) {
				// No information for this parameter available, so init it with some value (probably TOP)

				// param is its own allocation site
				paramHandler.initParam((Param) blockZeroNodes[i]);
			} else {
				// This parameter is remapped, so do nothing
			}

		}

		if (firstMemory == null) {
			// no mem node or no children found --> malformed MethodGraph!
			throw new IllegalArgumentException("Given method " + methodGraph.getName()
					+ " is malformed: No PARAM <MEM> in Block 0 (or it has no users).");
		}

		// init out-state for Block 0
		globals.methodMem.updateOutBlock();

		// put all users of first memory into the workstack
		for (Iterator<Node> memUserIt = firstMemory.usersIterator(); memUserIt.hasNext();) {
			Node memUser = memUserIt.next();

			// Create in-Block memory
			globals.methodMem.updateInBlock(memUser.getBlock(), startBlock);
			workstack.add(new WorklistEntry(startBlock, (ValueNode) memUser));
		}

		return workstack;
	}

	/**
	 * Processes the Nodes on the given workstack, walking down the Nodes' users until all Nodes have been
	 * analyzed exhaustively. Recursion and progressing is realized with the workstack and an extra stack -
	 * the method itself is non-recursive.
	 * 
	 * @param workstack
	 *            The workstack which is to be processed (will be empty on return)
	 * @param methodMem
	 *            The memory of the current method
	 * @throws InterruptedException
	 */
	private void process(Stack<WorklistEntry> workstack, Globals globals) {

		// Get the methodMemory from globals
		MethodMemory methodMem = globals.methodMem;

		// The inner stack just is another workstack, which enables the algorithm to decide whether to
		// schedule a node in this same path or to be processed when the whole path has ended (to minimize
		// recalculation count of PHI for example)
		// XXX FOR EVAL OF TWO WORKLIST: set stack to workstack
		Stack<WorklistEntry> stack = globals.analysisConfig.walkerUseOnlyOneWorklist() ? workstack
				: new Stack<WorklistEntry>();

		// Pick the next ValueNode from the outer workstack and follow its path until analysis of the path
		// terminates. If the workstack is empty when a path terminates, the analysis is finished.
		while (!workstack.isEmpty()) {

			// assert stack.isEmpty() : "Error: Stack not empty when taking new el't from workstack";
			if (log.trace())
				log.trace("Worklist: " + workstack.toString(), WALKER_LOG_TAG);
			stack.push(workstack.pop());

			while (!stack.isEmpty()) {
				// log.debug("  Worklist: " + workstack.toString());
				// log.debug("  Stack: " + stack.toString());
				// BLOCK
				WorklistEntry entry = stack.pop();
				Block block = entry.firstBlockVNode.getBlock();
				if (log.trace())
					log.trace(" Processing Worklist item: " + entry, WALKER_LOG_TAG);

				globals.methodMem.loadBlockMem(block);

				blockHandler.handleBlock(block, entry.firstBlockVNode, entry.prevBlock);

				ValueNode lastMem = processEntry(entry);

				if (!methodMem.currentOutStateSaved()) {
					// save the current block's out-state
					methodMem.updateOutBlock();
				}

				if (lastMem != null) {

					for (Iterator<Node> uIt = lastMem.usersIterator(); uIt.hasNext();) {
						Node node = uIt.next();
						if (log.trace())
							log.trace(String.format(
									" fromBlock==%s -> block==%s -> lastMem==%s -> memChild==%s",
									entry.prevBlock, block, lastMem, node), WALKER_LOG_TAG);
						Block nodeBlock = node.getBlock();

						// update the destination block's memory
						if (!methodMem.updateInBlock(nodeBlock, block)) {
							// this did not update the in state of the target Block, so do not revisit it
							continue;
						}

						if (node instanceof Block) {
							// control flow nodes (RETURN) have blocks as children, ignore them.
							// they will be visited if they contain a memory node
							continue;
						}

						ValueNode vNode = (ValueNode) node;
						WorklistEntry newEntry = new WorklistEntry(block, vNode);
						if (vNode instanceof Phi && ((Phi) vNode).getMode() == Mode.M) {
							// memory phi, just put it to the outer workstack to evaluate it later.

							// if it is already on the workstack, do not add a second time. one iteration
							// suffices because in-states are always merged
							if (!workstack.contains(newEntry)) {
								workstack.add(newEntry);
							}
						} else {
							stack.add(newEntry);
						}

					}
					// log.trace(String.format(" valueNode==%s -> lastMem==%s -- no more users", vNode,
					// lastMem),WALKER_LOG_TAG);

				} else {
					if (log.trace())
						log.trace(String.format(" fromBlock==%s -> block==%s -> lastMem==null",
								entry.prevBlock, block), WALKER_LOG_TAG);
				}
			}
		}
		// Worklist empty: MethodGraph is done
		log.trace("Worklist cleared.", WALKER_LOG_TAG);
		methodMem.printDumpFinalMem();
	}

	@SuppressWarnings("unchecked")
	private ValueNode processEntry(WorklistEntry entry) {

		ValueNode vNode = entry.firstBlockVNode;
		Block currentBlock = vNode.getBlock();
		ValueNode nextMem = null;

		while (vNode != null) {
			boolean endTrace = false;

			if (log.trace())
				log.trace(" Processing ValueNode: " + vNode, WALKER_LOG_TAG);
			switch (vNode.getOpcode()) {

			case CALL:
				nextMem = callHandler.evaluateCall((Call) vNode);
				break;
			case NEW:
				nextMem = newHandler.evaluateNew((NewObject) vNode);
				break;
			case RETURN:
				nextMem = returnHandler.evaluateReturn(vNode);
				endTrace = true;
				break;
			case GETFIELD:
				nextMem = arrayFieldHandler.handleGetField((GetField) vNode);
				break;
			case GETSTATIC:
				nextMem = arrayFieldHandler.handleGetStatic((GetStatic) vNode);
				break;
			case ALOAD:
				nextMem = arrayFieldHandler.handleLoad((ArrayLoad) vNode);
				break;
			case PUTFIELD:
				nextMem = arrayFieldHandler.handlePutField((PutField) vNode);
				break;
			case PUTSTATIC:
				nextMem = arrayFieldHandler.handlePutStatic((PutStatic) vNode);
				break;
			case ASTORE:
				nextMem = arrayFieldHandler.handleStore((ArrayStore) vNode);
				break;
			case PHI:
				// Might never encounter a reference phi node here
				assert vNode.getMode() == Mode.M : "FirmWalker found a data Phi node: " + vNode;

				// it's a memory phi, so it is the next memory node itself
				nextMem = vNode;
				break;
			case CAST:
				nextMem = castHandler.handleCast((Cast) vNode);
				break;
			case NEWARRAY:
				nextMem = arrayFieldHandler.handleNew((NewArray) vNode);
				break;
			case THROW:
				// TODO implement
				nextMem = ((Throw) vNode).getValue(MemExec.MEM);
				break;
			case SYNC:
				// TODO implement
				throw new UnsupportedOperationException("SYNC not implemented yet - " + vNode);
			case PARAM:
				throw new UnsupportedOperationException("PARAM may not be encountered - " + vNode);
			default:
				log.debug(" Walker met unhandled vNode: " + vNode, WALKER_LOG_TAG);
				if (vNode instanceof EnumTupleNode<?>) {
					try {
						nextMem = ((EnumTupleNode<MemExec>) vNode).getValue(MemExec.MEM);
					} catch (ClassCastException e) {
					}
					if (nextMem == null) {
						try {
							nextMem = ((EnumTupleNode<DataMemExec>) vNode).getValue(DataMemExec.MEM);
						} catch (ClassCastException f) {
						}
						if (nextMem != null) {
							break;
						}
						log.error(" Could not find next memory Node from here, trace ends", WALKER_LOG_TAG);
						nextMem = null;
					}
				}
			}
			if (!endTrace) {
				if (nextMem == null) {
					// trace ended unexpectedly
					throw new UnsupportedOperationException(
							"Could not handle instruction type of " + vNode
									+ " and could not find next memory node to just skip it either.");
				} else if (nextMem.getMode() != Mode.M) {
					// illegal next memory: no users for memory
					throw new MalformedMethodGraphException(
							"Memory chain contains non-memory node " + nextMem
									+ " (current instruction node: " + vNode + ")");
				} else if (!nextMem.usersIterator().hasNext()) {
					// illegal next memory: no users for memory
					throw new MalformedMethodGraphException(
							"Memory chain unexpectedly ended but not at end of method graph (last instruction node: "
									+ vNode + ")");
				}
			}
			vNode = nextNodeInBlock(nextMem, currentBlock);
		}
		return nextMem;
	}

	/**
	 * Returns the next node in this block, or <code>null</code> if the given one is the last one in the
	 * block.
	 * 
	 * @param node
	 *            The node whose successor shall be found
	 * @param block
	 *            The current block
	 * @return
	 */
	private ValueNode nextNodeInBlock(ValueNode node, Block block) {

		// Inside a block, a memory can only have one user (or it is the last memory in the block)
		Iterator<Node> childIter = node.usersIterator();
		if (childIter.hasNext()) {
			Node child = childIter.next();
			if (!childIter.hasNext() && child.getBlock() == block) {
				// exactly one child in this same block -> not end of block
				assert child instanceof ValueNode;
				return (ValueNode) child;
			}
		}
		// fallthrough: either no or more than one child, or children in different blocks
		return null;
	}

}
