package de.unisb.prog.typestate.ir;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.typestate.AnalysisLogger;

public class MemorySnapshot {

	private static final AnalysisLogger	log	= AnalysisLogger.getInstance();

	public static enum NodeType {
		ALLOC, REF
	}

	private HashMap<ContextAndNode, AllocSite>	allocSites;
	private HashMap<ContextAndNode, Reference>	references;
	private HashMap<FieldID, Reference>			fieldRefs;

	public MemorySnapshot() {
		allocSites = new HashMap<ContextAndNode, AllocSite>();
		references = new HashMap<ContextAndNode, Reference>();
		fieldRefs = new HashMap<FieldID, Reference>();
	}

	public boolean contains(NodeType type, ContextAndNode key) {
		switch (type) {
		case ALLOC:
			return allocSites.containsKey(key);
		case REF:
			return references.containsKey(key);
		default:
			throw new IllegalArgumentException("Illegal type " + type);
		}
	}

	public boolean contains(Reference ref) {
		return references.containsValue(ref) || fieldRefs.containsValue(ref);
	}

	public boolean containsField(FieldID field) {
		return fieldRefs.containsKey(field);
	}

	public void putRef(ContextAndNode key, Reference ref) {
		// if (references.values().contains(ref)) {
		// ContextAndNode k = null;
		// for (Entry<ContextAndNode, Reference> e : references.entrySet()) {
		// if (e.getValue() == ref) {
		// k = e.getKey();
		// break;
		// }
		//					
		// }
		// if (!k.equals(key))
		// System.out.println("want to map " + ref + " to " + key + " but already mapped to " + k);
		// }
		// if (fieldRefs.values().contains(ref)) {
		// System.out.println("kj");
		// }
		references.put(key, ref);
	}

	public void putAlloc(ContextAndNode key, AllocSite alloc) {
		allocSites.put(key, alloc);
	}

	public void putFieldRef(FieldID field, Reference ref) {
		// if (fieldRefs.values().contains(ref)) {
		// FieldID k = null;
		// for (Entry<FieldID, Reference> e : fieldRefs.entrySet()) {
		// if (e.getValue() == ref) {
		// k = e.getKey();
		// break;
		// }
		//					
		// }
		// if (!k.equals(field))
		// System.out.println("want to map " + ref + " to " + field + " but already mapped to " + k);
		// }
		// if (references.values().contains(ref)) {
		// System.out.println("kj");
		// }
		fieldRefs.put(field, ref);
	}

	public Reference getRef(ContextAndNode key) {
		return references.get(key);
	}

	public AllocSite getAlloc(ContextAndNode key) {
		return allocSites.get(key);
	}

	public Reference getFieldRef(FieldID field) {
		return fieldRefs.get(field);
	}

	public void killRef(ContextAndNode refEntry) {
		references.remove(refEntry);
	}

	/**
	 * Merge the input MemorySnapshot into this. In fact, merges the given allocation sites, references and
	 * fields into the current snapshot, keeping the links intact.
	 */
	public boolean merge(MemorySnapshot from) {
		boolean changed = false;

		// because the allocation sites keep links to their aliasing references, a mapping from old ref to new
		// ref must be built for correct duplication of the allocation sites
		HashMap<Reference, Reference> oldToNewRef = new HashMap<Reference, Reference>(from.references.size());

		// merge references
		changed = mergeReferences(from.references, oldToNewRef, false) || changed;

		// merge fields
		changed = mergeFields(from.fieldRefs, oldToNewRef, false) || changed;

		// merge allocation sites
		changed = mergeAllocs(from.allocSites, from.allocSites.keySet(), oldToNewRef, false) || changed;

		return changed;
	}

	public void mergeOnlyFieldsAndAllocs(MemorySnapshot from, Set<ContextAndNode> mask) {
		HashMap<Reference, Reference> oldToNewRef = new HashMap<Reference, Reference>();
		mergeFields(from.fieldRefs, oldToNewRef, true);
		mergeAllocs(from.allocSites, mask, oldToNewRef, true);

	}

	private boolean mergeAllocs(HashMap<ContextAndNode, AllocSite> allocs, Set<ContextAndNode> mask,
			HashMap<Reference, Reference> oldToNewRef, boolean strongUpdate) {
		boolean changed = false;

		if (allocs == null || mask == null) {
			throw new IllegalArgumentException("mergeFields(allocs==" + allocs + ", mask==" + mask
					+ "): No null allowed!");
		}

		// merge allocation sites
		for (ContextAndNode allocNode : mask) {
			AllocSite oldAlloc = allocs.get(allocNode);
			if (oldAlloc == null) {
				if (log.debug())
					log.debug("mergeFields: node " + allocNode + " in mask is unknown");
				continue;
			}

			AllocSite oldDestAlloc = allocSites.get(allocNode);
			if (oldDestAlloc == null) {
				// toBlock does not know this reference, so copy complete information from fromBlock
				allocSites.put(allocNode, oldAlloc.copy(oldToNewRef));
				changed = true;
			} else {
				// merge info from fromBlock into existing info in toBlock
				changed = oldDestAlloc.update(oldAlloc, strongUpdate, oldToNewRef) || changed;
			}
		}
		return changed;
	}

	private boolean mergeFields(HashMap<FieldID, Reference> fieldRefs, Map<Reference, Reference> oldToNewRef,
			boolean strongUpdate) {
		if (fieldRefs == null || oldToNewRef == null) {
			throw new IllegalArgumentException("mergeFields(fieldRefs==" + fieldRefs + ", oldToNewRef=="
					+ oldToNewRef + "): No null allowed!");
		}

		boolean changed = false;

		// merge fields
		for (Entry<FieldID, Reference> fieldEntry : fieldRefs.entrySet()) {
			FieldID field = fieldEntry.getKey();
			Reference oldRef = fieldEntry.getValue();
			Reference oldDestRef = this.fieldRefs.get(field);

			if (oldDestRef == null) {

				// toBlock does not know this field, so copy complete information from fromBlock
				Reference newRef = (Reference) oldRef.clone();

				// map the source reference to the new clone
				oldToNewRef.put(oldRef, newRef);

				// store new reference
				this.fieldRefs.put(field, newRef);
				changed = true;
			} else {
				// merge info from fromBlock into existing info in toBlock
				oldToNewRef.put(oldRef, oldDestRef);

				// map the source reference to the already existing reference in the target block
				changed = oldDestRef.update(oldRef, strongUpdate) || changed;
			}
		}
		return changed;

	}

	private boolean mergeReferences(HashMap<ContextAndNode, Reference> refs,
			HashMap<Reference, Reference> oldToNewRef, boolean strongUpdate) {
		boolean changed = false;

		for (Entry<ContextAndNode, Reference> refEntry : refs.entrySet()) {
			ContextAndNode refNode = refEntry.getKey();
			Reference oldRef = refEntry.getValue();
			Reference oldDestRef = references.get(refNode);

			if (oldDestRef == null) {
				// toBlock does not know this field, so copy complete information from fromBlock
				Reference newRef = (Reference) oldRef.clone();

				// map the source reference to the new clone
				oldToNewRef.put(oldRef, newRef);

				// store new reference
				references.put(refNode, newRef);
				changed = true;
			} else {
				// merge info from fromBlock into existing info in toBlock

				// map the source reference to the already existing reference in the target block
				oldToNewRef.put(oldRef, oldDestRef);

				changed = oldDestRef.update(oldRef, strongUpdate) || changed;
			}
		}

		return changed;
	}

	public MemorySnapshot deepCopy() {
		MemorySnapshot ret = new MemorySnapshot();

		HashMap<Reference, Reference> referenceMapping = new HashMap<Reference, Reference>();

		for (Entry<ContextAndNode, Reference> entry : references.entrySet()) {

			Reference ref = entry.getValue();
			Reference refCopy = (Reference) ref.clone();
			referenceMapping.put(ref, refCopy);
			ret.references.put(entry.getKey(), refCopy);
		}

		for (Entry<FieldID, Reference> entry : fieldRefs.entrySet()) {

			Reference ref = entry.getValue();
			Reference refCopy = (Reference) ref.clone();
			referenceMapping.put(ref, refCopy);
			ret.fieldRefs.put(entry.getKey(), refCopy);

		}

		for (Entry<ContextAndNode, AllocSite> entry : allocSites.entrySet()) {
			ret.allocSites.put(entry.getKey(), entry.getValue().copy(referenceMapping));
		}

		return ret;
	}

	public Set<Entry<ContextAndNode, Reference>> refEntrySet() {
		return references.entrySet();
	}

	public Set<Entry<ContextAndNode, AllocSite>> allocEntrySet() {
		return allocSites.entrySet();
	}

	public Set<Entry<FieldID, Reference>> fieldEntrySet() {
		return fieldRefs.entrySet();
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("ALLOCS:\n");
		for (Entry<ContextAndNode, AllocSite> entry : allocSites.entrySet()) {
			sb.append(String.format("  %s:\n   %s\n", entry.getKey(), entry.getValue()));
		}
		sb.append("REFS:\n");
		for (Entry<ContextAndNode, Reference> entry : references.entrySet()) {
			sb.append(String.format("  %s:\n   %s\n", entry.getKey(), entry.getValue()));
		}
		sb.append("FIELDS:\n");
		for (Entry<FieldID, Reference> entry : fieldRefs.entrySet()) {
			sb.append(String.format("  %s:\n   %s\n", entry.getKey(), entry.getValue()));
		}
		return sb.toString();
	}

}
