package de.unisb.prog.typestate.ir;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Map.Entry;

import de.unisb.prog.analysis.AnalysisContext;
import de.unisb.prog.analysis.Context;
import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.Block;
import de.unisb.prog.jfirm.ir.Call;
import de.unisb.prog.jfirm.ir.Constant;
import de.unisb.prog.jfirm.ir.MethodGraph;
import de.unisb.prog.jfirm.ir.Mode;
import de.unisb.prog.jfirm.ir.Opcode;
import de.unisb.prog.jfirm.ir.Param;
import de.unisb.prog.jfirm.ir.Type;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.pointsto.DemandPointsTo;
import de.unisb.prog.pointsto.HashPointsToSet;
import de.unisb.prog.pointsto.IDemandPointsToClient;
import de.unisb.prog.pointsto.PointsToSet;
import de.unisb.prog.pointsto.RefinePointsTo;
import de.unisb.prog.typestate.AnalysisLogger;
import de.unisb.prog.typestate.NullState;
import de.unisb.prog.typestate.TopState;
import de.unisb.prog.typestate.ir.BlockMemory.Mem;
import de.unisb.prog.typestate.ir.MemorySnapshot.NodeType;
import de.unisb.prog.typestate.util.NodeHelper;
import de.unisb.prog.typestate.util.ParamHelper;

/**
 * Contains the state information for one methodGraph in one context. It can be built from scratch (using
 * {@link #MethodMemory(MethodGraph)}) at the beginning of the analysis or by performing a context switch from
 * one methodGraph to a called one (using {@link #contextSwitchOut}). A memory which was created by a context
 * switch can be merged back into its &quot;parent memory&quot; by using
 * {@link #contextSwitchBack(MethodMemory)} on the parent with the child as parameter.<br>
 * <br>
 * Context-added node state information is stored as follows in a {@link MethodMemory} object:<br>
 * <br>
 * <code>
 * &nbsp;&nbsp;&nbsp;+--------------&nbsp;METHOD&nbsp;MEMORY&nbsp;-------------+<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;+-------&nbsp;for&nbsp;each&nbsp;BASIC&nbsp;BLOCK&nbsp;-------+&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;+---&nbsp;IN&nbsp;BLOCK&nbsp;MEMORY&nbsp;SNAPSHOT&nbsp;---+&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;CNodes&nbsp;->&nbsp;Allocations&nbsp;sites&nbsp;&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;CNodes&nbsp;->&nbsp;References&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;CNodes&nbsp;->&nbsp;Fields&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;+--------------------------------+&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;+---&nbsp;OUT&nbsp;BLOCK&nbsp;MEMORY&nbsp;SNAPSHOT&nbsp;--+&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;CNodes&nbsp;->&nbsp;Allocations&nbsp;sites&nbsp;&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;CNodes&nbsp;->&nbsp;References&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;CNodes&nbsp;->&nbsp;Fields&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;+--------------------------------+&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;+------------------------------------+&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|<br>
 * &nbsp;&nbsp;&nbsp;+------------------------------------------+<br>
 * </code> <br>
 * An IN memory snapshot for a Block must first be loaded (or newly created) before the state of reference
 * nodes can be read ({@link #loadBlockMem(Block)} or {@link #createInBlock(Block)}). If changes to the state
 * are made ({@link #updateState(ContextAndNode, StateTuple, boolean)},...) this will not affect the original
 * IN memory but changes will be kept in a new local memory snapshot (intermediate memory), which must finally
 * be written back as the OUT memory snapshot of the block <strong>before</strong> another IN memory is loaded
 * ({@link #updateOutBlock()}).<br>
 * <br>
 * Context-switching from a memory to a new one is used to handle methodGraph invocations which are analyzed
 * under the current memory snapshot. A new memory object, containing all information in the current
 * intermediate memory snapshot, will be available in the child memory as IN memory for Block 0. In addition,
 * the called methodGraph's PARAM ValueNodes will be mapped onto the provided parameter Nodes (see
 * {@link #getActualNode(Param)} and {@link #getParameterMapping()}) to enable the analysis to see the state
 * of the parameters. The changed state of the parameters and (if methodGraph is not <code>void</code>) the
 * state of the instance(s) being returned is stored back into the &quot;parent memory&quot; while the state
 * of purely local instances (e.g. local variables which are not visible to the calling context after the
 * called methodGraph returns) is discarded.
 * 
 * @author mikerofone
 * 
 */
public class MethodMemory {

	private static final char			METHODMEM_LOG_TAG	= 'm';

	private static final ContextAndNode	NULL_CONTEXTNODE	= new ContextAndNode(new AnalysisContext(), null);

	/* ----------------------------------------- */
	/* STATIC FIELDS & METHODS */
	/* ----------------------------------------- */

	/**
	 * For convenience, the two values of the {@link Mem} enum are mapped to static fields such that IN and
	 * OUT can be used without always saying "Mem.IN"
	 */
	private static Mem					IN					= Mem.IN, OUT = Mem.OUT;

	private static DemandPointsTo		pointsTo;

	/**
	 * Set the current top-level method graph and specify max context depth for points-to.
	 * 
	 * @param methodGraph
	 * @param maxDepth
	 */
	public static void initMemory(MethodGraph methodGraph, int maxDepth) {

		IDemandPointsToClient client = new IDemandPointsToClient() {

			public boolean answersQuery(ContextAndNode conANode, PointsToSet ptsTo) {
				return false;
			}
		};
		MethodMemory.pointsTo = new RefinePointsTo(methodGraph, 0, null, client, true);
		// MethodMemory.pointsTo = new IntraRegularPointsTo(methodGraph);
		// MethodMemory.globalChanges = new HashMap<ContextAndNode, Integer>();
	}

	/* ----------------------------------------- */
	/* FIELDS */
	/* ----------------------------------------- */

	private MethodGraph					methodGraph;
	// private HashMap<ContextAndNode,Integer> nonLocalChanges;
	private ContextAndNode[]			params;

	private ContextAndNode				returnTargetNode;
	private MutableStateTuple			returnTargetState;

	private HashMap<Block, BlockMemory>	blockMems;

	private BlockMemory					currentMem;
	private MemorySnapshot				currentMemSnapshot;
	private boolean						currentMemSaved;

	private Block						finalMemoryBlock;

	private AnalysisLogger				log;

	private Set<ContextAndNode>			refsBuildInThisMethod;

	/* ----------------------------------------- */
	/* CONSTRUCTORS */
	/* ----------------------------------------- */

	/**
	 * Creates a new MethodMemory for the given {@link MethodGraph}. Use for beginning of analysis:
	 * MethodGraph parameters are not re-mapped to external {@link ValueNode}s and must be initialized
	 * manually to some state (for example, {@link TopState}). The state of the return value (if any) will not
	 * be mapped to an external {@link ValueNode} either.
	 * 
	 * @param methodGraph
	 *            The methodGraph this memory is created for
	 */
	public MethodMemory(MethodGraph methodGraph) {
		if (methodGraph == null) {
			throw new IllegalArgumentException("MethodMemory(methodGraph==" + methodGraph
					+ "): methodGraph might not be null");
		}

		this.methodGraph = methodGraph;
		// nonLocalChanges = new HashMap<ContextAndNode, Integer>();
		currentMem = null;
		currentMemSaved = true;
		currentMemSnapshot = null;
		log = AnalysisLogger.getInstance();
		finalMemoryBlock = methodGraph.getEndBlock();
		params = new ContextAndNode[0];
		returnTargetNode = null;
		blockMems = new HashMap<Block, BlockMemory>();
		refsBuildInThisMethod = new HashSet<ContextAndNode>();

	}

	/* ----------------------------------------- */
	/* GETTERS / SETTERS */
	/* ----------------------------------------- */

	/**
	 * Returns the actual mapping for the PARAM nodes of the methodGraph to (external) {@link ValueNode}s if
	 * this memory was created as a result of a context switch. If this memory was created from scratch, the
	 * return value might be <code>null</code>.
	 * 
	 * @return An array containing the mapping from {@link Param}s to {@link ValueNode}s. See
	 *         {@link #contextSwitchOut(MethodGraph, ValueNode, ValueNode...)} for details.
	 */
	public ContextAndNode[] getParameterMapping() {
		return params;
	}

	/* ----------------------------------------- */
	/* IN/OUT BLOCK STATE MANIPULATION */
	/* ----------------------------------------- */

	/**
	 * Creates an empty IN-memory for a block.
	 * 
	 * @param memBlock
	 *            the Block for which a new entry memory is to be created
	 * @throws IllegalArgumentException
	 *             if memory already exists
	 */
	public void createInBlock(Block memBlock) {
		BlockMemory blockMem = getBlockMem(memBlock);
		if (blockMem.hasMem(IN)) {
			// memory already exists
			throw new IllegalArgumentException("MethodMemory.createMem(memBlock==" + memBlock
					+ "): memory already exists");
		}

		// create empty mem
		blockMem.createMem(IN);

		currentMemSnapshot = blockMem.getDeepMemCopy(IN);
		currentMem = blockMem;
	}

	/**
	 * Retrieves the {@link MemorySnapshot} for the given block and creates it, if first access.
	 * 
	 * @param memBlock
	 * @return
	 */
	private BlockMemory getBlockMem(Block memBlock) {
		BlockMemory blockMem = blockMems.get(memBlock);
		if (blockMem == null) {
			blockMem = new BlockMemory(memBlock);
			blockMems.put(memBlock, blockMem);
		}
		return blockMem;

	}

	/**
	 * Loads the IN-memory for a Block as the current memory.
	 * 
	 * @param memBlock
	 *            the desired Block
	 */
	public void loadBlockMem(Block memBlock) {
		if (currentMem == null || currentMem.getBlock() != memBlock) {
			if (log.trace())
				log.trace("Loading memory for " + memBlock, METHODMEM_LOG_TAG);
			if (!currentMemSaved) {
				throw new IllegalStateException("MethodMemory.loadBlockMem(memBlock==" + memBlock
						+ "): changed memory for " + currentMem + " has not been stored in out-state");
			}

			BlockMemory blockMem = blockMems.get(memBlock);

			// make a deep copy
			currentMemSnapshot = blockMem.getDeepMemCopy(IN);
			currentMem = blockMem;
		}
	}

	/**
	 * Stores the current memory snapshot as the OUT-memory of the current block.
	 */
	public void updateOutBlock() {
		if (log.trace())
			log.trace("updating OUT-memory for " + currentMem, METHODMEM_LOG_TAG);

		if (currentMem == null) {
			throw new IllegalArgumentException("MethodMemory.updateBlock(): no current memory loaded");
		}
		currentMem.setMemory(OUT, currentMemSnapshot);
		currentMemSaved = true;
		printDumpCurrMem();
	}

	/**
	 * Updates (&rarr; set union) the IN-memory of a Block with the OUT-memory of another block.
	 * 
	 * @param toBlock
	 *            the Block for which the IN-memory will be updated
	 * 
	 * @param fromBlock
	 *            the predecessor Block (whose OUT-memory is copied and merged into the <code>toBlock</code>
	 *            IN-memory)
	 * @return whether this changed the entry state of <code>toBlock</code> or not
	 */
	public boolean updateInBlock(Block toBlock, Block fromBlock) {
		if (toBlock == null || fromBlock == null) {
			throw new IllegalArgumentException("MethodMemory.updateBlock(toBlock==" + toBlock
					+ ", fromBlock==" + fromBlock + "): no argument might be null");
		}

		if (log.trace())
			log.trace("updating IN-memory of " + toBlock + " from " + fromBlock, METHODMEM_LOG_TAG);

		// retrieve memory for target block
		BlockMemory toMem = blockMems.get(toBlock);
		if (toMem == null) {
			// First definition of this block
			toMem = new BlockMemory(toBlock);
			blockMems.put(toBlock, toMem);
		}

		boolean changed = false;

		// create the toBlock IN-memory if not already done
		if (!toMem.hasMem(IN)) {
			toMem.createMem(IN);
			changed = true;
		}

		// Merge information from fromBlock out-memory into toBlock in-memory
		if (currentMem.getBlock() != fromBlock) {
			// not cached, use BlockMemory
			BlockMemory fromMem = blockMems.get(fromBlock);
			assert fromMem != null : "blockmemory for " + fromBlock + "==null";

			// check if required OUT memory is available
			if (!fromMem.hasMem(OUT)) {
				// previous memory not defined -> error
				throw new IllegalArgumentException("MethodMemory.update(toBlock==" + toBlock
						+ ", fromBlock==" + fromBlock + "): fromBlock has no OUT state");
			}

			changed = toMem.updateMemory(IN, fromMem, OUT) || changed;
		} else {
			// use from cache
			changed = toMem.updateMemory(IN, currentMemSnapshot) || changed;
		}
		return changed;
	}

	/**
	 * Determines if the current memory snapshot was changed since either the last
	 * {@link #loadBlockMem(Block)} or {@link #updateOutBlock()} operation (which would mean that the current
	 * memory snapshot might differ from all IN and OUT memories and must be saved prior loading a different
	 * one).
	 * 
	 * @return <code>true</code> if it was either saved after the last change or not changed after loading,
	 *         <code>false</code> else
	 */
	public boolean currentOutStateSaved() {
		return currentMemSaved;
	}

	/* ----------------------------------------- */
	/* RETURN VALUE HANDLING */
	/* ----------------------------------------- */

	/**
	 * Gets the ValueNode to which the state of the return value of the methodGraph will be written, if any.
	 * (In interprocedural case)
	 * 
	 * @return The corresponding <code>PROJ DATA</code> ValueNode of the {@link Call} Node from which this
	 *         memory stems. See {@link #contextSwitchOut(MethodGraph, ValueNode, ValueNode...)} for details.
	 */
	public ContextAndNode getReturnNode() {
		return returnTargetNode;
	}

	/**
	 * Adds the given state set to the state of the return value
	 * 
	 * @param newState
	 * @return whether the state changed
	 */
	public boolean updateReturnState(StateTuple newState) {
		if (returnTargetState == null) {
			// First definition of return type
			returnTargetState = new MutableStateTuple(newState.getType());
		}
		return returnTargetState.update(newState, false);
	}

	/**
	 * Get the state of the return value. Can be <code>null</code> if none.
	 * 
	 * @return
	 */
	public StateTuple getReturnState() {
		return returnTargetState;
	}

	/* ----------------------------------------- */
	/* MARKING ESCAPING STATE INFORMATION */
	/* ----------------------------------------- */

	// public void setNonLocal(ContextAndNode node) {
	// Integer currVal = nonLocalChanges.get(node);
	// currVal = currVal == null ? 1 : ++currVal;
	// nonLocalChanges.put(node,currVal);
	// if (log.debug())
	// log.debug("StateTuple for " + node + " escapes to caller (" + currVal + " times)", METHODMEM_LOG_TAG);
	// }
	//	
	// public void removeNonLocal(ContextAndNode node) {
	// Integer currVal = nonLocalChanges.get(node);
	// if (currVal == null) {
	// log.warn("StateTuple for " + node + " to be unmarked as nonlocal, but it already isnt",
	// METHODMEM_LOG_TAG);
	// return;
	// }
	// if (--currVal>0)
	// nonLocalChanges.put(node,currVal);
	// if (log.debug())
	// log.debug("StateTuple for " + node + " escapes to caller (" + currVal + " times)", METHODMEM_LOG_TAG);
	// }
	//
	//
	// public void setGlobal(ContextAndNode node) {
	// Integer currVal = globalChanges.get(node);
	// currVal = currVal == null ? 1 : ++currVal;
	// globalChanges.put(node,currVal);
	// if (log.debug())
	// log.debug("StateTuple for " + node + " escapes globally (" + currVal + " times)", METHODMEM_LOG_TAG);
	// }
	//	
	// public void removeGlobal(ContextAndNode node) {
	// Integer currVal = globalChanges.get(node);
	// if (currVal == null) {
	// log.warn("StateTuple for " + node + " to be unmarked as global, but it already isnt",
	// METHODMEM_LOG_TAG);
	// return;
	// }
	// if (--currVal>0)
	// globalChanges.put(node,currVal);
	// if (log.debug())
	// log.debug("StateTuple for " + node + " escapes globally (" + currVal + " times)", METHODMEM_LOG_TAG);
	// }

	/* ----------------------------------------- */
	/* COMMON STATE HANDLING */
	/* ----------------------------------------- */

	/**
	 * Reads state information stored for the given node, if any. It will first try to find a reference stored
	 * for the given node and return its state. If there is no such reference, there must be an allocation
	 * site stored for its producer, which will then be read. If there is none either, <code>null</code> is
	 * returned.
	 * 
	 * @param node
	 *            the requested node
	 * @return a {@link StateTuple} containing the requested state, <code>null</code> if node is unknown. If
	 *         the given node marks a reference as well as an allocation site, the reference state is
	 *         returned.
	 */
	public StateTuple readState(ContextAndNode node) {
		ValueNode vNode = node.getNode();
		if (vNode.getOpcode() == Opcode.PARAM) {
			// parameters might be remapped
			node = getActualNode(node);
			vNode = node.getNode();
		}
		// should not be necessary as a cast may never be encountered
		// //HACK: Ignore casts
		// else if (vNode.getOpcode() == Opcode.CAST){
		// if (AnalysisLogger.getInstance().debug()) {
		// AnalysisLogger.getInstance().debug("CAST HACK: skipping " + vNode +
		// " while retrieving Typestate information", METHODMEM_LOG_TAG);
		// }
		// return readState(new ContextAndNode (node.getContext(),NodeHelper.getDataNode(vNode)));
		// }

		StateTuple preCheckState = preCheckReadState(node);
		if (preCheckState != null) {
			// some special case CONST, null or whatever occurred
			return preCheckState;
		} else {
			// acutally check for stored state
			if (containsReference(node)) {
				return readReference(node);
			} else {
				if (containsAllocSite(node)) {
					return readAllocSite(node);
				}
			}
		}
		// not found
		return null;
	}

	/**
	 * Performs checks required before looking up allocation sites/references in the data structure. Returns
	 * predefined values for Constants
	 * 
	 * @param node
	 *            the node for which the state is enquired
	 */
	private StateTuple preCheckReadState(ContextAndNode node) {
		ValueNode vNode = node.getNode();
		if (vNode.getOpcode() == Opcode.CONST) {
			if (((Constant) vNode).getValue() == null) {
				Type nullConstType = NodeHelper.getType(vNode);
				MutableStateTuple nullConstState = new MutableStateTuple(nullConstType);
				nullConstState.addState(new NullState());
				return nullConstState;
			} else {
				log.debug("readState(" + node + "): only NULL constants do have state, returning BOTTOM");
				return new StateTuple(NodeHelper.getType(vNode));
			}
		}
		// no special case
		return null;
	}

	/**
	 * as {@link #readState(ContextAndNode)}, but allows reading this information from another block's OUT
	 * set.
	 * 
	 * @param node
	 *            the node whose state to get
	 * @param block
	 *            the block to read from
	 * @return the state or <code>null</code> if nott existing
	 */
	public StateTuple readState(ContextAndNode node, Block block) {
		if (currentMem.getBlock().equals(block)) {
			// read from current memory snapshot. calls covered by this case will happen if phi have self
			// loops
			log.debug("readState(node==" + node + ", block==" + block + ") while currentMem.getBlock()=="
					+ currentMem.getBlock() + ", redirecting to currentMemSnapshot");
			return readState(node);
		}
		ValueNode vNode = node.getNode();
		if (vNode.getOpcode() == Opcode.PARAM) {
			// parameters might be remapped
			node = getActualNode(node);
			vNode = node.getNode();
		}

		StateTuple preCheckState = preCheckReadState(node);
		if (preCheckState != null) {
			// some special case CONST, null or whatever occurred
			return preCheckState;
		} else {
			// actually check for stored state
			BlockMemory blockMem = blockMems.get(block);
			if (blockMem != null) {
				// blockmem exists
				MemorySnapshot blockOutSnap = blockMem.getView(OUT);
				if (blockOutSnap != null) {
					// OUT snapshot exists as well
					// look for refs and then allocs
					if (blockOutSnap.contains(NodeType.REF, node)) {
						return blockOutSnap.getRef(node);
					} else {
						if (blockOutSnap.contains(NodeType.ALLOC, node)) {
							return blockOutSnap.getAlloc(node);
						}
					}
				}
			}
		}
		// not found
		return null;
	}

	/**
	 * Updates state information stored for the given node, or throws an {@link IllegalArgumentException} if
	 * none. Like {@link #readState(ContextAndNode)}, it will update the reference for the given node or, if
	 * no such reference exists, the allocation site stored for its producer. If both are unknown, an
	 * {@link IllegalArgumentException} is thrown.
	 * 
	 * @param node
	 *            the requested node
	 * 
	 * @return whether the operation changed state of the target
	 */
	public boolean updateState(ContextAndNode node, StateTuple withState, boolean strongUpdate) {
		if (node.getNode().getOpcode() == Opcode.PARAM) {
			// parameters might be remapped
			node = getActualNode(node);
		}
		if (node.getNode().getOpcode() == Opcode.CONST) {
			// ignore constants
			return false;
		}

		// check for ref
		if (containsReference(node)) {
			return updateReferenceState(node, withState);
		} else {
			// check for alloc
			ContextAndNode producerNode = new ContextAndNode(node.getContext(), (ValueNode) node.getNode()
					.getProducer());
			if (containsAllocSite(producerNode)) {
				return updateAllocSite(producerNode, withState, strongUpdate, null);
			}
		}

		// not found
		throw new IllegalArgumentException("Cannot update state for unknown node " + node);
	}

	/* ----------------------------------------- */
	/* REFERENCE NODE STATE MANIPULATION */
	/* ----------------------------------------- */

	/** Return whether the current memory snapshot has a reference stored for the given node. */
	private boolean containsReference(ContextAndNode refNode) {
		return currentMemSnapshot.contains(NodeType.REF, refNode);
	}

	/**
	 * Creates a reference for the given ContextAndNode. If first enquiry for this node, a new reference will
	 * be created and equipped with typestate information. It will NOT be registered at its alloc sites! If
	 * reference already exists, it remains untouched and is returned. It will NOT be stored in the current
	 * snapshot!
	 * 
	 * @param refNode
	 * @return
	 */
	public Reference getLinkedReference(ContextAndNode refNode) {
		Reference newRef = currentMemSnapshot.getRef(refNode);
		boolean firstDefinition = newRef == null;

		if (firstDefinition) {
			// first definition, create new
			newRef = new Reference(NodeHelper.getType(refNode.getNode()));
			linkReference(refNode, newRef);
		} else {
			// already defined, just return

		}

		return newRef;
	}

	/**
	 * Adds points-to information to the given reference for all allocation sites pointed at by the given
	 * ContextAndNode.
	 * 
	 * @param refNode
	 * @param ref
	 */
	public void linkReference(ContextAndNode refNode, Reference ref) {
		// Find all reachable allocation Sites and register reference as user
		PointsToSet ptSet = getAllocSites(refNode);
		ref.setPtsTop(ptSet.isTop());
		ref.setPtsFromArray(ptSet.isFromArray());
		for (ContextAndNode allocNode : ptSet) {
			ValueNode node = NodeHelper.getDataNode(allocNode.getNode());
			if (node == null || node.getOpcode() == Opcode.CONST || node.getMode() != Mode.A) {
				// consts or non-reference nodes are skipped
				continue;
			}
			ref.addPointsToTarget(allocNode);
		}
	}

	/**
	 * Register the given reference at all its allocation sites.
	 * 
	 * @param ref
	 *            the ref to register
	 * @param mergeAllocStateInRef
	 *            if <code>true</code>, also weak-updates the alloc's states from the ref
	 * @return <code>true</code> if it changed something
	 */
	private boolean registerRefAtAllocs(Reference ref, boolean mergeAllocStateInRef) {

		boolean mustRegisterAsUser = !ref.isRegisteredAtAllocs();
		// if already registered and no state merge required, nothing to do
		if (!mustRegisterAsUser && !mergeAllocStateInRef) {
			return false;
		}

		boolean changed = false;

		// get pointsto info from reference
		Set<ContextAndNode> pts = ref.getPointsToTargets();

		boolean registeredToAll = true;
		for (ContextAndNode allocNode : pts) {
			AllocSite allocSite = currentMemSnapshot.getAlloc(allocNode);
			if (allocSite == null) {
				if (log.debug())
					log.debug("PointsToSet for " + ref + " contains unknown allocation site " + allocNode
							+ ", crossing a domination border here?", METHODMEM_LOG_TAG);
				continue;
			}
			if (mustRegisterAsUser) {
				if (allocSite != null) {
					changed = allocSite.addUser(ref) || changed;
				} else {
					registeredToAll = false;
				}
			}
			// add state of the allocation site to state of the reference
			if (mergeAllocStateInRef) {
				changed = ref.update(allocSite, false) || changed;
			}
		}
		ref.setRegisteredAtAllocs(registeredToAll);
		return changed;
	}

	// private void unregisterRefAtAllocs(Reference ref) {
	// List<ContextAndNode> pts = ref.getPointsToTargets();
	// ref.setRegisteredAtAllocs(pts.size() == 0);
	// for (ContextAndNode allocNode : pts) {
	// AllocSite allocSite = currentMemSnapshot.getAlloc(allocNode);
	// allocSite.removeUser(ref);
	// }
	// }

	// private void clearReference(Reference ref) {
	// // unregister
	// unregisterRefAtAllocs(ref);
	// // remove pointsto info
	// ref.clearPointsTo();
	// // remove typestate info
	// ref.clearTypestate();
	// }

	/* ----------------------------------------- */
	/* METHOD-MEMORY MANIPULATION */
	/* ----------------------------------------- */

	/**
	 * Creates a new reference for the given node, links it to every contributing allocation site and merges
	 * all alloc sites' states together as state for the new reference, if no state is given as second
	 * parameter (i.e. if its <code>null</code>). The new reference is returned. If a reference already
	 * exists, it will be reused: It will be linked again to it's allocation sites.<br>
	 * 
	 * @return the new {@link Reference} or <code>null</code>, if the given node is not valid (CONST)
	 */
	public Reference makeReference(ContextAndNode refNode, StateTuple withState) {

		ValueNode dataNode = NodeHelper.getDataNode(refNode.getNode());
		if (dataNode.getOpcode() == Opcode.CONST) {
			if (log.debug())
				log.debug("makeReference on " + refNode + " ignored because it's a CONSTant!",
						METHODMEM_LOG_TAG);
			return null;
		}
		if (dataNode.getMode() != Mode.A) {
			if (log.debug())
				log.debug("makeReference on " + refNode + "ignored because it's not Mode.A!",
						METHODMEM_LOG_TAG);
			return null;
		}

		// Create/Reset reference with pointsto information
		boolean makeNewRef = !currentMemSnapshot.contains(NodeType.REF, refNode);
		Reference newRef = getLinkedReference(refNode);

		if (!makeNewRef) {
			// Clear old information
			newRef.clearTypestate();
		} else {
			currentMemSnapshot.putRef(refNode, newRef);
			refsBuildInThisMethod.add(refNode);
		}

		// Link it to all allocs and merge their state information in the ref
		registerRefAtAllocs(newRef, withState == null);

		if (withState != null) {
			newRef.update(withState, true);
		}

		return newRef;
	}

	/** Stores the given reference for the given node, and links it to every contributing allocation site. */
	public void putAsReference(ContextAndNode refNode, Reference ref) {

		ValueNode dataNode = NodeHelper.getDataNode(refNode.getNode());
		if (dataNode.getOpcode() == Opcode.CONST) {
			if (log.debug())
				log.debug("putAsReference on " + refNode + " with " + ref
						+ "ignored because it's a CONSTant!", METHODMEM_LOG_TAG);
			return;
		}
		if (dataNode.getMode() != Mode.A) {
			if (log.debug())
				log.debug("putAsReference on " + refNode + " with " + ref
						+ "ignored because it's not Mode.A!", METHODMEM_LOG_TAG);
			return;
		}

		// Put ref into memory and link it
		Reference currRef = currentMemSnapshot.getRef(refNode);
		if (currRef == null) {
			linkReference(refNode, ref);
			currentMemSnapshot.putRef(refNode, ref);
			refsBuildInThisMethod.add(refNode);
			currRef = ref;
		} else {
			// update state
			currRef.overwrite(ref);
		}

		// must also be done for existing references to handle dominance borders
		registerRefAtAllocs(ref, false);

	}

	/**
	 * (Hard-)updates the state of all allocation sites referenced by the given ValueNode in the current
	 * memory snapshot
	 * 
	 * @param refNode
	 *            the desired ValueNode (must have <code>Mode.A</code>)
	 * @param newState
	 *            the new state to store, overwriting or updating the old one
	 */
	private boolean updateReferenceState(ContextAndNode refNode, StateTuple newState) {
		assert currentMemSnapshot != null : "Must first load a block memory before accessing local memory";
		if (refNode.getNode().getOpcode() == Opcode.CONST) {
			log.warn("updateReferenceState on " + refNode + " with " + newState
					+ "ignored because it's a CONSTant!", METHODMEM_LOG_TAG);
			return false;
		}

		boolean changed = false;
		// The propagation of the state change to all of the affected references might only encounter each
		// reference once.
		HashSet<Reference> seen = new HashSet<Reference>();

		// First update the state of the given node itself
		Reference refState = currentMemSnapshot.getRef(refNode);
		if (refState == null) {
			throw new IllegalArgumentException("updateReferenceState for node " + refNode
					+ " but no ref stored for that node");
		}
		// this node has an own state, always perform a strong update on it
		changed = refState.update(newState, true);
		seen.add(refState);

		// then perform an update on all possible points-to targets of this node
		boolean strongUpdate = true;
		// Determine if a strong update can be performed
		// TOP always performs strong updates
		// TOP pointsto, references to arrays or not single concrete state information leads to weak updates
		strongUpdate = newState.containsTop()
				|| (!refState.isPtsTop() && !refState.isPtsFromArray() && refState.isPtsSingleConcrete());

		if (refState.isPtsTop()) {
			// FIXME: ignored for now, but must be considered when pointsToTargets information becomes better
			// and
			// types are available
			if (log.debug())
				log
						.debug("updateReferenceState: points to information for " + refNode
								+ " returned TOP " + refState.getPointsToTargets()
								+ " TODO: Implement TOP handling!", METHODMEM_LOG_TAG);
		}
		if (refState.getPointsToTargets().size() == 0) {
			// If no allocsites are given, assume that every allocation site with compatible type (NOT
			// IMPLEMENTED YET) will be updated
			if (log.debug())
				log.debug("updateReferenceState: PointsToSet for '" + refNode
						+ "' returned zero targets. TODO: Implement TOP handling!", METHODMEM_LOG_TAG);

		}

		// if this is a getfield reference, update the fields explicitly
		if (refState instanceof FieldReference) {
			// perform strongupdate on the fields, if required
			FieldReference fieldRefState = (FieldReference) refState;
			Set<FieldID> fields = fieldRefState.getFields();
			if (fields.size() == 1) {
				for (FieldID field : fields) {
					Reference fieldRef = getFieldRef(field);
					fieldRef.update(newState, true);

					// avoid more (weak) updates on those refs
					seen.add(fieldRef);
				}
			}
		}

		// check if reference is correctly linked
		changed = registerRefAtAllocs(refState, false) || changed;

		// perform update on the alloc sites registered with the reference
		for (ContextAndNode allocNode : refState.getPointsToTargets()) {
			AllocSite alloc = currentMemSnapshot.getAlloc(allocNode);
			if (alloc == null) {
				log.debug("updateReferenceState: Skipping points to target " + allocNode
						+ " because unknown to memory (dominance border here?)", METHODMEM_LOG_TAG);
				continue;
			}
			changed = alloc.update(newState, strongUpdate) || changed;
			// propagate to user refs
			changed = alloc.propagate(newState, strongUpdate, seen) || changed;
		}
		return changed;
	}

	/**
	 * Gets the state saved for this reference. It does not consider the contributing allocation sites, use
	 * {@link #getAllocSites(ContextAndNode)} and {@link #readAllocSite(ContextAndNode)} to access their
	 * state.
	 * 
	 * @param refNode
	 *            the desired reference {@link ContextAndNode}
	 * @return the {@link Reference} stored for this node
	 */
	private Reference readReference(ContextAndNode refNode) {
		assert currentMemSnapshot != null : "Must first load a block memory before accessing local memory";

		// retrieve the state of the given reference node
		Reference state = currentMemSnapshot.getRef(refNode);

		if (state == null) {
			throw new IllegalArgumentException("The given reference " + refNode
					+ " is unknown in the current memory snapshot");
		}

		return state;
	}

	/* ----------------------------------------- */
	/* ALLOCATION SITE MANIPULATION */
	/* ----------------------------------------- */

	/** Return whether the current memory snapshot has an allocation site stored for the given node. */
	private boolean containsAllocSite(ContextAndNode refNode) {
		return currentMemSnapshot.contains(NodeType.ALLOC, refNode);
		// return currentMemSnapshot.contains(NodeType.ALLOC, getDataNode(refNode));
	}

	public void createAllocSite(ContextAndNode allocNode, StateTuple newState) {
		assert currentMemSnapshot != null : "Must first load a block memory before accessing local memory";

		ValueNode firmNode = allocNode.getNode();

		switch (firmNode.getOpcode()) {
		case CONST:
			// constants do not have any state (or they can at least not change it)
			// so just ignore this request
			return;
		default:
			// if (firmNode instanceof EnumTupleNode<?>) {
			// allocNode = getDataNode(allocNode);
			// }
			// if existent, find the previous version of the allocSite
			AllocSite existingAlloc = currentMemSnapshot.getAlloc(allocNode);
			if (existingAlloc == null) {
				// not existent, so create new
				AllocSite newAlloc = new AllocSite(allocNode, NodeHelper.getType(allocNode.getNode()));
				newAlloc.overwrite(newState);
				currentMemSnapshot.putAlloc(allocNode, newAlloc);
				break;
			} else {
				// old version of alloc site exists,

				// The disconnection step should only be done if
				// - Pathsensitive analysis is desired,
				// - Better aliasing handling (not through data in alloc site) is implemented
				// Otherwise, some references are disconnected from each other and aliasing is broken
				// // so cleanly disconnect it from any users and write newstate
				// // --> remove all pointsToTargets information (nothing to do here)
				//
				// // --> "unregister" all users
				// existingAlloc.disconnect();

				// Write new empty state
				existingAlloc.update(newState, true);
				break;
			}
		}
		currentMemSaved = false;

	}

	private AllocSite readAllocSite(ContextAndNode allocNode) {

		AllocSite state;
		ValueNode allocValueNode = allocNode.getNode();
		switch (allocValueNode.getOpcode()) {
		case CONST:

			// if refNode is a constant, return the empty state set (constants don't have any state or at
			// least can never change it)
			state = new AllocSite(allocNode, NodeHelper.getType(allocValueNode));
			// Check if this might be actually a reference being <null>
			if (((Constant) allocValueNode).getValue() == null) {
				// add a meta "null" state
				state.addState(new NullState());
			}
			break;
		default:
			// state = currentMemSnapshot.getAlloc(getDataNode(allocNode));
			state = currentMemSnapshot.getAlloc(allocNode);
			if (state == null) {
				throw new IllegalArgumentException("MethodMemory.readAllocSite(allocNode==" + allocNode
						+ "): allocNode unknown to current memory (" + currentMemSnapshot + ")");
			}
			break;
		}
		return state;
	}

	private boolean updateAllocSite(ContextAndNode siteNode, StateTuple withState, boolean strongUpd,
			HashSet<Reference> seen) {
		assert currentMemSnapshot != null : "Must first load a block memory before accessing local memory";

		boolean changed = false;
		switch (siteNode.getNode().getOpcode()) {
		case CONST:
			// constants do not have any state (or they can at least not change it)
			// so just ignore this request
			break;
		default:
			// AllocSite alloc = currentMemSnapshot.getAlloc(getDataNode(siteNode));
			AllocSite alloc = currentMemSnapshot.getAlloc(siteNode);
			assert alloc != null : "updateAllocSite on target '" + siteNode + "' was mapped to '" + alloc
					+ "' which is unknown";

			// Perform an update on the AllocSite itself
			changed = alloc.update(withState, strongUpd);

			// Does this stateset only consist of the start state?
			if (withState.stateCount() == 1 && withState.iterator().next().getName().equals("start")) {
				// Yes, so do not forward to references
				if (log.debug())
					log.debug("Skipped propagation of start state for alloc site " + alloc + " stored for "
							+ siteNode, METHODMEM_LOG_TAG);
				break;
			}
			// Propagate the change to all of its users
			changed = alloc.propagate(withState, strongUpd, seen) || changed;
			break;
		}
		currentMemSaved = false;

		return changed;
	}

	// public void updateAllocSiteState(ContextAndNode allocSite, HashSet<IState> newState, boolean strongUpd)
	// {
	// assert currentMemSnapshot != null : "Must first load a block memory before accessing local memory";
	//
	// switch (allocSite.getNode().getOpcode()) {
	// case CONST:
	// // constants do not have any state (or they can at least not change it)
	// // so just ignore this request
	// return;
	// default:
	// AllocSite oldState = currentMemSnapshot.getAlloc(allocSite);
	// if (oldState == null) {
	// throw new IllegalStateException("Updating allocation Site " + allocSite
	// + " with new state + " + newState + "failed: Must be created first");
	// }
	//
	// // Perform (strong or weak) update
	// oldState.update(newState, strongUpd);
	// break;
	// }
	// currentMemSaved = false;
	//
	// }

	public PointsToSet getAllocSites(ContextAndNode refNode) {
		// look up all possible allocation sites for refNode

		ValueNode node = refNode.getNode();
		// consider remapping of parameters
		if (node instanceof Param) {
			refNode = getActualNode(refNode);
			node = refNode.getNode();
		}

		// determine the type of node to find out what pointsto information is needed
		switch (node.getOpcode()) {
		case CONST:

			// if it's a null-const it must be returned as a possible target
			// if not just return an empty set
			if (((Constant) refNode.getNode()).getValue() == null) {
				return new HashPointsToSet(refNode);
			} else {
				// default constant, return empty pointstoset
				return EmptyPointsToSet.getInstance();
			}

		case NEW:

			// the alloc site is the only target
			return new HashPointsToSet(refNode);

		default:
			// use pointsto implementation to get the allocation sites

			// find the DATA element of the given node
			refNode = NodeHelper.getDataNode(refNode);

			try {
				PointsToSet pts = pointsTo.pointsTo(refNode);
				// HACK: For some reason pointsTo can return "null" ContextAndNodes
				try {
					pts.remove(NULL_CONTEXTNODE);
				} catch (UnsupportedOperationException e) {
					// immutable set, create mutable copy if necessary
					if (pts.contains(NULL_CONTEXTNODE)) {
						ModifiablePointsToSet newPts = new ModifiablePointsToSet();
						newPts.addAll(pts);
						newPts.remove(NULL_CONTEXTNODE);
						pts = newPts;
					}
				}
				if (pts.size() == 0) {
					log.debug("PointsTo set returned for " + refNode + " was empty!", METHODMEM_LOG_TAG);
				}
				return pts;
			} catch (IllegalArgumentException e) {
				throw e;
			} catch (Exception e) {
				log.error("getAllocSites(" + refNode + "): Exception in staj-pointsto: " + e
						+ "\nAssume no targets being returned.", METHODMEM_LOG_TAG);
				return EmptyPointsToSet.getInstance();
			}
		}
	}

	/* ----------------------------------------- */
	/* FIELD HANDLING */
	/* ----------------------------------------- */

	/**
	 * Gets the Reference currently assigned to the given field, or <code>null</code> if not yet defined.
	 * 
	 * @param field
	 *            The field whose reference is requested
	 * @return the reference or <code>null</code>, if not defined yet
	 */
	public Reference getFieldRef(FieldID field) {
		if (field == null) {
			throw new IllegalArgumentException("getFieldState(field==null): no null allowed");
		}
		if (currentMemSnapshot == null) {
			throw new IllegalArgumentException("getFieldState(field==" + field
					+ "): no memory snapshot loaded");
		}

		return currentMemSnapshot.getFieldRef(field);
	}

	/**
	 * Assigns the given reference to the given field. If the field was previously assigned another reference,
	 * the previous one will remain unchanged, such that all nodes still using it won't see a difference.
	 * 
	 * @param field
	 *            The field to create or update
	 * @param fieldRef
	 *            The reference to set
	 * @param strongUpdate
	 */
	public Reference setFieldRef(FieldID field, StateTuple fieldState, ContextAndNode fromNode,
			boolean strongUpdate) {
		if (field == null || fieldState == null) {
			throw new IllegalArgumentException("setFieldRef(field==" + field + ",fieldState==" + fieldState
					+ "fromNode==" + fromNode + "): No null allowed for first two args");
		}

		Reference curFieldRef = currentMemSnapshot.getFieldRef(field);
		if (curFieldRef == null) {
			// build a new reference using the given information

			// if TOP field, there is no "fromNode"
			if (fromNode != null) {
				curFieldRef = getLinkedReference(fromNode); // fieldRef;
			} else {
				curFieldRef = new Reference(field.getField().getType());
			}
			curFieldRef.update(fieldState, true);
			currentMemSnapshot.putFieldRef(field, curFieldRef);
		} else {
			curFieldRef.update(fieldState, strongUpdate);
		}

		// TOP fields do not belong to any object, such that fromNode can be null
		if (fromNode != null) {
			// link to allocsites
			registerRefAtAllocs(curFieldRef, false);
		}

		return curFieldRef;
	}

	/* ----------------------------------------- */
	/* METHOD-MEMORY MANIPULATION */
	/* ----------------------------------------- */

	/**
	 * Generates a new MethodMemory object for a new methodGraph, which will contain the memory snapshot valid
	 * at the moment of the invocation of this methodGraph. This is used to perform a methodGraph context
	 * switch when another methodGraph is called.
	 * 
	 * @param methodGraph
	 *            the jFirm methodGraph for which this memory is created
	 * @param returnTargetNode
	 *            the ValueNode to which the return Value will be written
	 * @param params
	 *            the ordered parameters for this methodGraph
	 * @param isStatic
	 * @return the new MethodMemory object
	 */
	public MethodMemory contextSwitchOut(MethodGraph methodGraph, Context toContext,
			ContextAndNode returnNode, ContextAndNode... params) {
		MethodMemory newMem = new MethodMemory(methodGraph);
		if (methodGraph == null || params == null) {
			throw new IllegalArgumentException("contextSwitchOut(methodGraph==" + methodGraph + ", params=="
					+ params + "): nothing might be null");
		}
		if (currentMemSnapshot == null) {
			throw new IllegalStateException(
					"contextSwitchOut(methodGraph=="
							+ methodGraph
							+ ", params=="
							+ params
							+ "): cannot switch context from a memory which has no current memory loaded (this.currentMem == null)");
		}

		// no need to deep copy currentMemSnapshot as the IN memory of the start block will not be modified
		Block startBlock = methodGraph.getStartBlock();
		BlockMemory startBlockMem = new BlockMemory(startBlock);
		newMem.blockMems.put(startBlock, startBlockMem);
		// startBlockMem.createMem(IN);
		startBlockMem.setMemory(IN, currentMemSnapshot.deepCopy());// updateMemory(IN, currentMemSnapshot);
		newMem.loadBlockMem(startBlock);
		newMem.params = params;

		// (will finally be resolved by pointsto - only mark the parameters as escaping then)
		// changes to the parameters must be returned upon methodGraph completion
		// ValueNode[] paramNodes = methodGraph.getParameters();
		for (int i = 0, n = params.length; i < n; i++) {
			// not every param is remapped, only process those which are
			if (params[i] != null) {
				// resolve potential remapping of the parameters
				if (params[i].getNode() instanceof Param) {
					params[i] = getActualNode(params[i]);
				}

				// SHOULD NOT BE NECESSARY IF CORRECTLY RE-LINKED
				/*
				 * ContextAndNode paramNode = new ContextAndNode(toContext, paramNodes[i]);
				 * 
				 * // Create a new reference linking this param node to the real argument Reference paramRef =
				 * new Reference(); StateTuple paramState = newMem.readState(params[i]); if (paramState ==
				 * null) { log.error("contextSwitchOut: State for parameter " + paramNode + " mapped to " +
				 * params[i] + " unknown, not remapping"); continue; } paramRef.overwrite(paramState);
				 * newMem.putAsReference(paramNode, paramRef);
				 */

				// // mark all affected allocation sites (as they live in the calling context)
				// PointsToSet paramTargets = getAllocSites(params[i]);
				// for (ContextAndNode paramTarget : paramTargets) {
				// setNonLocal(paramTarget);
				// }
				// newMem.addPointsToTargets(paramNode, paramTargets);
			}
		}

		// handle the return value of the called methodGraph - if type is a reference (and thus not void, too)
		if (returnNode != null && NodeHelper.getDataNode(returnNode.getNode()).getMode() == Mode.A) {
			// tell the memory where the result must be returned to
			newMem.returnTargetNode = returnNode;
		}

		// log.trace("PARAM mapping: " + Arrays.toString(params), METHODMEM_LOG_TAG);

		return newMem;
	}

	/**
	 * After a context switch (see {@link #contextSwitchOut(MethodGraph, ValueNode, ValueNode...)}) to another
	 * methodGraph, the state of the parameters and the return-target ValueNode (see {@link #getReturnNode()})
	 * will be copied back from the IN-memory of the final <code>Block</code> (see
	 * {@link #markReturnBlock(Block)}) from the given child {@link MethodMemory}.
	 * 
	 * @param childMem
	 *            the <code>MethodMemory</code> instance which was created by calling
	 *            {@link #contextSwitchOut(MethodGraph, ValueNode, ValueNode...)} on this methodGraph memory.
	 * @param childContext
	 */
	public void contextSwitchBack(MethodMemory childMem, Context childContext) {
		if (childMem == null) {
			throw new IllegalArgumentException("contextSwitchBack(childMem==" + childMem
					+ "): argument may not be null");
		}
		if (currentMemSnapshot == null) {
			throw new IllegalStateException("contextSwitchBack(childMem==" + childMem
					+ "): cannot context switch back to a methodGraph memory with currentMemSnapshot == null");
		}

		// Get the final memory state of the childMem
		assert childMem.finalMemoryBlock != null : "MethodMemory.contextSwitchBack(childMem==" + childMem
				+ "): childMem.finalMemoryBlock==null";
		BlockMemory finalChildMem = childMem.blockMems.get(childMem.finalMemoryBlock);
		MemorySnapshot finalChildSnapshot = finalChildMem.getView(IN);

		// just take the last snapshot and replace the current one with it
		currentMemSnapshot = finalChildSnapshot;
		// remove all references which have been created in the submethod, as they never can be used again
		for (ContextAndNode refEntry : childMem.refsBuildInThisMethod) {
			Reference refToKill = currentMemSnapshot.getRef(refEntry);
			refToKill.disconnect();
			currentMemSnapshot.killRef(refEntry);
		}
		// Nodes mapped as parameters could have received users as well, which need to be removed
		for (ContextAndNode param : childMem.params) {
			AllocSite alloc = currentMemSnapshot.getAlloc(param);
			if (alloc != null) {
				alloc.removeUsersNotIn(currentMemSnapshot);
			}
		}

		// // copy any globally visible information (parameters, assignments to fields etc) back
		// if (log.trace())
		// log.trace("Merging " + childMem + " into " + this + ", changes from childMem: "
		// + childMem.nonLocalChanges, METHODMEM_LOG_TAG);
		//
		// // All allocation sites that are passed from the child methodGraph must be stored in caller, as
		// well as ALL
		// // field information. References are discarded because they are methodGraph-local.
		// Set<ContextAndNode> returningNodes = new
		// HashSet<ContextAndNode>(childMem.nonLocalChanges.keySet());
		// returningNodes.addAll(MethodMemory.globalChanges.keySet());
		// currentMemSnapshot.mergeOnlyFieldsAndAllocs(finalChildSnapshot, returningNodes);

		// Create a reference node for the return value
		if (childMem.returnTargetNode != null) {
			Reference returnReference = new Reference(NodeHelper.getType(childMem.returnTargetNode.getNode()));

			// //Read state of the all the contributing returns (returns are successors of the end block)
			// for (int i = 0, n = childMem.finalMemoryBlock.getN(); i < n; i++) {
			//				
			// ValueNode returnNode = childMem.finalMemoryBlock.getIn(i);
			// ContextAndNode returnValue = new ContextAndNode(childContext,
			// NodeHelper.getInput(returnNode,1));
			//				
			// //merge the states -> perform weak updates
			// returnReference.update(childMem.readState(returnValue), false);
			// }

			// Get state of the return value
			StateTuple returnState = childMem.getReturnState();

			assert returnState != null : "Return value expected for " + childMem.returnTargetNode
					+ " but was unknown in child memory.";
			// Set reference to state of the return value
			returnReference.overwrite(returnState);

			// Store the reference with the merged state in the memory
			putAsReference(childMem.returnTargetNode, returnReference);
		}

		// for (ContextAndNode escapingNode : MethodMemory.globalChanges) {
		// NodeType type = finalChildSnapshot.getType(escapingNode);
		// // The returning methodGraph must contain information about the nonlocal changes
		// assert type != null;
		// // copy state from called methodGraph (hard update)
		// switch (type) {
		// case ALLOC:
		// currentMemSnapshot.putAlloc(escapingNode, finalChildSnapshot.getAlloc(escapingNode));
		// break;
		// case REF:
		// currentMemSnapshot.putRef(escapingNode, finalChildSnapshot.getRef(escapingNode));
		// break;
		// }
		// }
		//
		// // Link the return-target reference node to the correct allocation sites
		// if (childMem.returnNode != null) {
		// pointsToSets.put(childMem.returnNode, new PointsToModifiable(childMem
		// .getAllocSites(childMem.returnNode)));
		// }

		printDumpCurrMem();
	}

	/* ----------------------------------------- */
	/* INTERNAL HELPER METHODS */
	/* ----------------------------------------- */

	/**
	 * Routes request for parameters to the actual ValueNode that is passed as parameter. If no remapping is
	 * given, it will return the argument.
	 * 
	 * @param paramNode
	 *            The {@link Param} node whose mapping shall be resolved.
	 * @return the mapped (external) {@link ValueNode}, if specified, and <code>paramNode</code>, if not.
	 */
	public ContextAndNode getActualNode(ContextAndNode paramNode) {
		if (paramNode.getNode() instanceof Param) {
			int pIndex = ParamHelper.getIndex((Param) paramNode.getNode(), methodGraph.getParameters());
			if (pIndex < params.length && pIndex > -1 && params[pIndex] != null) {
				if (log.debug())
					log.debug("  " + paramNode + " was mapped to " + params[pIndex], METHODMEM_LOG_TAG);
				return params[pIndex];
			}
			// if not found: fall through
		} else {
			// passed node is no PARAM
		}
		return paramNode;
	}

	/* ----------------------------------------- */
	/* DEBUGGING METHODS */
	/* ----------------------------------------- */

	public void printDumpCurrMem() {

		if (log.trace()) {
			log.trace("--------CURRMEM (" + methodGraph.getName() + ") " + currentMem + " ------------",
					METHODMEM_LOG_TAG);
			log.trace("   ALLOCS", METHODMEM_LOG_TAG);
			for (Entry<ContextAndNode, AllocSite> entry : currentMemSnapshot.allocEntrySet()) {
				log
						.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
								METHODMEM_LOG_TAG);
			}
			log.trace("   REFS", METHODMEM_LOG_TAG);
			for (Entry<ContextAndNode, Reference> entry : currentMemSnapshot.refEntrySet()) {
				log
						.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
								METHODMEM_LOG_TAG);
			}
			log.trace(" FIELDS", METHODMEM_LOG_TAG);
			for (Entry<FieldID, Reference> entry : currentMemSnapshot.fieldEntrySet()) {
				log
						.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
								METHODMEM_LOG_TAG);
			}
			if (returnTargetState != null) {
				log.trace("   RETURN VALUE STATE:", METHODMEM_LOG_TAG);
				log.trace(String.format("    %-40s: %s", returnTargetNode, returnTargetState));
			}
		}
	}

	public void printDumpFinalMem() {
		if (log.trace()) {
			log.trace("--------FINAL MEM (" + methodGraph.getName() + ")------------", METHODMEM_LOG_TAG);
			try {
				log.trace("   ALLOCS", METHODMEM_LOG_TAG);
				for (Entry<ContextAndNode, AllocSite> entry : blockMems.get(methodGraph.getEndBlock())
						.getView(IN).allocEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}
				log.trace("   REFS", METHODMEM_LOG_TAG);
				for (Entry<ContextAndNode, Reference> entry : blockMems.get(methodGraph.getEndBlock())
						.getView(IN).refEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}
				log.trace(" FIELDS", METHODMEM_LOG_TAG);
				for (Entry<FieldID, Reference> entry : blockMems.get(methodGraph.getEndBlock()).getView(IN)
						.fieldEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}
				if (returnTargetState != null) {
					log.trace("   RETURN VALUE STATE:", METHODMEM_LOG_TAG);
					log.trace(String.format("    %-40s: %s", returnTargetNode, returnTargetState),
							METHODMEM_LOG_TAG);
				}
			} catch (NullPointerException e) {
				log.trace("  -- End Block Memory not found!! --", METHODMEM_LOG_TAG);
			}
		}
	}

	public void printDump() {
		if (log.trace()) {
			log.trace("-----MEMDUMP(" + methodGraph.getName() + "):-----", METHODMEM_LOG_TAG);
			log.trace(" ENTRY STATES:", METHODMEM_LOG_TAG);
			for (BlockMemory blockMem : blockMems.values()) {
				log.trace("   ALLOCS", METHODMEM_LOG_TAG);
				for (Entry<ContextAndNode, AllocSite> entry : blockMem.getView(IN).allocEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}
				log.trace("   REFS", METHODMEM_LOG_TAG);
				for (Entry<ContextAndNode, Reference> entry : blockMem.getView(IN).refEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}
				log.trace(" FIELDS", METHODMEM_LOG_TAG);
				for (Entry<FieldID, Reference> entry : blockMem.getView(IN).fieldEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}

			}
			log.trace(" OUT STATES:", METHODMEM_LOG_TAG);
			for (BlockMemory blockMem : blockMems.values()) {
				log.trace("   ALLOCS", METHODMEM_LOG_TAG);
				for (Entry<ContextAndNode, AllocSite> entry : blockMem.getView(OUT).allocEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}
				log.trace("   REFS", METHODMEM_LOG_TAG);
				for (Entry<ContextAndNode, Reference> entry : blockMem.getView(OUT).refEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}
				log.trace(" FIELDS", METHODMEM_LOG_TAG);
				for (Entry<FieldID, Reference> entry : blockMem.getView(OUT).fieldEntrySet()) {
					log.trace(String.format("    %-40s: %s", entry.getKey(), entry.getValue()),
							METHODMEM_LOG_TAG);
				}

			}
			log.trace(" CURRENT MEMORY:", METHODMEM_LOG_TAG);
			log.trace(String.format("  CUR %9s: %s", currentMem.toString(), currentMemSnapshot.toString()),
					METHODMEM_LOG_TAG);
			log.trace("END MEMDUMP", METHODMEM_LOG_TAG);
		}
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("MethodMemory (" + methodGraph.getName() + ")");
		// sb.append("\n ENTRY STATES: " + inStates + "\n");
		// sb.append(" OUT STATES: " + outStates + "\n");
		// sb.append(" CURRENT MEMORY: " + currentMem + " " + currentMemSnapshot + "\n");
		return sb.toString();
	}

}
