package de.unisb.prog.typestate.ir;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.Type;
import de.unisb.prog.typestate.IState;

public class Reference extends MutableStateTuple implements Cloneable {

	private static int		counter	= 1;
	private int				id;
	HashSet<ContextAndNode>	ptsTargets;
	HashSet<AllocSite>		registeredAt;
	boolean					ptsTop, ptsFromArray;
	private boolean			isRegistered;

	// protected int numSites = 0;

	/** Create new empty reference node tuple */
	public Reference(Type type) {
		super(type);
		this.id = counter++;
		this.ptsTargets = new HashSet<ContextAndNode>();
		this.registeredAt = new HashSet<AllocSite>();
	}

	/** Copy Constructor */
	public Reference(Reference copyFrom) {
		super(copyFrom);
		this.id = counter++;
		// this.numSites = copyFrom.numSites;
		this.ptsTargets = new HashSet<ContextAndNode>(copyFrom.ptsTargets);
		this.ptsFromArray = copyFrom.ptsFromArray;
		this.ptsTop = copyFrom.ptsTop;
		
		//DO NOT COPY REGISTERED AT because the references might be invalid. the new alloc sites will register themselves
		this.registeredAt = new HashSet<AllocSite>();
	}

	public boolean receiveUpdate(StateTuple newState, boolean strongUpdateOnAlloc) {
		// TODO some more sophisticated checking could enable strong updates on the reference nodes

		// perform a strong update upon reception if
		// 1. there was a strong update on the propagating allocation site
		// 2. the user only depends on one exactly this single allocation site -> COSTLY?

		// -> 2. if the reference depends on only one target, it must be the update-propagating alloc site, so
		// a strong update can be performed

		// boolean strongUpd = strongUpdateOnAlloc && numSites < 2;

		boolean strongUpd = strongUpdateOnAlloc && isPtsSingleConcrete();
		return update(newState, strongUpd);
	}

	public int getId() {
		return id;
	}

	public boolean update(Reference fromRef, boolean strongUpdate) {
		boolean changed = false;
		if (strongUpdate) {
			changed = ptsTargets.size() > 0 || changed;
			disconnect();
			clearPointsTo();
		}

		// add the new pointsto targets
		boolean addedPtsTargets = this.ptsTargets.addAll(fromRef.ptsTargets);
		isRegistered &= !addedPtsTargets;
		changed = addedPtsTargets || changed;

		// set the pointsto parameters
		changed = this.ptsFromArray != fromRef.ptsFromArray || this.ptsTop != fromRef.ptsTop || changed;
		this.ptsFromArray |= fromRef.ptsFromArray;
		this.ptsTop |= fromRef.ptsTop;

		// update the state information
		return super.update(fromRef, strongUpdate) || changed;
	}

	public boolean isRegisteredAtAllocs() {
		return isRegistered;
	}

	public void setRegisteredAtAllocs(boolean isRegistered) {
		this.isRegistered = isRegistered;
	}

	/**
	 * Is called by every allocation site to which this Reference is registered as a user.
	 * 
	 * @param target
	 */
	public void registered(AllocSite site) {

		registeredAt.add(site);
		// Deprecated...
		// // Just increase counter, ignore the site
		// this.numSites++;
	}

	/**
	 * Is called by every allocation site to which this Reference was registered as a user and is now
	 * unregistered.
	 * 
	 * @param target
	 */
	public void unregistered(AllocSite site) {

		registeredAt.remove(site);
		// Deprecated...
		// // Just decrease counter, ignore the site
		// this.numSites--;
	}

	public boolean addPointsToTarget(ContextAndNode target) {
		boolean changed = ptsTargets.add(target);
		isRegistered &= !changed;
		return changed;
	}

	public boolean addPointsToTargets(Collection<ContextAndNode> targets) {
		boolean changed = ptsTargets.addAll(targets);
		isRegistered &= !changed;
		return changed;
	}

	public Set<ContextAndNode> getPointsToTargets() {
		return Collections.unmodifiableSet(ptsTargets);
	}

	/**
	 * @return the ptsFromArray
	 */
	public boolean isPtsFromArray() {
		return this.ptsFromArray;
	}

	/**
	 * @param ptsFromArray
	 *            the ptsFromArray to set
	 */
	public void setPtsFromArray(boolean ptsFromArray) {
		this.ptsFromArray = ptsFromArray;
	}

	/**
	 * @return the ptsTop
	 */
	public boolean isPtsTop() {
		return this.ptsTop;
	}

	/**
	 * @param ptsTop
	 *            the ptsTop to set
	 */
	public void setPtsTop(boolean ptsTop) {
		this.ptsTop = ptsTop;
	}

	public boolean isPtsSingleConcrete() {
		return ptsTop && ptsFromArray && ptsTargets.size() == 1;
	}

	public void disconnect() {
		// Must create copy of registeredAt because otherwise it is modified while being iterated
		ArrayList<AllocSite> regAtCopy = new ArrayList<AllocSite>(registeredAt);

		for (AllocSite alloc : regAtCopy) {
			alloc.removeUser(this);
			// calls unregistered()
		}
		// should not be necessary because removeUser calls unregistered()
		registeredAt.clear();

		isRegistered = ptsTargets.size() == 0;
	}

	public void clearPointsTo() {
		setPtsFromArray(false);
		setPtsTop(false);
		this.ptsTargets = new HashSet<ContextAndNode>();
	}

	@Override
	public String toString() {
		return id + "R" + super.toString()  + " @" + registeredAt + " " + " p("
				+ (isPtsSingleConcrete() ? "S" : "") + (isPtsTop() ? "T" : "")
				+ (isPtsFromArray() ? "A" : "") + "):" + ptsTargets;
	}
	
	@Override
	protected Object clone() {
		Reference clone = null;
		try {
			clone = (Reference) super.clone();
		} catch (CloneNotSupportedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		clone.registeredAt= new HashSet<AllocSite>();
		clone.ptsTargets = new HashSet<ContextAndNode>(this.ptsTargets);
		clone.typeState = new HashSet<IState>(this.typeState);
		//clone.update(this,true);
		clone.id = counter++;
		return clone;
	}

}
