package de.unisb.prog.typestate.util;

import de.unisb.prog.analysis.ContextAndNode;
import de.unisb.prog.jfirm.ir.Call;
import de.unisb.prog.jfirm.ir.Cast;
import de.unisb.prog.jfirm.ir.Constant;
import de.unisb.prog.jfirm.ir.DataMemExec;
import de.unisb.prog.jfirm.ir.EnumTupleNode;
import de.unisb.prog.jfirm.ir.GetField;
import de.unisb.prog.jfirm.ir.GetStatic;
import de.unisb.prog.jfirm.ir.Mode;
import de.unisb.prog.jfirm.ir.NewArray;
import de.unisb.prog.jfirm.ir.NewObject;
import de.unisb.prog.jfirm.ir.Node;
import de.unisb.prog.jfirm.ir.Opcode;
import de.unisb.prog.jfirm.ir.Param;
import de.unisb.prog.jfirm.ir.ProjValueNode;
import de.unisb.prog.jfirm.ir.PutField;
import de.unisb.prog.jfirm.ir.PutStatic;
import de.unisb.prog.jfirm.ir.Type;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.typestate.AnalysisLogger;

/**
 * Provides some static methods useful for handling jFirm nodes.
 * 
 * @author mikerofone
 * 
 */
public abstract class NodeHelper {

	private static final char	NODEHLP_LOG_TAG	= 'N';

	/**
	 * Returns the producer of input <code>index</code> of the given jFirm <code>of</code> node.
	 * 
	 * @param of
	 *            The jFirm node to get the inputs from
	 * @param index
	 *            the index of the desired input
	 * @return the producer of the input node (which might be the node itself, e.g. Phi)
	 */
	public static ValueNode getInput(ValueNode of, int index) {
		Node prod = of.getIn(index).getProducer();
		if (!(prod instanceof ValueNode)) {
			throw new IllegalArgumentException("Producer of node " + of + " was " + prod
					+ " which is not a ValueNode");
		}
		//HACK: ignore casts
		if (prod instanceof Cast) {
			if (AnalysisLogger.getInstance().debug()) {
				AnalysisLogger.getInstance().debug("CAST HACK: skipping " + prod + " while getting Input " + index + " of " + of);
			}
			return getInput((ValueNode) prod,1);
		}
		return (ValueNode) prod;
	}

	public static ContextAndNode getDataNode(ContextAndNode contextNode) {
		return new ContextAndNode(contextNode.getContext(), getDataNode(contextNode.getNode()));
	}

	public static ValueNode getDataNode(ValueNode firmNode) {
		//HACK: Ignore casts as much as possible due to insufficient handling in Ptsto
		if (firmNode instanceof Cast) {
			if (AnalysisLogger.getInstance().debug()) {
				AnalysisLogger.getInstance().debug("CAST HACK: skipping " + firmNode + " while getting DATA Node");
			}
			return getDataNode(firmNode.getIn(1));
		} else if (firmNode.getN() > 0 && firmNode.getIn(0) instanceof Cast) {
			if (AnalysisLogger.getInstance().debug()) {
				AnalysisLogger.getInstance().debug("CAST HACK: skipping parent "+ firmNode.getIn(0) + " of " + firmNode + " while getting DATA Node");
			}
			return getDataNode(firmNode.getIn(0).getIn(1));
		}
			if (firmNode instanceof EnumTupleNode<?>) {
			try {
				@SuppressWarnings("unchecked")
				ValueNode dataNode = ((EnumTupleNode<DataMemExec>) firmNode).getValue(DataMemExec.DATA);
				if (dataNode != null) {
					// Found data node of EnumTupleNode
					return dataNode;
				}
				// strange enum tuple node
				AnalysisLogger.getInstance().warn(
						"getDataFromEnumNode with DataMemExec EnumTuple Node " + firmNode
								+ " which has no PROJ DATA node", NODEHLP_LOG_TAG);
			} catch (ClassCastException e) {
				// strange enum tuple node
				AnalysisLogger.getInstance().warn(
						"getDataFromEnumNode with non-DataMemExec EnumTuple Node " + firmNode,
						NODEHLP_LOG_TAG);
			}
		}
		// The node itself was the data node
		return firmNode;
	}

	/**
	 * Determines if typestate information for a {@link ValueNode} may be assigned or requested.
	 */
	public static boolean isDataNode(ValueNode node) {
		// If it's a EnumTupleNode or similar, fetch the PROJ DATA output.
		node = getDataNode(node);
		return node.getMode() == Mode.A && node.getOpcode() != Opcode.CONST;
	}

	public static Type getType(ValueNode node) {
		switch (node.getOpcode()) {
		case GETFIELD:
			return ((GetField) node).getTargetField().getType();
		case PUTFIELD:
			return ((PutField) node).getTargetField().getType();
		case GETSTATIC:
			return ((GetStatic) node).getTargetField().getType();
		case PUTSTATIC:
			return ((PutStatic) node).getTargetField().getType();
		case CALL:
			return ((Call) node).getTargetMethod().getReturnType();
		case ALOAD:
		case ASTORE:
		case THROW:
			return getType(node.getIn(1));
		case NEWARRAY:
			return ((NewArray) node).getElementType();
		case NEW:
			return ((NewObject) node).getObjectType();
		case CONST:
			Object val = ((Constant) node).getValue();
			return Type.getByDesc(val == null ? "Ljava/lang/Object;" : "L" + val.getClass().getName().replace('.', '/') + ";");
		case PHI:
			// FIXME this is wrong for PHI, but unless there is no correct type hierarchy just take the first
			// type you find
		case RETURN:
			return getType(node.getIn(0));
		case CAST:
			return ((Cast) node).getClassType();
		case PARAM:
			return ((Param) node).getType();
		case PROJ:
			return getType(((ProjValueNode<?>) node).getProducer());

		default:
			AnalysisLogger.getInstance().error(
					"NodeHelper.getType(" + node + "): Missing rule for Opcode " + node.getOpcode());
			return null;
		}
	}
}
