package de.unisb.prog.typestate.util;

import de.unisb.prog.jfirm.ir.Param;
import de.unisb.prog.jfirm.ir.Type;
import de.unisb.prog.jfirm.ir.ValueNode;
import de.unisb.prog.jfirm.ir.Type.ReferenceType;
import de.unisb.prog.typestate.AnalysisLogger;

public abstract class Types {

	/**
	 * Returns the fully qualified method name
	 * 
	 * @param method
	 *            the method
	 * @return the signature as String
	 */
	public static final String getFullName(de.unisb.prog.jfirm.ir.Type.ClassType.Method method) {
		if (method == null) {
			AnalysisLogger.getInstance().error("Types.getFullName(method==null) might never happen");
			return "";
		}
		return method.getOwner().getName() + "/" + getSignature(method);
	}

	/**
	 * Returns the textual representation of a method-type's signature
	 * 
	 * @param method
	 *            the method
	 * @return the signature as String
	 */
	public static final String getFullShortName(de.unisb.prog.jfirm.ir.Type.ClassType.Method method) {
		if (method == null) {
			AnalysisLogger.getInstance().error("Types.getFullShortName(method==null) might never happen");
			return "";
		}
		StringBuilder sb = new StringBuilder(shortName(method.getOwner().toString()));
		sb.append("." + method.getName() + "(");
		for (Type par : method.getParameters()) {
			sb.append(shortName(bytecodeify(par)));
		}
		sb.append(")");
		// sb.append(shortName(method.getReturnType().toString()));
		return sb.toString();
	}

	/**
	 * Returns the textual representation of a method-type's signature
	 * 
	 * @param method
	 *            the method
	 * @return the signature as String
	 */
	public static final String getSignature(de.unisb.prog.jfirm.ir.Type.ClassType.Method methodType) {
		if (methodType == null) {
			AnalysisLogger.getInstance().error("Types.getSignature(methodType==null) might never happen");
			return "";
		}
		StringBuilder sb = new StringBuilder(methodType.getName());
		sb.append("(");
		for (Type par : methodType.getParameters()) {
			sb.append(bytecodeify(par));
		}
		sb.append(")");
		sb.append(bytecodeify(methodType.getReturnType()));
		return sb.toString();
	}

	/**
	 * Returns the textual representation of a jFirm-method's signature
	 * 
	 * @param method
	 *            the method
	 * @return the signature as String
	 */
	public static final String getSignature(de.unisb.prog.jfirm.ir.MethodGraph firmMethod) {
		if (firmMethod == null) {
			AnalysisLogger.getInstance().error("Types.getSignature(firmMethod==null) might never happen");
			return "";
		}
		// FIXME add return value and class path
		StringBuilder sb = new StringBuilder(firmMethod.getName());
		sb.append("(");
		ValueNode[] pars = firmMethod.getParameters();
		for (int i = 1, n = pars.length; i < n; i++) {
			assert pars[i] instanceof Param : "a parameter of " + firmMethod.getName() + " is no PARAM instance";

			Param par = (Param) pars[i];
			sb.append(bytecodeify(par.getType()));
		}
		sb.append(")");
		return sb.toString();
	}

	/**
	 * Returns the bytecode representation of a Firm <code>Type</code>.
	 * 
	 * @param type
	 *            the Type
	 * @return the textual representation of the <code>Type</code>
	 */
	public static String bytecodeify(Type type) {
		if (type == null) {
			AnalysisLogger.getInstance().error("Types.bytecodify(type==null) might never happen");
			return "";
		}
		if (type instanceof de.unisb.prog.jfirm.ir.Type.ClassType.ArrayType) {
			// ArrayType aType = (de.unisb.prog.jfirm.ir.Type.ClassType.ArrayType) type;
			// return "[" + bytecodeify(aType.getElement());
			// toString() is already correct
			return ((de.unisb.prog.jfirm.ir.Type.ClassType.ArrayType) type).getName();
		} else if (type instanceof de.unisb.prog.jfirm.ir.Type.ClassType.ReferenceType) {
			ReferenceType rType = (de.unisb.prog.jfirm.ir.Type.ClassType.ReferenceType) type;
			return "L" + rType.getName() + ";";
		} else {
			return type.getMode().name();
		}
	}

	/**
	 * Strips away package information in the given typename.
	 * 
	 * @param typeName
	 * @return
	 */
	public static String shortName(String typeName) {
		if (typeName == null) {
			AnalysisLogger.getInstance().error("Types.shortName(method==null) might never happen");
			return "";
		}
		int startIndex = typeName.lastIndexOf('/');
		return startIndex == -1 ? typeName : typeName.substring(startIndex + 1);
	}

	/**
	 * Strips away package information in the given typename.
	 * 
	 * @param typeName
	 * @return
	 */
	public static String methodName(String bytecodeName) {
		if (bytecodeName == null) {
			AnalysisLogger.getInstance().error("Types.getFullShortName(bytecodeName==null) might never happen");
			return "";
		}
		int startIndex = bytecodeName.indexOf('.');
		if (startIndex < 0)
			startIndex = 0;
		int endIndex = bytecodeName.indexOf('(');
		if (endIndex < 0) {
			endIndex = bytecodeName.length();
		}
		return bytecodeName.substring(startIndex + 1, endIndex);
	}

	public static String paramRetTypes(String bytecodeName) {
		if (bytecodeName == null) {
			AnalysisLogger.getInstance().error("Types.paramRetTypes(bytecodeName==null) might never happen");
			return "";
		}
		int startIndex = bytecodeName.indexOf('(');
		if (startIndex < 0) {
			throw new IllegalArgumentException("given string is not a valid bytecode name for a method: "
					+ bytecodeName);
		}
		return bytecodeName.substring(startIndex);
	}

	public static String getClassName(String bytecodeName) {
		if (bytecodeName == null) {
			AnalysisLogger.getInstance().error("Types.getClassName(bytecodeName==null) might never happen");
			return "";
		}
		int endIndex = bytecodeName.indexOf('.');
		if (endIndex < 0) {
			throw new IllegalArgumentException("given string is not a valid bytecode name for a method: "
					+ bytecodeName);
		}
		return bytecodeName.substring(0, endIndex);
	}
}
