package de.unisb.prog.typestate.violation;

import java.util.HashSet;
import java.util.Set;

/**
 * 
 * @author mikerofone
 * 
 */
public class AlternateMethodsSuggestion implements ISuggestion {

	/** Methods which take instance to a state where the initial transition is valid without potential errors */
	public Set<String>	safeEnablingTrans;

	/**
	 * Methods which take instance to a state where the initial transition is valid but may cause
	 * non-deterministic error transitions
	 */
	public Set<String>	unsafeEnablingTrans;

	/** Methods which do not leave the current state */
	public Set<String>	sameState;

	/** Methods which change the current state AND might lead to an error */
	public Set<String>	changing;

	/** Methods which safely change to another state (no error possible) */
	public Set<String>	safeChanging;

	/**
	 * Creates an empty suggestion which can be extended by using the getters and then adding entries
	 */
	public AlternateMethodsSuggestion() {
		this(new HashSet<String>(),new HashSet<String>(),new HashSet<String>(),new HashSet<String>(),new HashSet<String>());
	}
	
	/**
	 * @param safeEnablingTrans
	 *            methods which take instance to a state where the initial transition is valid without
	 *            potential errors
	 * @param unsafeEnablingTrans
	 *            methods which take instance to a state where the initial transition is valid but may cause
	 *            non-deterministic error transitions
	 * @param safeChanging
	 *            methods which change the state WITHOUT possibly leading to the error state
	 * @param changing
	 *            methods which change the state AND possibly lead to error transitions
	 * @param sameState
	 *            methods which do not change the current state (which might be quite useless)
	 */
	public AlternateMethodsSuggestion(Set<String> safeEnablingTrans, Set<String> unsafeEnablingTrans,
			Set<String> safeChanging, Set<String> changing, Set<String> sameState) {
		this.safeEnablingTrans = safeEnablingTrans;
		this.unsafeEnablingTrans = unsafeEnablingTrans;
		this.safeChanging = safeChanging;
		this.changing = changing;
		this.sameState = sameState;

	}

	
	
	/**
	 * @return the set of methods safely enabling the initially wanted transition
	 */
	public Set<String> getSafeEnablingTrans() {
		return this.safeEnablingTrans;
	}



	/**
	 * @return the set of methods unsafely enabling the initially wanted transition
	 */
	public Set<String> getUnsafeEnablingTrans() {
		return this.unsafeEnablingTrans;
	}



	/**
	 * @return the set of methods which do not leave the current state
	 */
	public Set<String> getSameState() {
		return this.sameState;
	}

	/**
	 * @return the set of methods which change the current state AND might lead to an error
	 */
	public Set<String> getChanging() {
		return this.changing;
	}

	/**
	 * @return the set of methods which safely change to another state (no error possible)
	 */
	public Set<String> getSafeChanging() {
		return this.safeChanging;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.unisb.prog.typestate.violation.IViolationInfo#getType()
	 */
	public SuggestionType getType() {
		return SuggestionType.ALTERNATE_METHODS;
	}

	/**
	 * The textual representation returns contents of all three sets grouped together
	 * 
	 * @see de.unisb.prog.typestate.violation.IViolationInfo#text()
	 */
	public String text() {
		StringBuilder sb = new StringBuilder();
		if (!safeEnablingTrans.isEmpty()) {
			sb.append("Methods safely enabling the wanted method: ");
			sb.append(safeEnablingTrans.toString());
		}
		if (!unsafeEnablingTrans.isEmpty()) {
			if (sb.length() > 0 && sb.charAt(sb.length() - 1) != '\n')
				sb.append('\n');
			sb.append("Methods unsafely enabling the wanted method: ");
			sb.append(unsafeEnablingTrans.toString());
		}
		if (!safeChanging.isEmpty()) {
			if (sb.length() > 0 && sb.charAt(sb.length() - 1) != '\n')
				sb.append('\n');
			sb.append("Methods safely changing states: ");
			sb.append(safeChanging.toString());
		}
		if (!changing.isEmpty()) {
			if (sb.length() > 0 && sb.charAt(sb.length() - 1) != '\n')
				sb.append('\n');
			sb.append("Methods not safely changing state: ");
			sb.append(changing.toString());
		}
		if (!sameState.isEmpty()) {
			if (sb.length() > 0 && sb.charAt(sb.length() - 1) != '\n')
				sb.append('\n');
			sb.append("Methods not changing state: ");
			sb.append(sameState.toString());
		}
		String msg = sb.toString();
		return msg; //"".equals(msg) ? "No method propositions available" : msg;
	}

	public Object getSuggestions() {
		// TODO Auto-generated method stub
		return null;
	}

}
